#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include <dbus/dbus-glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glade/glade.h>

#include "bmpx/dbus.h"

#define DBUS_TYPE_G_STRING_VALUE_HASHTABLE (dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_VALUE))

gboolean enqueue = FALSE;
gboolean play    = FALSE;

static GOptionEntry bmpx_remote_options[] =
{
    {"enqueue", 'e', 0, G_OPTION_ARG_NONE, &enqueue,
     N_("Enqueue files to BMPx"), NULL},
    {"play", 'p', 0,   G_OPTION_ARG_NONE,  &play,
     N_("Play files in BMPx"), NULL},
    { NULL }
};

static  DBusGConnection *bus   = NULL;
static  DBusGProxy	*bmpx  = NULL;
static  GError		*error = NULL;

static	gint		 _n_args = 0;
static  gchar**		 _args   = NULL;

static void
setup_combo_box_lists (GtkWidget *widget)
{
  gchar **array_lists;
  gint    current_list;
  gint	  n = 0;

  if (!dbus_g_proxy_call (bmpx, "GetTracklists", &error,
			  G_TYPE_INVALID,
			  G_TYPE_STRV, 
			  &array_lists, 
			  G_TYPE_INVALID))
    {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "failed to complete GetTracklists: %s", error->message);
	exit(EXIT_FAILURE);
    }

  if (!dbus_g_proxy_call (bmpx, "GetCurrentTracklistNumber", &error,
			  G_TYPE_INVALID,
			  G_TYPE_INT, 
			  &current_list, 
			  G_TYPE_INVALID))
    {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "failed to complete GetTracklists: %s", error->message);
	exit(EXIT_FAILURE);
    }

    while (array_lists[n])
      {
	  gtk_combo_box_append_text (GTK_COMBO_BOX(widget), array_lists[n]); 
	  n++;
      }

    gtk_combo_box_set_active (GTK_COMBO_BOX(widget), current_list);
}

void
on_dialog_enqueue_response (GtkDialog *dialog,
			    gint arg1,
			    GtkWidget *widget)
{
  gchar **uri_list;
  gint	  tracklist, position, n , idx;

  if (arg1 == GTK_RESPONSE_CANCEL)
    exit (EXIT_SUCCESS);

  tracklist = gtk_combo_box_get_active (GTK_COMBO_BOX(widget)); 
  position  = -1; 

  uri_list = g_new (gchar*, _n_args);

  for (n = 1; n < _n_args; n++)
    {
      uri_list[n-1] = _args[n];
    }
  uri_list[n] = NULL;


  if (!dbus_g_proxy_call (bmpx, "AddUriList", &error,
			  G_TYPE_STRV, uri_list,
			  G_TYPE_INT, tracklist, 
			  G_TYPE_INT, position, 
			  G_TYPE_INVALID,
			  G_TYPE_INT, &idx,
			  G_TYPE_INVALID)) 
    {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Couldn't run AddUriList: %s", error->message);
	exit (EXIT_FAILURE);
    }

  exit (EXIT_SUCCESS);
}

gint
main (gint n_args, gchar **args)
{
    GladeXML  *xml;
    GtkWidget *dialog;
    GtkWidget *cbox;
    GtkWidget *label;
    gchar     *text;

    gtk_init_with_args (&n_args, &args, "", &bmpx_remote_options, NULL, NULL);
    dbus_g_type_specialized_init ();

    _n_args = n_args;
    _args = args;

    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (!bus)
      {
  	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Couldn't connect to session bus: %s", error->message);
	exit (EXIT_FAILURE);
      }
  
    bmpx = dbus_g_proxy_new_for_name (bus,
				      BMP_DBUS_SERVICE,
				      BMP_DBUS_PATH_SYSTEMCONTROL,
				      BMP_DBUS_INTERFACE);

    if (enqueue)
      {
	xml = glade_xml_new (DATA_DIR "/glade/bmpx-remote.glade", NULL, NULL);
        if (!xml)
	  exit (EXIT_FAILURE);

	label = glade_xml_get_widget (xml, "label1");
	text = g_strdup_printf ("Enqueue <b>%d</b> %s To:", n_args-1, ((n_args-1) == 1) ? "Item" : "Items");
	gtk_label_set_markup (label, text);
	g_free (text);

        cbox = glade_xml_get_widget (xml, "cbox_lists");
	gtk_combo_box_remove_text (GTK_COMBO_BOX(cbox), 0);
        setup_combo_box_lists (cbox);

	dialog = glade_xml_get_widget (xml, "dialog_enqueue");
        g_object_connect (G_OBJECT(dialog),
		      "signal::response",
		      G_CALLBACK(on_dialog_enqueue_response),
		      cbox,
		      NULL);

	gtk_dialog_run (GTK_DIALOG (dialog));
      }

    gtk_main ();

    exit (EXIT_SUCCESS);
}
