/*  libhrel - Holyrel library

    Copyright (C) 2005 Chong Kai Xiong <descender@phreaker.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the
    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA  02111-1307  USA.
*/

#include <config.h>

#include "hash_index.h"
#include "hash_table.h"
#include "misc.h"

struct _HHashIndex
{
  HHashTable *hash_table;
  GSList     *list;
};

HHashIndex *
h_hash_index_new (GDestroyNotify  key_destroy_func,
                  GDestroyNotify  value_destroy_func,
                  ...)
{
  va_list     args;
  HHashIndex *index;
  GSList     *list;

  va_start (args, value_destroy_func);
  list = h_va_list_to_slist (args);
  va_end (args);

  index = h_hash_index_new_l (key_destroy_func, value_destroy_func, list);
  g_slist_free (list);

  return index;
}


HHashIndex *
h_hash_index_new_l (GDestroyNotify  key_destroy_func,
                    GDestroyNotify  value_destroy_func,
                    GSList         *list)
{
  HHashIndex *index;

  index = g_new (HHashIndex, 1);

  /* copy attribute name list */
  index->list = NULL;
  while (list)
    {
      index->list = g_slist_append (index->list, g_strdup ((const gchar *) list->data));
      list = list->next;
    }

  index->hash_table = h_hash_table_new_full ((HHashFunc) h_tuple_hash_elems_l,
                                             index->list,
                                             (HEqualFunc) h_tuple_equal_elems_l,
                                             index->list,
                                             key_destroy_func,
                                             value_destroy_func);

  return index;
}

void
h_hash_index_free (HHashIndex *index)
{
  g_return_if_fail (index != NULL);

  g_slist_foreach (index->list, (GFunc) g_free, NULL);
  g_slist_free (index->list);

  h_hash_table_destroy (index->hash_table);

  g_free (index);
}

GSList *
h_hash_index_get_attrs (HHashIndex *index)
{
  g_return_val_if_fail (index != NULL, NULL);

  return index->list;
}

gpointer
h_hash_index_lookup (HHashIndex *index,
                     HTuple     *tuple)
{
  g_return_val_if_fail (index != NULL, NULL);

  return h_hash_table_lookup (index->hash_table, tuple);
}

void
h_hash_index_insert (HHashIndex *index,
                     HTuple     *tuple,
                     gpointer    value)
{
  g_return_if_fail (index != NULL);

  h_hash_table_insert (index->hash_table, tuple, value);
}

gboolean
h_hash_index_remove (HHashIndex *index,
                     HTuple     *tuple)
{
  g_return_val_if_fail (index != NULL, FALSE);

  return h_hash_table_remove (index->hash_table, tuple);
}

guint
h_hash_index_get_size (HHashIndex *index)
{
  g_return_val_if_fail (index != NULL, 0);

  return h_hash_table_size (index->hash_table);
}

void
h_hash_index_foreach (HHashIndex *index,
                      HIndexFunc  func,
                      gpointer    user_data)
{
  g_return_if_fail (index != NULL);

  h_hash_table_foreach (index->hash_table, (HHFunc) func, user_data);
}
