#ifndef _STORE_STORE_H
#define _STORE_STORE_H

#include <glib-object.h>
#include <bmpx/list.h>

#include "bstore.h"

#define B_TYPE_STORE		  (b_store_get_type ())
#define B_STORE(obj)		  (G_TYPE_CHECK_INSTANCE_CAST ((obj), B_TYPE_STORE, BStore))
#define B_STORE_CLASS(klass)	  (G_TYPE_CHECK_CLASS_CAST ((klass), B_TYPE_STORE, BStoreClass))
#define B_IS_STORE(obj)		  (G_TYPE_CHECK_INSTANCE_TYPE ((obj), B_TYPE_STORE))
#define B_IS_STORE_CLASS(klass)	  (G_TYPE_CHECK_CLASS_TYPE ((klass), B_TYPE_STORE))
#define B_STORE_GET_CLASS(obj)	  (G_TYPE_INSTANCE_GET_CLASS ((obj), B_TYPE_STORE, BStoreClass))

#define B_TYPE_STORE_ROW	  (b_store_row_get_type ())

typedef struct _BStore BStore;
typedef struct _BStoreClass BStoreClass;
typedef struct _BStorePrivate BStorePrivate;

struct _BStore {
    GObject parent;

    /*
     * instance members 
     */
    BStorePrivate *private;
};

struct _BStoreClass {
    GObjectClass parent;

};

typedef struct _BStoreRow BStoreRow;
struct _BStoreRow {
    gchar        *uri;          /* file URI */
    RowGUID	 *guid;
};

/*StoreRow stuff*/
GType
b_store_row_get_type (void);

void
b_store_row_free (BStoreRow *store_row);

BStoreRow*
b_store_row_copy (BStoreRow *store_row);

BStoreRow*
b_store_row_new (const gchar *uri);
#define B_STORE_ROW(x) ((BStoreRow*)x)

/*Store stuff*/
GType
b_store_get_type (void);

BStore*
b_store_new (const gchar *name);

time_t
b_store_get_time (BStore *self);

const gchar*
b_store_get_name (BStore *self);
 
BStoreRow*
b_store_row_fetch (BStore *self,
		   GtkTreeIter *iter);

void
b_store_row_append (BStore *self, 
		    BStoreRow *row);

BStoreRow*
b_store_row_get (BStore *self,
		 GtkTreeIter *iter);

BStoreRow*
b_store_row_get_by_index (BStore *self,
			  gint _index);

gint
b_store_get_n_rows (BStore *self);

gint
b_store_iter_get_index (BStore *self, GtkTreeIter *iter);

gboolean
b_store_get_iter_next (BStore *self,
			    GtkTreeIter *iter);
gboolean
b_store_get_iter_prev (BStore *self,
			    GtkTreeIter *iter);
gboolean
b_store_get_iter_current (BStore *self,
			       GtkTreeIter *iter);

void
b_store_store_clear (BStore *self);

void
b_store_row_add (BStore *self, BStoreRow *row);

void
b_store_remove_rows (BStore *self, gint *rows);

#if 0
gboolean
b_store_store_row_remove(BStore *self, GtkTreeIter * iter);

gboolean
b_store_store_row_move_down(BStore *self, GtkTreeIter * iter);

gboolean
b_store_store_row_move_up(BStore *self, GtkTreeIter * iter);
#endif

#endif
