/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Chroma widget set
 * (C) 2005 M. Derezynski <internalerror@gmail.com> 
 */
 
#include <gtk/gtk.h>

#include "ui_util.h"
#include "marshalers.h"

#include "goa/libgoa.h"

#include "chroma_button.h"

typedef enum {
	CHROMA_BUTTON_SIGNAL_CLICKED,
	CHROMA_BUTTON_SIGNAL_EXPOSE_REQUEST,

	CHROMA_BUTTON_N_SIGNALS
} ChromaButtonSignals;
static guint chroma_button_signals[CHROMA_BUTTON_N_SIGNALS] = { 0, 0 };

typedef enum {
	C_BUTTON_PROP_X = 1,
	C_BUTTON_PROP_Y,
	C_BUTTON_PROP_W,
	C_BUTTON_PROP_H,
	
	C_BUTTON_N_PROPERTIES
} ChromaButtonProperties;

struct _ChromaButtonPrivate
{
        gboolean dispose_has_run;

	gint position_x;
	gint position_y;
	gint width;
	gint height;

	gboolean active;
	gboolean inside;

	GdkPixmap *pixmap;
};

static void
chroma_button_instance_init (GTypeInstance   *instance,
                         gpointer         g_class)
{
        ChromaButton *self = (ChromaButton *)instance;
        self->private = g_new (ChromaButtonPrivate, 1);
        self->private->dispose_has_run = FALSE;
	self->private->pixmap = NULL;
	self->private->active = FALSE;
	self->private->inside = FALSE;
}

static void
chroma_button_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{
        

        switch (property_id) {

	case C_BUTTON_PROP_X:
	{
		CHROMA_BUTTON(object)->private->position_x = g_value_get_int (value);
	}
	break;

	case C_BUTTON_PROP_Y:
	{
		CHROMA_BUTTON(object)->private->position_y = g_value_get_int (value);
	}
	break;

	case C_BUTTON_PROP_W:
	{
		CHROMA_BUTTON(object)->private->width = g_value_get_int (value);
	}
	break;

	case C_BUTTON_PROP_H:
	{
		CHROMA_BUTTON(object)->private->height = g_value_get_int (value);
	}
	break;

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static void
chroma_button_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{
        

        switch (property_id) {

 	case C_BUTTON_PROP_X:
	{
		g_value_set_int (value, CHROMA_BUTTON(object)->private->position_x); 
	}
	break;

	case C_BUTTON_PROP_Y:
	{
		g_value_set_int (value, CHROMA_BUTTON(object)->private->position_y);
	}
	break;

	case C_BUTTON_PROP_W:
	{
		g_value_set_int (value, CHROMA_BUTTON(object)->private->width);
	}
	break;

	case C_BUTTON_PROP_H:
	{
		g_value_set_int (value, CHROMA_BUTTON(object)->private->height);
	}
	break;

       default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static GObject *
chroma_button_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                ChromaButtonClass *klass;
                GObjectClass *parent_class;  
                klass = CHROMA_BUTTON_CLASS (g_type_class_peek (CHROMA_TYPE_BUTTON));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }
        
        /* do stuff. */

        return obj;
}

static void
chroma_button_dispose (GObject *obj)
{
        ChromaButton *self = (ChromaButton *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /* 
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a 
         * reference.
         */
}

static void
chroma_button_finalize (GObject *obj)
{
        ChromaButton *self = (ChromaButton *)obj;

        /*
         * Here, complete object destruction.
         * You might not need to do much...
         */
        g_free (self->private);
}

GObject*
chroma_button_new (gint position_x, gint position_y, gint width, gint height)
{
  ChromaButton *self;
  self = g_object_new (chroma_button_get_type (), "pos_x", position_x, "pos_y", position_y, "width", width, "height", height, NULL);
 
  return G_OBJECT (self);
}

void
chroma_button_renew (ChromaButton *self, GtkWidget *widget, unsigned char data[])
{
	GdkColor *fg_color, *fg_current, *bg_normal, *bg_selected;
	GdkGC *gc_local;

	gc_local  = gdk_gc_new (GDK_DRAWABLE(widget->window));

	fg_color    = bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_NORMAL);
	fg_current  = bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT);
	bg_normal    = bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL);
	bg_selected = bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_SELECTED);

        gdk_color_alloc(gdk_window_get_colormap(widget->window), bg_normal); 
        gdk_color_alloc(gdk_window_get_colormap(widget->window), fg_current); 

	self->private->pixmap = gdk_pixmap_create_from_data (widget->window,
					 	(const gchar*)data,	
						16, 16, 24, bg_normal, fg_current);						

#if 0
	gdk_gc_set_rgb_fg_color (gc_local, bg_normal);
	gdk_draw_point (self->private->pixmap,
			gc_local,
			0, 0);
	
	gdk_draw_point (self->private->pixmap,
			gc_local,
			0, 15);

	gdk_draw_point (self->private->pixmap,
			gc_local,
			15, 0);

	gdk_draw_point (self->private->pixmap,
			gc_local,
			15, 15);
#endif
}

void
chroma_button_inside(ChromaButton *self, gint x,  gint y) 
{
	self->private->inside =	((x >= self->private->position_x) &&
		 (x <= (self->private->position_x + self->private->width)) &&
		 (y >= self->private->position_y) && 
		 (y <= (self->private->position_y + self->private->height))); 
}

gboolean
chroma_button_dispatch_button_release (ChromaButton *self, gint event_x, gint event_y)
{
  gint control_y;

  chroma_button_inside (self, event_x, event_y);

  control_y = event_y - self->private->position_y;

  if (self->private->active == TRUE) self->private->active = FALSE;

  g_signal_emit (self, chroma_button_signals[CHROMA_BUTTON_SIGNAL_EXPOSE_REQUEST], 0,
		g_object_get_int (self, "pos_x"),
		g_object_get_int (self, "pos_y"),
		g_object_get_int (self, "width") + 1,
		g_object_get_int (self, "height") + 1
	  );

  return self->private->inside;
}



gboolean
chroma_button_dispatch_button_press (ChromaButton *self, gint event_x, gint event_y, gboolean active, guint modifier_state)
{
  gint control_y;

  chroma_button_inside (self, event_x, event_y);

  control_y = event_y - self->private->position_y;

  self->private->active = TRUE;

  g_signal_emit (self, chroma_button_signals[CHROMA_BUTTON_SIGNAL_EXPOSE_REQUEST], 0,
		g_object_get_int (self, "pos_x"),
		g_object_get_int (self, "pos_y"),
		g_object_get_int (self, "width") + 1,
		g_object_get_int (self, "height") + 1
	  );

  return self->private->inside;
}


gboolean
chroma_button_dispatch_motion (ChromaButton *self, gint event_x, gint event_y)
{
  gint control_y;

  chroma_button_inside (self, event_x, event_y); 

  control_y = event_y - self->private->position_y;

  g_signal_emit (self, chroma_button_signals[CHROMA_BUTTON_SIGNAL_EXPOSE_REQUEST], 0,
		g_object_get_int (self, "pos_x"),
		g_object_get_int (self, "pos_y"),
		g_object_get_int (self, "width") + 1,
		g_object_get_int (self, "height") + 1
	  );

  return self->private->inside;
}

void
chroma_button_draw (ChromaButton *self, GtkWidget *widget)
{
	GdkColor *fg_color, *fg_current, *bg_normal, *bg_selected;
	GdkGC *gc_local;
	gint offset, x, y;

	g_return_if_fail (CHROMA_IS_BUTTON(self));

	gc_local  = gdk_gc_new (GDK_DRAWABLE(widget->window));

	fg_color    = bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_NORMAL);
	fg_current  = bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT);
	bg_normal    = bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL);
	bg_selected = bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_SELECTED);

	gdk_gc_set_rgb_fg_color (gc_local, fg_color);
	gdk_gc_set_rgb_bg_color (gc_local, bg_normal);

	x = g_object_get_int (self, "pos_x");
	y = g_object_get_int (self, "pos_y");

	if ( ((self->private->active) && !(self->private->inside)) || (!self->private->active) )
	{
		gdk_gc_set_rgb_fg_color(gc_local, bmp_colors_get_color (BMP_SKINCOLOR_PL_BG_SELECTED)); 
        	gdk_draw_rectangle(widget->window,
			gc_local,
                        TRUE,
			g_object_get_int (self, "pos_x") + 1,
			g_object_get_int (self, "pos_y") + 1,
			g_object_get_int (self, "width"),
			g_object_get_int (self, "height"));

#if 0
		gdk_gc_set_rgb_fg_color(gc_local, bmp_colors_get_color (BMP_SKINCOLOR_PL_BG_NORMAL)); 
		gdk_draw_point (widget->window,
			gc_local,
			x+1, y+1);

		gdk_draw_point (widget->window,
			gc_local,
			x+16, y+16);

		gdk_draw_point (widget->window,
			gc_local,
			x+1, y+16);

		gdk_draw_point (widget->window,
			gc_local,
			x+16, y+1);
#endif
	}

	offset = ( (self->private->active == TRUE) && (self->private->inside) )? 1 : 0;
	gdk_draw_drawable (widget->window, gc_local, self->private->pixmap, 
			0, 0, 
			g_object_get_int (self, "pos_x") + offset,
			g_object_get_int (self, "pos_y") + offset,
			g_object_get_int (self, "width"),
			g_object_get_int (self, "height")
			);

#if 0
	if ( ((self->private->active) && !(self->private->inside)) || (!self->private->active) )
	{
		gdk_gc_set_rgb_fg_color(gc_local, bmp_colors_get_color (BMP_SKINCOLOR_PL_BG_SELECTED)); 
		gdk_draw_point (widget->window,
			gc_local,
			x+15, y+15);

	}
#endif

}

static void
chroma_button_class_init (gpointer g_class,
                      gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
        
        GParamSpec *pspec;

        gobject_class->set_property = chroma_button_set_property;
        gobject_class->get_property = chroma_button_get_property;
        gobject_class->dispose = chroma_button_dispose;
        gobject_class->finalize = chroma_button_finalize;
        gobject_class->constructor = chroma_button_constructor;

        pspec = g_param_spec_int ("pos_x",
                                  "pos_x",
                                  "pos_x", 0, G_MAXINT, 0,
                                  G_PARAM_READWRITE);
        g_object_class_install_property (gobject_class,
                                         C_BUTTON_PROP_X,
                                         pspec);

        pspec = g_param_spec_int ("pos_y",
                                  "pos_y",
                                  "pos_y", 0, G_MAXINT, 0,
                                  G_PARAM_READWRITE);
        g_object_class_install_property (gobject_class,
                                         C_BUTTON_PROP_Y,
                                         pspec);

        pspec = g_param_spec_int ("width",
                                  "width",
                                  "width", 0, G_MAXINT, 0,
                                  G_PARAM_READWRITE);
        g_object_class_install_property (gobject_class,
                                         C_BUTTON_PROP_W,
                                         pspec);

        pspec = g_param_spec_int ("height",
                                  "height",
                                  "height", 0, G_MAXINT, 0,
                                  G_PARAM_READWRITE);
        g_object_class_install_property (gobject_class,
                                         C_BUTTON_PROP_H,
                                         pspec);

        chroma_button_signals[CHROMA_BUTTON_SIGNAL_EXPOSE_REQUEST] =
        g_signal_new ("expose_request",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (ChromaButtonClass, expose_request),
                  NULL, NULL,
                  bmpx_VOID__INT_INT_INT_INT,
                  G_TYPE_NONE, 4, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);

        chroma_button_signals[CHROMA_BUTTON_SIGNAL_CLICKED] =
        g_signal_new ("clicked",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (ChromaButtonClass, clicked),
                  NULL, NULL,
                  g_cclosure_marshal_VOID__VOID,
                  G_TYPE_NONE, 0);
}

GType
chroma_button_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (ChromaButtonClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        chroma_button_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (ChromaButton),
                        0,      /* n_preallocs */
                        chroma_button_instance_init    /* instance_init */
                };
                type = g_type_register_static (G_TYPE_OBJECT,
                                               "ChromaButtonType",
                                               &info, 0);
        }
        return type;
}
