/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * Chroma widget set
 * (C) 2005 M. Derezynski <internalerror@gmail.com>
 */

#include <gtk/gtk.h>
#include <math.h>

#include "chroma_vscrollbar.h"
#include "chroma_list.h"

#include "ui.h"
#include "ui_util.h"

#include "marshalers.h"
#include "goa/libgoa.h"

#define  SCROLLBAR_GRIP_MINIMAL_SIZE 12

typedef enum
{
	AREA_NONE,
	AREA_ARROW_UP,
	AREA_GRIP,
	AREA_ARROW_DOWN

} ChromaVScrollbarActiveArea;


typedef enum {
        CHROMA_VSCROLLBAR_SIGNAL_POSITION,
	CHROMA_VSCROLLBAR_SIGNAL_EXPOSE_REQUEST,

	CHROMA_VSCROLLBAR_N_SIGNALS
} ChromaVScrollbarSignals;
static guint chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_N_SIGNALS] = { 0, 0 };

typedef enum {
	C_VSCROLL_PROP_X = 1,
	C_VSCROLL_PROP_Y,
	C_VSCROLL_PROP_W,
	C_VSCROLL_PROP_H,
	C_VSCROLL_PROP_GRIP_HEIGHT,
	C_VSCROLL_PROP_POSITION,

	C_VSCROLL_N_PROPERTIES
} ChromaVScrollbarProperties;

struct _ChromaVScrollbarPrivate
{
        gboolean dispose_has_run;

	GtkAllocation allocation;

	gint    position_x, position_y;
	gint    height, width;
	gdouble scroll_position;
	gdouble grip_height;

	gint     active_area;
	gboolean active;
	gboolean inside;

	gint     last_y;
	gint 	 item_size;
};

void
chroma_vscrollbar_emit_expose_request (ChromaVScrollbar *self_arg);

void
chroma_vscrollbar_reset (ChromaVScrollbar *self)
{
	g_return_if_fail (CHROMA_IS_VSCROLLBAR(self));

	self->private->scroll_position = 0;
	chroma_vscrollbar_emit_expose_request (self);
}

static void
chroma_vscrollbar_instance_init (GTypeInstance   *instance,
                         gpointer         g_class)
{
        ChromaVScrollbar *self = (ChromaVScrollbar *)instance;
        self->private = g_new (ChromaVScrollbarPrivate, 1);
        self->private->dispose_has_run = FALSE;
	self->private->scroll_position = 0;
	self->private->grip_height = 0;
	self->private->active = FALSE;
	self->private->active_area = AREA_NONE;
}


static void
chroma_vscrollbar_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec)
{


        switch (property_id) {

	case C_VSCROLL_PROP_POSITION:
	{
		gdouble value_d;

		value_d = g_value_get_double (value);

		CHROMA_VSCROLLBAR(object)->private->scroll_position = value_d;
		g_signal_emit (object, chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_POSITION], 0, CHROMA_VSCROLLBAR(object)->private->scroll_position);
		chroma_vscrollbar_emit_expose_request(CHROMA_VSCROLLBAR(object));
	}
	break;

	case C_VSCROLL_PROP_GRIP_HEIGHT:
	{
		CHROMA_VSCROLLBAR(object)->private->grip_height = ceil (g_value_get_double (value)+0.5);
	}
	break;

        default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static void
chroma_vscrollbar_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec)
{


        switch (property_id) {

 	case C_VSCROLL_PROP_POSITION:
	{
		g_value_set_double (value, CHROMA_VSCROLLBAR(object)->private->scroll_position);
	}
	break;

 	case C_VSCROLL_PROP_X:
	{
		g_value_set_int (value, CHROMA_VSCROLLBAR(object)->private->allocation.x);
	}
	break;

	case C_VSCROLL_PROP_Y:
	{
		g_value_set_int (value, CHROMA_VSCROLLBAR(object)->private->allocation.y);
	}
	break;

	case C_VSCROLL_PROP_W:
	{
		g_value_set_int (value, CHROMA_VSCROLLBAR(object)->private->allocation.width);
	}
	break;

	case C_VSCROLL_PROP_H:
	{
		g_value_set_int (value, CHROMA_VSCROLLBAR(object)->private->allocation.height);
	}
	break;

	case C_VSCROLL_PROP_GRIP_HEIGHT:
	{
		g_value_set_double (value, CHROMA_VSCROLLBAR(object)->private->grip_height);
	}
	break;

       default:
                /* We don't have any other property... */
                g_assert (FALSE);
                break;
        }
}

static GObject *
chroma_vscrollbar_constructor (GType                  type,
                       guint                  n_construct_properties,
                       GObjectConstructParam *construct_properties)
{
        GObject *obj;

        {
                /* Invoke parent constructor. */
                ChromaVScrollbarClass *klass;
                GObjectClass *parent_class;
                klass = CHROMA_VSCROLLBAR_CLASS (g_type_class_peek (CHROMA_TYPE_VSCROLLBAR));
                parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
                obj = parent_class->constructor (type,
                                                 n_construct_properties,
                                                 construct_properties);
        }

        /* do stuff. */

        return obj;
}

static void
chroma_vscrollbar_dispose (GObject *obj)
{
        ChromaVScrollbar *self = (ChromaVScrollbar *)obj;

        if (self->private->dispose_has_run) {
                /* If dispose did already run, return. */
                return;
        }
        /* Make sure dispose does not run twice. */
        self->private->dispose_has_run = TRUE;

        /*
         * In dispose, you are supposed to free all types referenced from this
         * object which might themselves hold a reference to self. Generally,
         * the most simple solution is to unref all members on which you own a
         * reference.
         */
}

static void
chroma_vscrollbar_finalize (GObject *obj)
{
        ChromaVScrollbar *self = (ChromaVScrollbar *)obj;

        /*
         * Here, complete object destruction.
         * You might not need to do much...
         */
        g_free (self->private);
}

void
chroma_vscrollbar_move_resize(ChromaVScrollbar * self, gint pos_x, gint pos_y, gint w, gint h)
{
    ChromaList *list;
    gdouble grip_height;

    g_return_if_fail(CHROMA_IS_VSCROLLBAR(self));

    self->private->allocation.x = pos_x;
    self->private->allocation.y = pos_y;
    self->private->allocation.width = w;
    self->private->allocation.height = h;

    list = g_object_get_data (G_OBJECT (self), "list");

    if ((chroma_list_get_n_item_display(list) != 0) && (chroma_list_get_n_items(list) != 0)) {
	grip_height = (gdouble)chroma_list_get_n_items(list)/(gdouble)chroma_list_get_n_item_display(list);
	grip_height = (h - w*2 - 2) / grip_height;
	g_object_set (self, "grip_height", grip_height, NULL);
    }

    chroma_vscrollbar_emit_expose_request(self);
}

GObject*
chroma_vscrollbar_new (void)
{
  ChromaVScrollbar *scrollbar;

  scrollbar = g_object_new (chroma_vscrollbar_get_type (), NULL);
  return G_OBJECT (scrollbar);
}

#if 0
static gint
_scroll_grip_size (ChromaVScrollbar *self)
{
	ChromaList *list;
	gint row_height, grip_size;

	list = CHROMA_LIST (g_object_get_data (self, "list"));

	row_height = g_object_get_int (list, "row_height");

	grip_size = ( (self->private->allocation.height / row_height) * row_height);

	return grip_size;
}
#endif

void
chroma_vscrollbar_adjust_sizes (ChromaVScrollbar *self)
{
  gint control_y = 0, scroll_height, scroll_start, scroll_grip_size;
  gdouble position_percent, percent;

  scroll_start = self->private->allocation.width + 2;
  scroll_height = (self->private->allocation.height - ((self->private->allocation.width*2)+4));

  scroll_grip_size = self->private->allocation.height - ((self->private->allocation.width*2)+4);
  if (self->private->grip_height > 0)
	scroll_grip_size  = self->private->grip_height;

  scroll_grip_size = (scroll_grip_size < SCROLLBAR_GRIP_MINIMAL_SIZE) ? SCROLLBAR_GRIP_MINIMAL_SIZE : scroll_grip_size;

  percent = scroll_height / 100.0;
  position_percent = (gdouble)self->private->scroll_position / percent;

  self->private->scroll_position = (control_y - scroll_start) - (scroll_grip_size/2);

  if (self->private->scroll_position < 0) self->private->scroll_position = 0;
  if (self->private->scroll_position > (scroll_height - scroll_grip_size)) self->private->scroll_position = (scroll_height - scroll_grip_size);
}

gboolean
chroma_vscrollbar_dispatch_motion (ChromaVScrollbar *self, GdkEventMotion *event)
{
  gboolean inside;
  gdouble control_x, control_y, scroll_area, scroll_start, scroll_grip_size, scroll_step;
  gdouble scroll_percent;

  inside =
		((event->y >= self->private->allocation.y) &&
		(event->y <= (self->private->allocation.y + self->private->allocation.height)));


  self->private->inside =
		 ((event->x >= g_object_get_int (self, "pos_x")) &&
		 (event->x <= (g_object_get_int (self, "pos_x") + g_object_get_int(self, "width")))) ? TRUE : FALSE;

  self->private->last_y = event->y - g_object_get_int (self, "pos_y");

  if (!self->private->active) {
	return FALSE;
  }

  if (self->private->active_area != AREA_GRIP) {
	return FALSE;
  }

  scroll_start = self->private->allocation.width + 2;
  scroll_area =  self->private->allocation.height - ((self->private->allocation.width*2)+4);

  scroll_grip_size = self->private->allocation.height - ((self->private->allocation.width*2)+4);
  if (self->private->grip_height > 0)
	scroll_grip_size  = self->private->grip_height;
  scroll_grip_size = (scroll_grip_size < SCROLLBAR_GRIP_MINIMAL_SIZE) ? SCROLLBAR_GRIP_MINIMAL_SIZE : scroll_grip_size;

  scroll_area = scroll_area - scroll_grip_size;
  scroll_percent = scroll_area / 100;

  control_x = event->x - self->private->allocation.x;
  control_y = event->y - self->private->allocation.y;

  scroll_step = self->private->allocation.height - ((self->private->allocation.width*2)+4);
  if (self->private->grip_height > 0)
  	scroll_step /= self->private->grip_height;

  scroll_step = (scroll_step < SCROLLBAR_GRIP_MINIMAL_SIZE) ? SCROLLBAR_GRIP_MINIMAL_SIZE : scroll_step;
  scroll_step /= scroll_area;

  if ( (control_y > scroll_start) &&
       (control_y < (scroll_start + (scroll_area+scroll_grip_size))) )
    {
        control_y = event->y - self->private->allocation.y - (scroll_grip_size/2);
  	self->private->scroll_position = ((gdouble)control_y - (gdouble)scroll_start) / (gdouble)scroll_area;

#if 0
	aux_int0 = self->private->scroll_position*100;
	aux_int1 = scroll_step*100;

	aux_int2 = (aux_int0/aux_int1)*aux_int1;

	self->private->scroll_position = aux_int2/100.;
#endif

	if (self->private->scroll_position < 0) self->private->scroll_position = 0;
	if (self->private->scroll_position > 1) self->private->scroll_position = 1;

	chroma_vscrollbar_emit_expose_request (self);
	g_signal_emit (self, chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_POSITION], 0, self->private->scroll_position);
     }

   chroma_vscrollbar_emit_expose_request (self);
   return inside;
}

void
chroma_vscrollbar_inside(ChromaVScrollbar *self, gint x,  gint y)
{
	self->private->inside =	((x >= self->private->allocation.x) &&
		 (x <= (self->private->allocation.x + self->private->allocation.width)) &&
		 (y >= self->private->allocation.y) &&
		 (y <= (self->private->allocation.y + self->private->allocation.height + 1)));
}

gboolean
chroma_vscrollbar_get_inside(ChromaVScrollbar *self)
{
	return self->private->inside;
}

gboolean
chroma_vscrollbar_dispatch_button_release (ChromaVScrollbar *self, GdkEventButton *event)
{
  if (!self->private->active)
	return FALSE;

  chroma_vscrollbar_inside (self, event->x, event->y);

  self->private->last_y = event->y - g_object_get_int (self, "pos_y");

  if (!self->private->inside)
  {
	self->private->active = FALSE;
	chroma_vscrollbar_emit_expose_request (self);
	return FALSE;
  }

#if 0
  scroll_start = self->private->allocation.width + 2;
  scroll_area =  self->private->allocation.height - ((self->private->allocation.width*2)+4);

  scroll_grip_size = self->private->allocation.height - ((self->private->allocation.width*2)+4);
  if (self->private->grip_height > 0)
	scroll_grip_size  = self->private->grip_height;
  scroll_grip_size = (scroll_grip_size < SCROLLBAR_GRIP_MINIMAL_SIZE) ? SCROLLBAR_GRIP_MINIMAL_SIZE : scroll_grip_size;

  scroll_area = scroll_area - scroll_grip_size;
  scroll_percent = scroll_area / 100;

  control_x = event->x - self->private->allocation.x;
  control_y = event->y - self->private->allocation.y;

  scroll_step = self->private->allocation.height - ((self->private->allocation.width*2)+4);
  if (self->private->grip_height > 0)
  	scroll_step /= self->private->grip_height;

  scroll_step = (scroll_step < SCROLLBAR_GRIP_MINIMAL_SIZE) ? SCROLLBAR_GRIP_MINIMAL_SIZE : scroll_step;
  scroll_step /= scroll_area;

  if (!self->private->inside) return FALSE;

  // self->private->active = TRUE;
#endif

#if 0
  if ( (control_y > scroll_start) &&
       (control_y < (scroll_start + (scroll_area+scroll_grip_size))) )
    {
	gint clamp;

        control_y = event->y - self->private->allocation.y - (scroll_grip_size/2);
  	self->private->scroll_position = ((gdouble)control_y - (gdouble)scroll_start) / (gdouble)scroll_area;

	if (self->private->scroll_position < 0) self->private->scroll_position = 0;
	if (self->private->scroll_position > 1) self->private->scroll_position = 1;

	self->private->active_area = AREA_GRIP;
     }
  /* "Up" scroll arrow */
  else if ( (control_y <= self->private->allocation.width))
     {
	self->private->scroll_position -= scroll_step;
	if (self->private->scroll_position < 0) self->private->scroll_position = 0;
	self->private->active_area = AREA_ARROW_UP;
     }
  /* "Down" scroll arrow */
  else if ((control_y >= (self->private->allocation.height - self->private->allocation.width)) && (control_y <= self->private->allocation.height ))
     {
	self->private->scroll_position += scroll_step;
	if (self->private->scroll_position > 1) self->private->scroll_position = 1;

	self->private->active_area = AREA_ARROW_DOWN;
     }

  g_signal_emit (self, chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_POSITION], 0, self->private->scroll_position);
#endif

    self->private->active = FALSE;
    self->private->active_area = AREA_NONE;

    chroma_vscrollbar_emit_expose_request (self);

    return self->private->inside;
}

gboolean
chroma_vscrollbar_dispatch_button_press (ChromaVScrollbar *self, GdkEventButton *event)
{
  gdouble control_x, control_y, scroll_area, scroll_start, scroll_grip_size, scroll_step;
  gdouble scroll_percent;

  chroma_vscrollbar_inside(self,event->x,event->y);

  if (! self->private->inside) return FALSE;

  self->private->last_y = event->y - g_object_get_int (self, "pos_y");

  self->private->active = TRUE;

  scroll_start = self->private->allocation.width + 2;
  scroll_area =  self->private->allocation.height - ((self->private->allocation.width*2)+4);

  scroll_grip_size = self->private->allocation.height - ((self->private->allocation.width*2)+4);
  if (self->private->grip_height > 0)
	scroll_grip_size  = self->private->grip_height;
  scroll_grip_size = (scroll_grip_size < SCROLLBAR_GRIP_MINIMAL_SIZE) ? SCROLLBAR_GRIP_MINIMAL_SIZE : scroll_grip_size;

  scroll_area = scroll_area - scroll_grip_size;
  scroll_percent = scroll_area / 100;

  control_x = event->x - self->private->allocation.x;
  control_y = event->y - self->private->allocation.y;

  scroll_step = self->private->allocation.height - ((self->private->allocation.width*2)+4);
  if (self->private->grip_height > 0)
  	scroll_step /= self->private->grip_height;

  scroll_step = (scroll_step < SCROLLBAR_GRIP_MINIMAL_SIZE) ? SCROLLBAR_GRIP_MINIMAL_SIZE : scroll_step;
  scroll_step /= scroll_area;

  if (!self->private->inside) return FALSE;

  self->private->active = TRUE;

  if ( (control_y > scroll_start) &&
       (control_y < (scroll_start + (scroll_area+scroll_grip_size))) )
    {

        control_y = event->y - self->private->allocation.y - (scroll_grip_size/2);
  	self->private->scroll_position = ((gdouble)control_y - (gdouble)scroll_start) / (gdouble)scroll_area;

#if 0
	aux_int0 = self->private->scroll_position*100;
	aux_int1 = scroll_step*100;

	aux_int2 = (aux_int0/aux_int1)*aux_int1;

	self->private->scroll_position = aux_int2/100.;
#endif

	if (self->private->scroll_position < 0) self->private->scroll_position = 0;
	if (self->private->scroll_position > 1) self->private->scroll_position = 1;

	self->private->active_area = AREA_GRIP;

  	g_signal_emit (self, chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_POSITION], 0, self->private->scroll_position);
     }
  /* "Up" scroll arrow */
  else if ( (control_y <= self->private->allocation.width))
     {
	self->private->scroll_position -= scroll_step;
	if (self->private->scroll_position < 0) self->private->scroll_position = 0;
  	g_signal_emit (self, chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_POSITION], 0, self->private->scroll_position);

	self->private->active_area = AREA_ARROW_UP;
     }
  /* "Down" scroll arrow */
  else if ((control_y >= (self->private->allocation.height - self->private->allocation.width)) && (control_y <= self->private->allocation.height ))
     {
	self->private->scroll_position += scroll_step;
	if (self->private->scroll_position > 1) self->private->scroll_position = 1;
  	g_signal_emit (self, chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_POSITION], 0, self->private->scroll_position);

	self->private->active_area = AREA_ARROW_DOWN;
     }


    chroma_vscrollbar_emit_expose_request (self);

    return self->private->inside;
}

#define CALCULATE_SCROLL_VALUES() \
  	scroll_grip_size = self->private->allocation.height - ((self->private->allocation.width*2)+4); \
	\
	if (self->private->grip_height > 0) \
		scroll_grip_size  = self->private->grip_height; \
	scroll_grip_size = (scroll_grip_size < SCROLLBAR_GRIP_MINIMAL_SIZE) ? SCROLLBAR_GRIP_MINIMAL_SIZE : scroll_grip_size; \
	\
        scroll_area = (self->private->allocation.height - ((self->private->allocation.width*2)+4));  \
	scroll_area -= scroll_grip_size; \
	\
	scroll_step = self->private->allocation.height - ((self->private->allocation.width*2)+4); \
	  if (self->private->grip_height > 0) \
  		scroll_step = self->private->grip_height; \
	\
	scroll_step /= scroll_area;

#if 0
 	mod = (((scroll_step / 2) * 2) == scroll_step) ? 1 : 0;
	scroll_step -= mod;
#endif

/* Added by Kenneth "Langly" Østby */
gboolean
chroma_vscrollbar_dispatch_scroll (ChromaVScrollbar *self, GdkEventScroll *event)
{
	gdouble scroll_step,scroll_area,scroll_grip_size;

	CALCULATE_SCROLL_VALUES();

	chroma_vscrollbar_inside(self, event->x, event->y);

	if (!self->private->inside) return FALSE;

	/*
	 * Have to handle them both. Some mouse types can scroll
	 * sideways and that is supported in GDK
	 */
	if(event->direction == GDK_SCROLL_UP){
		self->private->scroll_position -= (scroll_step/6.); /* FIXME: Do not hardcode this */
		if (self->private->scroll_position < 0) self->private->scroll_position = 0;
	}else if(event->direction == GDK_SCROLL_DOWN) {
		self->private->scroll_position += (scroll_step/6.);
		if (self->private->scroll_position > 1) self->private->scroll_position = 1;
	}

	chroma_vscrollbar_emit_expose_request (self);
	g_signal_emit (self, chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_POSITION], 0, self->private->scroll_position);

	return TRUE;
}

gboolean
chroma_vscrollbar_dispatch_scroll_direct (ChromaVScrollbar *self, GdkEventScroll *event)
{
	gdouble scroll_step,scroll_area,scroll_grip_size;

	CALCULATE_SCROLL_VALUES();

	/*
	 * Have to handle them both. Some mouse types can scroll
	 * sideways and that is supported in GDK
	 */
	if(event->direction == GDK_SCROLL_UP){
		self->private->scroll_position -= (scroll_step/6.); /* FIXME: Do not hardcode this */
		if (self->private->scroll_position < 0) self->private->scroll_position = 0;
	}else if(event->direction == GDK_SCROLL_DOWN) {
		self->private->scroll_position += (scroll_step/6.);
		if (self->private->scroll_position > 1) self->private->scroll_position = 1;
	}

	chroma_vscrollbar_emit_expose_request (self);
	g_signal_emit (self, chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_POSITION], 0, self->private->scroll_position);

	return TRUE;
}

void
chroma_vscrollbar_draw (ChromaVScrollbar *self, GdkDrawable *drawable, cairo_t *cr)
{
	BmpPoint points[3];
	GdkColor *fg_color, *bg_color, *fg_color_shaded;
	gint scroll_grip_size, scroll_area, pressed_offset_arrow_up = 0, pressed_offset_arrow_down = 0;
	gdouble x, y, w, h, xr, yr, wr, hr;

	g_return_if_fail (CHROMA_IS_VSCROLLBAR(self));

	fg_color = bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT);
	bg_color = bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL);
	fg_color_shaded = bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_SELECTED);

	x = g_object_get_int (self, "pos_x")+0.5;
	y = g_object_get_int (self, "pos_y");
	w = g_object_get_int (self, "width") - 2;
	h = g_object_get_int (self, "height");

	/* Top (up) arrow */
	if ( (self->private->active) && (self->private->active_area == AREA_ARROW_UP) && (self->private->inside) && (self->private->last_y <= self->private->allocation.width) )
		draw_rounded_rectangle (cr, x, y, w, w, fg_color, 0.2, 1.0, TRUE,  1.5);

	draw_rounded_rectangle (cr, x, y, w, w, fg_color, 1.0, 0.7, FALSE, 1.5);


	/* Draw proportional scroll grip */
	scroll_grip_size = self->private->allocation.height - ((self->private->allocation.width*2)+4);
	if (self->private->grip_height > 0) scroll_grip_size  = self->private->grip_height;
	scroll_grip_size = (scroll_grip_size < SCROLLBAR_GRIP_MINIMAL_SIZE) ? SCROLLBAR_GRIP_MINIMAL_SIZE : scroll_grip_size;
        scroll_area = (self->private->allocation.height - ((self->private->allocation.width*2)+4));
	xr = x+0.15;
	yr = y + w + 3 + ((gdouble)self->private->scroll_position * ((gdouble)scroll_area - (gdouble)scroll_grip_size));
	wr = w-0.3;
	hr = scroll_grip_size-1;
	draw_rounded_rectangle (cr, xr, yr, wr, hr, fg_color, 0.2, 1.0, TRUE, 1.5);
	draw_rounded_rectangle (cr, xr, yr, wr, hr, fg_color, 1.0, 0.7, FALSE, 1.5);

	/* Bottom (down) arrow */
	if ( (self->private->active) && (self->private->active_area == AREA_ARROW_DOWN) && (self->private->inside) && (self->private->last_y >= (self->private->allocation.height - self->private->allocation.width)))
		draw_rounded_rectangle (cr, x, y + (self->private->allocation.height - self->private->allocation.width), w, w, fg_color, 0.2, 1.0, TRUE, 1.5);

	draw_rounded_rectangle (cr, x, y + (self->private->allocation.height - self->private->allocation.width), w, w, fg_color, 1.0, 0.7, FALSE, 1.5);

	points[0].x = x + (w / 2.) 		+ pressed_offset_arrow_up;
	points[0].y = y + 2. 			+ pressed_offset_arrow_up + 1.;
	points[1].x = x + w - 3. 		+ pressed_offset_arrow_up;
	points[1].y = y + w - 5. 		+ pressed_offset_arrow_up + 1.;
	points[2].x = x + 3. 			+ pressed_offset_arrow_up;
	points[2].y = y + w - 5. 		+ pressed_offset_arrow_up + 1.;
	draw_polygon (cr, fg_color, 0.8, points, 3, TRUE);

	points[0].x = x + (w / 2.) 		+ pressed_offset_arrow_down;
	points[0].y = (y + h) - 2. 		+ pressed_offset_arrow_down - 2.;
	points[1].x = x + w - 3. 		+ pressed_offset_arrow_down;
	points[1].y = y + (h - w) + 4. 		+ pressed_offset_arrow_down - 1.;
	points[2].x = x + 3. 			+ pressed_offset_arrow_down;
	points[2].y = y + (h - w) + 4. 		+ pressed_offset_arrow_down - 1.;
	draw_polygon (cr, fg_color, 0.8, points, 3, TRUE);
}

void
chroma_vscrollbar_emit_expose_request (ChromaVScrollbar *self)
{
    g_return_if_fail (CHROMA_IS_VSCROLLBAR (self));

    g_signal_emit (self, chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_EXPOSE_REQUEST],
		0,
		g_object_get_int (self, "pos_x"),
		g_object_get_int (self, "pos_y"),
		g_object_get_int (self, "width"),
		g_object_get_int (self, "height"));
}

static void
chroma_vscrollbar_class_init (gpointer g_class,
                      gpointer g_class_data)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
        GParamSpec *pspec;

        gobject_class->set_property = chroma_vscrollbar_set_property;
        gobject_class->get_property = chroma_vscrollbar_get_property;
        gobject_class->dispose = chroma_vscrollbar_dispose;
        gobject_class->finalize = chroma_vscrollbar_finalize;
        gobject_class->constructor = chroma_vscrollbar_constructor;

        pspec = g_param_spec_int ("pos_x",
                                  "pos_x",
                                  "pos_x", 0, G_MAXINT, 0,
                                  G_PARAM_READABLE);
        g_object_class_install_property (gobject_class,
                                         C_VSCROLL_PROP_X,
                                         pspec);

        pspec = g_param_spec_int ("pos_y",
                                  "pos_y",
                                  "pos_y", 0, G_MAXINT, 0,
                                  G_PARAM_READABLE);
        g_object_class_install_property (gobject_class,
                                         C_VSCROLL_PROP_Y,
                                         pspec);

        pspec = g_param_spec_int ("width",
                                  "width",
                                  "width", 0, G_MAXINT, 0,
                                  G_PARAM_READABLE);
        g_object_class_install_property (gobject_class,
                                         C_VSCROLL_PROP_W,
                                         pspec);

        pspec = g_param_spec_int ("height",
                                  "height",
                                  "height", 0, G_MAXINT, 0,
                                  G_PARAM_READABLE);
        g_object_class_install_property (gobject_class,
                                         C_VSCROLL_PROP_H,
                                         pspec);

        pspec = g_param_spec_double ("grip_height",
                                  "grip_height",
                                  "grip_height", 0, G_MAXINT, 0,
                                  G_PARAM_READWRITE);
        g_object_class_install_property (gobject_class,
                                         C_VSCROLL_PROP_GRIP_HEIGHT,
                                         pspec);

	pspec = g_param_spec_double ("position",
					"position",
					"position",
					0,
					1,
					0,
					G_PARAM_READWRITE);
        g_object_class_install_property (gobject_class,
                                         C_VSCROLL_PROP_POSITION,
                                         pspec);

        chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_POSITION] =
        g_signal_new ("position",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (ChromaVScrollbarClass, position),
                  NULL, NULL,
                  g_cclosure_marshal_VOID__DOUBLE,
                  G_TYPE_NONE, 1, G_TYPE_DOUBLE);

        chroma_vscrollbar_signals[CHROMA_VSCROLLBAR_SIGNAL_EXPOSE_REQUEST] =
        g_signal_new ("expose_request",
                  G_OBJECT_CLASS_TYPE (gobject_class),
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (ChromaVScrollbarClass, expose_request),
                  NULL, NULL,
                  bmpx_VOID__INT_INT_INT_INT,
                  G_TYPE_NONE, 4, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);
}

GType
chroma_vscrollbar_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (ChromaVScrollbarClass),
                        NULL,   /* base_init */
                        NULL,   /* base_finalize */
                        chroma_vscrollbar_class_init,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        sizeof (ChromaVScrollbar),
                        0,      /* n_preallocs */
                        chroma_vscrollbar_instance_init    /* instance_init */
                };
                type = g_type_register_static (G_TYPE_OBJECT,
                                               "ChromaVScrollbarType",
                                               &info, 0);
        }
        return type;
}
