/* NanoGC */

#include <glib.h>
#include <gc.h>

/* FIXME: Move the lock inside the NGC struct so we can actually have multiple instances (if we need them...) */
G_LOCK_DEFINE (GC_LOCK);

typedef struct {

    gpointer pointer;
    GDestroyNotify destroy;

} GcItem;


NGC*
gc_new (void)
{
    NGC *gc;

    gc = g_new0 (NGC, 1);

    gc->stack = NULL;

    return gc;
}

void
gc_destroy (NGC *gc)
{
    g_free (gc); 
}

void
gc_destroy_free (NGC *gc)
{
}

void
gc_mark (NGC *gc)
{
    GList *collector_stack = NULL;

    G_LOCK (GC_LOCK);

    gc->stack = g_list_prepend (gc->stack, collector_stack);
    gc->ctr = 0;


    G_UNLOCK (GC_LOCK);
}

static void
free_func (gpointer pointer, gpointer gcpointer)
{
    GcItem *item = (GcItem*)pointer;
    NGC *gc = (NGC*)gcpointer;

    if (! pointer)
	return;

    if (item->destroy)
	{
	    item->destroy (item->pointer);
	}
    else
	{
	    g_free (item->pointer);
	}

    g_free (item);
    gc->ctr--;

}

void
gc_free (NGC *gc)
{
    GList *collector_stack;

    g_return_if_fail (gc->stack != NULL);

    G_LOCK (GC_LOCK);

    collector_stack = gc->stack->data; 

    g_list_foreach (collector_stack, free_func, gc);
    gc->stack = g_list_remove (gc->stack, collector_stack);
    g_list_free (collector_stack);

    G_UNLOCK (GC_LOCK);
}

gpointer
gc_collect_d (NGC *gc, gpointer collect_pointer, GDestroyNotify destroy)
{
    GcItem *item;
    GList *collector_stack;

    G_LOCK (GC_LOCK);

    collector_stack = gc->stack->data;  

    item = g_new0 (GcItem, 1);
    item->pointer = collect_pointer;
    item->destroy = destroy;

    collector_stack = g_list_append (collector_stack, item);

    G_UNLOCK (GC_LOCK);

    return collect_pointer;
}

gpointer
gc_collect (NGC *gc, gpointer collect_pointer)
{
    GcItem *item;
    GList *collector_stack;

    G_LOCK (GC_LOCK);

    collector_stack = gc->stack->data;  

    item = g_new0 (GcItem, 1);
    item->pointer = collect_pointer;
    item->destroy = g_free;

    collector_stack = g_list_append (collector_stack, item);
    gc->ctr++;

    G_UNLOCK (GC_LOCK);

    return collect_pointer;
}
