/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */

#ifndef BMP_SYSTEM_CONTROL_H
#define BMP_SYSTEM_CONTROL_H

/*
 * Potentially, include other headers on which this header depends.
 */
#include <glib-object.h>
#include "bmpx/list.h"

#define BMP_TYPE_SYSTEM_CONTROL		  (bmp_system_control_get_type ())
#define BMP_SYSTEM_CONTROL(obj)		  (G_TYPE_CHECK_INSTANCE_CAST ((obj), BMP_TYPE_SYSTEM_CONTROL, BmpSystemControl))
#define BMP_SYSTEM_CONTROL_CLASS(klass)	  (G_TYPE_CHECK_CLASS_CAST ((klass), BMP_TYPE_SYSTEM_CONTROL, BmpSystemControlClass))
#define BMP_IS_SYSTEM_CONTROL(obj)	  (G_TYPE_CHECK_INSTANCE_TYPE ((obj), BMP_TYPE_SYSTEM_CONTROL))
#define BMP_IS_SYSTEM_CONTROL_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), BMP_TYPE_SYSTEM_CONTROL))
#define BMP_SYSTEM_CONTROL_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), BMP_TYPE_SYSTEM_CONTROL, BmpSystemControlClass))

typedef struct _BmpSystemControl BmpSystemControl;
typedef struct _BmpSystemControlClass BmpSystemControlClass;
typedef struct _BmpSystemControlPrivate BmpSystemControlPrivate;

struct _BmpSystemControl {
    GObject parent;

    BmpSystemControlPrivate *private;
};

struct _BmpSystemControlClass {
    GObjectClass parent;

    /* Signals */
    void (*track_change)	     (BmpSystemControl * self);

    void (*set_current_tracklist)    (BmpSystemControl * self, gint list);
    void (*set_active_tracklist)     (BmpSystemControl * self, gint list);
    void (*set_stop_after_current)   (BmpSystemControl * self, gboolean stop);
    void (*set_stream_pos)	     (BmpSystemControl * self, gint stream_pos);
    void (*set_playstatus)	     (BmpSystemControl * self, gint playstatus);
    void (*set_title)		     (BmpSystemControl * self, gchar *title);
    void (*set_volume)		     (BmpSystemControl * self, gint volume);
    void (*set_eq)		     (BmpSystemControl * self, gint eq_id, gint value);
    void (*set_shuffle)		     (BmpSystemControl * self, gboolean shuffle);
    void (*set_repeat)		     (BmpSystemControl * self, gboolean repeat);

    void (*list_deleted)	     (BmpSystemControl *self, gint list);
    void (*list_added)		     (BmpSystemControl *self, gint list);
    void (*list_renamed)	     (BmpSystemControl *self, gint list);

    void (*tracklist_list_sorted)    (BmpSystemControl *self, gint list, gpointer array, gint n_items);
    void (*tracklist_items_removed)  (BmpSystemControl *self, gint list, gpointer array, gint n_items);
    void (*tracklist_items_added)    (BmpSystemControl *self, gint list, gpointer array, gint n_items);
    void (*tracklist_rows_swapped)   (BmpSystemControl *self, gint list, gint row_a, gint row_b); 

    void (*startup_complete)	     (BmpSystemControl *self);
    void (*shutdown_complete)	     (BmpSystemControl *self);
};

typedef enum {
    BMP_SYSTEM_MESSAGE_CURRENT,
    BMP_SYSTEM_MESSAGE_NEXT,
    BMP_SYSTEM_MESSAGE_PREV,
    BMP_SYSTEM_MESSAGE_STOP,
    BMP_SYSTEM_MESSAGE_PAUSE,
    BMP_SYSTEM_MESSAGE_SEEK,
    BMP_SYSTEM_MESSAGE_SEEK_PERCENT,
    BMP_SYSTEM_MESSAGE_PLAY_LIST,
    BMP_SYSTEM_MESSAGE_PLAY_TRACK,
    BMP_SYSTEM_MESSAGE_PLAY_LIST_TRACK,
    BMP_SYSTEM_MESSAGE_SET_CURRENT_TRACKLIST,
    BMP_SYSTEM_MESSAGE_STOP_AFTER_CURRENT,
    BMP_SYSTEM_MESSAGE_VOLUME,
    BMP_SYSTEM_MESSAGE_EQ,
    BMP_SYSTEM_MESSAGE_BITRATE,
    BMP_SYSTEM_MESSAGE_SAMPLERATE,
    BMP_SYSTEM_MESSAGE_REPEAT,
    BMP_SYSTEM_MESSAGE_SHUFFLE,
    BMP_SYSTEM_MESSAGE_UI_START,
    BMP_SYSTEM_MESSAGE_UI_STOP,
    BMP_SYSTEM_MESSAGE_SHOW_PL_WINDOW,
    BMP_SYSTEM_MESSAGE_SAVE_STATE,
    BMP_SYSTEM_MESSAGE_QUIT
} BmpSystemMessage;

typedef struct {
    BmpSystemMessage message_type;
    gpointer	     user_data;
    GDestroyNotify   user_data_destroy;
} bmp_message_t;

GType
bmp_system_control_get_type (void);

BmpSystemControl*
bmp_system_control_new (gboolean no_remote);

// FIXME: Fix up DBus exported methods to be really compliant, as in, return a gboolean and take in a GError

void bmp_system_message_send (BmpSystemControl	*self,
			      BmpSystemMessage	message_type,
			      gpointer		user_data,
			      GDestroyNotify	user_data_destroy);

void bmp_system_control_show_pl_window (BmpSystemControl *self,
					gboolean show);

void bmp_system_control_current_tracklist_set (BmpSystemControl *self,
					       gint _idx);

void bmp_system_control_move_start (BmpSystemControl *self);

void bmp_system_control_move_end (BmpSystemControl *self);

void bmp_system_control_ui_start (BmpSystemControl *self);

void bmp_system_control_ui_stop (BmpSystemControl *self);

void bmp_system_control_quit (BmpSystemControl *self);

void bmp_system_control_save_state (BmpSystemControl *self);

void bmp_system_control_bitrate (BmpSystemControl *self,
				 gint bitrate);

void bmp_system_control_samplerate (BmpSystemControl *self,
				    gint samplerate);

void bmp_system_control_seek_percent (BmpSystemControl * self, gint position);

void bmp_system_control_play_track (BmpSystemControl * self, gint _idx);

void bmp_system_control_play_list (BmpSystemControl * self, gint _idx_list);

void bmp_system_control_play_list_track (BmpSystemControl * self, gint _idx_list,
				    gint _idx);

void bmp_system_control_play_pause (BmpSystemControl * self);

void bmp_system_control_rows_swapped (BmpSystemControl *self,
				      gint list, 
				      gint row_a, 
				      gint row_b); 

void bmp_system_control_play_current (BmpSystemControl *self);

void bmp_system_control_send_startup_complete (BmpSystemControl *self);

void bmp_system_control_send_shutdown_complete (BmpSystemControl *self);

void bmp_system_control_app_busy (BmpSystemControl *self);

void bmp_system_control_app_idle (BmpSystemControl *self);

gboolean
bmp_system_control_export_tracklist (BmpSystemControl *self,
				     const gchar *container_plugin,
				     gint         _idx,
				     gboolean	  all,
				     const gchar *filename,
				     GError **error);

gboolean
bmp_system_control_volume_set (BmpSystemControl * self, gint volume, GError ** error);

gboolean
bmp_system_control_eq_set (BmpSystemControl *self,
			    BmpEqEnum eq,
			    gint value,
			    GError **error);

gboolean
bmp_system_control_play_stop (BmpSystemControl * self, GError * error);

gboolean
bmp_system_control_play_prev (BmpSystemControl * self, GError * error);

gboolean
bmp_system_control_play_next (BmpSystemControl * self, GError * error);

gboolean
bmp_system_control_add_uri_list (BmpSystemControl * self, const gchar **uri_list,
				 gint tracklist, gint position,
				 gint * position_out, GError ** error);

gboolean
bmp_system_control_clear_playback_history (BmpSystemControl *self, GError **error);

gboolean
bmp_system_control_tracklist_sort (BmpSystemControl * self,
				   gint _idx,
				   gint tag_id,
				   GError ** error);

gboolean
bmp_system_control_tracklist_remove (BmpSystemControl * self,
				 gint list_index,
				 GError ** error);

gboolean
bmp_system_control_stop_after_current_set (BmpSystemControl * self, gboolean stop, GError **error);

gboolean
bmp_system_control_repeat_set (BmpSystemControl *self, gboolean repeat, GError **error);

gboolean
bmp_system_control_repeat_get (BmpSystemControl *self, gboolean *repeat, GError **error);

gboolean
bmp_system_control_shuffle_set (BmpSystemControl *self, gboolean shuffle, GError **error);

gboolean
bmp_system_control_shuffle_get (BmpSystemControl *self, gboolean *shuffle, GError **error);


#endif /* BMP_SYSTEM_CONTROL_H */
