#include <stdlib.h>
#include <string.h>

#include <gdk/gdkkeysyms.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include <libskinned/bmp_window.h>
#include <libskinned/bmp_slider.h>

#include "main.h"
#include "system_control.h"
#include "resource_manager.h"

#include <bmpx/util.h>
#include <bmpx/uri.h>
#include <bmpx/playlist.h>

#include "xcs/xcs.h"
#include "goa/libgoa.h"

#include "glade.h"

#include "ui.h"
#include "ui_playlist.h"
#include "ui_prefswin.h"

#include "ui_dialog_about.h"
#include "ui_dialog_library.h"

#include "chroma_list.h"
#include "plugin_priv.h"
#include "embedded-images/media-pixbufs.h"


static GtkWidget *fi_dialog = NULL;

void
bmp_ui_callback_about (GtkAction *widget, gpointer data)
{
	bmp_about_dialog_show ();
}

void
bmp_ui_callback_preferences (GtkAction *widget, gpointer data)
{
    BmpWindowPreferences *window_preferences;

    window_preferences = RM_GET_ITEM("windows", "window-preferences", NOLOCK);

    bmp_preferences_window_show (window_preferences);
}

void
bmp_ui_callback_quit (GtkAction *widget, gpointer data)
{
	bmp_system_control_quit (bmp_system_control);
}

void
bmp_ui_callback_clear_playback_history (GtkAction *action, gpointer data)
{
    bmp_system_control_clear_playback_history (bmp_system_control, NULL);
    gtk_action_set_sensitive (gtk_action_group_get_action(bmp_actions, BMP_ACTION_CLEAR_PLAYBACK_HISTORY), FALSE);
}

void
bmp_ui_callback_iconify (GtkAction *widget, gpointer data)
{
	BmpWindowMain     *window_main;
	BmpWindowPlaylist *window_playlist;

	window_main     = RM_GET_ITEM("windows", "window-main", NOLOCK);
	window_playlist = RM_GET_ITEM("windows", "window-playlist", NOLOCK);

	gtk_window_iconify (GTK_WINDOW(window_main->window));
	gtk_window_iconify (GTK_WINDOW(window_playlist->window));
}

void
bmp_ui_callback_prev(GtkAction *widget, gpointer data)
{
	bmp_system_control_play_prev (bmp_system_control, NULL);
}

void
bmp_ui_callback_play(GtkAction *widget, gpointer data)
{
	bmp_system_control_play_current (bmp_system_control);
}

void
bmp_ui_callback_pause(GtkAction *widget, gpointer data)
{
	bmp_system_control_play_pause (bmp_system_control);
}

void
bmp_ui_callback_stop(GtkAction *widget, gpointer data)
{
	bmp_system_control_play_stop (bmp_system_control, NULL);
}

void
bmp_ui_callback_next(GtkAction *widget, gpointer data)
{
	bmp_system_control_play_next (bmp_system_control, NULL);
}

void
bmp_ui_callback_toggle_shuffle(GtkToggleAction *action, gpointer data)
{
  if (gtk_toggle_action_get_active(action))
    {
      bmp_system_control_shuffle_set (bmp_system_control, TRUE, NULL);
    }
  else
    {
      bmp_system_control_shuffle_set (bmp_system_control, FALSE, NULL);
    }
}

void
bmp_ui_callback_toggle_repeat(GtkToggleAction *action, gpointer data)
{

 if (gtk_toggle_action_get_active(action))
   {
      bmp_system_control_repeat_set (bmp_system_control, TRUE, NULL);
   }
  else
   {
      bmp_system_control_repeat_set (bmp_system_control, FALSE, NULL);
   }

}

void
bmp_ui_callback_toggle_pl(GtkToggleAction *action, gpointer data) 
{

    BmpWindowPlaylist *window_playlist;

    window_playlist = RM_GET_ITEM("windows", "window-playlist", NOLOCK);

    if (!gtk_toggle_action_get_active(action))
      {
	gtk_widget_hide (window_playlist->window);
      }
    else if ((window_playlist) && GTK_WIDGET_REALIZED(window_playlist->window))
      {
	gtk_widget_show_all (window_playlist->window);
	gtk_window_present (GTK_WINDOW(window_playlist->window));
	bmp_window_playlist_resize_widgets(window_playlist);
      }
}

void
bmp_ui_callback_seek_moved(GtkWidget *widget, gpointer data)
{
    BmpWindowMain *window_main;

    gint position = bmp_slider_get_position(BMP_SLIDER(widget));

    window_main = RM_GET_ITEM("windows", "window-main", NOLOCK);

    bmp_system_control_seek_percent (bmp_system_control, position);
    bmp_window_main_set_seek_request (window_main, position);
}

void
bmp_ui_callback_volume_moved(GtkWidget *widget, gpointer data)
{
    gint volume;

    volume = bmp_slider_get_position(BMP_SLIDER(widget));
    bmp_system_control_volume_set (bmp_system_control, volume, NULL);
}

void
bmp_ui_callback_balance_moved(GtkWidget *widget, gpointer data)
{
#if 0
    gint position = bmp_slider_get_position(BMP_SLIDER(widget));
#endif
}

static gchar*
new_tracklist (BmpPlaylist *self)
{
    BmpListType *list_t;
    gchar *aux;
    gboolean match = TRUE;
    gint ctr = 0;
    gint list;

    while (match)
      {
	ctr++;
	aux = 
	  g_strdup_printf("%s%d", Q_("New Playlist #"), ctr);
	list_t = 
	  bmp_playlist_tracklist_get (self, aux, NULL);
	g_free (aux);

	if (!list_t)
	    match = FALSE;
      }

    aux = 
      g_strdup_printf("%s%d", Q_("New Playlist #"), ctr);
    list = 
      bmp_playlist_tracklist_new (self, aux);

    return (list != -1) ? aux : NULL;
}

void
bmp_ui_callback_tracklist_new (GtkAction *action, gpointer data)
{
    BmpWindowPlaylist *window_playlist;
    gchar *listname;

    window_playlist =
	RM_GET_ITEM("windows", "window-playlist", NOLOCK);
    listname =
	new_tracklist (bmp_playlist);

    if (listname) {
	g_free (listname);
	chroma_list_emit_changed (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget));
	bmp_window_playlist_resize_widgets (window_playlist);
    }
}

/* FIXME: Make system control methods */
void
bmp_ui_callback_tracklist_remove (GtkAction *action, gpointer data)
{
    bmp_system_control_tracklist_remove (bmp_system_control, g_object_get_int (bmp_system_control, "current_tracklist"), NULL);
}

void
bmp_ui_callback_tracklists_remove_all (GtkAction *action, gpointer data)
{
    BmpWindowPlaylist *window_playlist;

    window_playlist =
	RM_GET_ITEM("windows", "window-playlist", NOLOCK);

    bmp_playlist_tracklists_remove_all (bmp_playlist);
}
/* </FIXME> */

static void
rename_tracklist (gchar *suggestion)
{
    /* FIXME: This isn't the best place for it */
    /* FIXME: The whole code is a little jack wacky */
    BmpWindowPlaylist *window_playlist;
    GladeXML *xml;
    GtkWidget *dialog;
    GtkWidget *entry;
    gint response, _idx;
    gchar *name_new = NULL, *title;

    window_playlist = RM_GET_ITEM("windows", "window-playlist", NOLOCK);

    xml = glade_xml_new_or_die(Q_("Dialogs"),  DATA_DIR "/glade/dialog_rename_tracklist.glade",
			       NULL,
                               NULL);
    dialog =
	glade_xml_get_widget (xml, "dialog_rename_tracklist");
    entry =
	glade_xml_get_widget (xml, "dialog_rename_tracklist_entry");

    if (suggestion != NULL)
	    gtk_entry_set_text (GTK_ENTRY(entry), suggestion);
    else
	    gtk_entry_set_text (GTK_ENTRY(entry), bmp_playlist_get_nth_tracklist_name (bmp_playlist, g_object_get_int (bmp_system_control, "current_tracklist")));

    gtk_editable_select_region (GTK_EDITABLE(entry), 0, -1);

    _idx =
	g_object_get_int (bmp_system_control, "current_tracklist");

    gtk_window_set_transient_for (GTK_WINDOW(dialog), GTK_WINDOW(window_playlist->window));

    repeat_rename:

    response = gtk_dialog_run (GTK_DIALOG(dialog));

    switch (response)
    {
	case GTK_RESPONSE_OK:
	  {
		name_new = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
		if (! name_new)
		{
			gtk_widget_destroy (dialog);
			return;
		}

		if (strlen(name_new) == 0)
		{
			gtk_widget_destroy (dialog);
			return;
		}
		break;
	  }
	case GTK_RESPONSE_CANCEL:
          {
    		gtk_widget_destroy (dialog);
	    	return;
          }
        default: break;
    }
    gtk_widget_hide (dialog);

    if (bmp_playlist_tracklist_rename (bmp_playlist, _idx, name_new) == -1) goto repeat_rename;
    gtk_widget_destroy (dialog);

    chroma_list_select_name (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_PLAYLISTS]->widget), name_new);

    title = bmp_playlist_get_current_title (bmp_playlist);
    bmp_ui_set_trayicon_text (bmp_ui, title);
    g_free (title);
}

/* FIXME: Make system control method */
void
bmp_ui_callback_tracklist_rename (GtkAction *action, gpointer data)
{
    rename_tracklist(NULL);
}

void
bmp_ui_callback_set_playlist_filename (GtkAction *action, gpointer data)
{
}

static gboolean
audio_files_filter (const GtkFileFilterInfo *filter_info,
		    gpointer                 data)
{
	if (is_audio_file(filter_info->uri)) return TRUE;

	return FALSE;
}

static void
_filebrowser_play (GtkWidget *widget, GtkFileChooser *chooser)
{
    BmpWindowPlaylist *window_playlist;
    GSList *filenames;
    gchar **uri_list;
    gint new_ctr = 0;

    window_playlist =
	RM_GET_ITEM("windows", "window-playlist", NOLOCK);
    filenames =
	gtk_file_chooser_get_uris (GTK_FILE_CHOOSER(chooser));

    bmp_window_set_busy (GTK_WIDGET(chooser));

    xcs_key_set_string (xcs, 
				      "bmp", 
				      "file-chooser-path", 
				      gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER(chooser)));

    if (filenames)
      {
	bmp_playlist_tracklist_remove_rows_all (bmp_playlist, g_object_get_int (bmp_system_control, "current_tracklist"));
	uri_list = gslist_to_strv (filenames);
	bmp_system_control_add_uri_list (bmp_system_control, (const gchar**)uri_list, -1, -1, &new_ctr, NULL);
	gtk_file_chooser_unselect_all (GTK_FILE_CHOOSER(chooser));
	bmp_system_control_play_list_track (bmp_system_control, g_object_get_int(bmp_system_control, "current_tracklist"), 0);
      }

    bmp_window_set_busy (GTK_WIDGET(chooser));
}
static void
_filebrowser_add (GtkWidget *widget, GtkFileChooser *chooser)
{
    BmpWindowPlaylist *window_playlist;
    GSList *filenames;
    gchar **uri_list;
    gint new_ctr = 0;

    window_playlist =
	RM_GET_ITEM("windows", "window-playlist", NOLOCK);
    filenames =
	gtk_file_chooser_get_uris (GTK_FILE_CHOOSER(chooser));

    bmp_window_set_busy (GTK_WIDGET(chooser));

    xcs_key_set_string (xcs, 
				      "bmp", 
				      "file-chooser-path", 
				      gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER(chooser)));

    if (filenames)
      {
	uri_list = gslist_to_strv (filenames);
	bmp_system_control_add_uri_list (bmp_system_control, (const gchar**)uri_list, -1, -1, &new_ctr, NULL);
	gtk_file_chooser_unselect_all (GTK_FILE_CHOOSER(chooser));
	/* FIXME: Free GSlist */
      }

    bmp_window_set_idle (GTK_WIDGET(chooser));
}

static void
_filebrowser_do_hide_open (GtkWidget *w1, GtkWidget *w2)
{
    if (xcs_key_fetch_boolean(xcs,"bmp","file-chooser-close-on-open"))
	gtk_widget_hide (w2);
}

static void
_filebrowser_do_hide_add (GtkWidget *w1, GtkWidget *w2)
{
    if (xcs_key_fetch_boolean(xcs,"bmp","file-chooser-close-on-add"))
	gtk_widget_hide (w2);
}

static void
_filebrowser_check_hide_open (GtkToggleButton *button, gpointer user_data)
{
    xcs_key_set_boolean(xcs,"bmp","file-chooser-close-on-open",  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)));
}
static void
_filebrowser_check_hide_add (GtkToggleButton *button, gpointer user_data)
{
    xcs_key_set_boolean(xcs,"bmp","file-chooser-close-on-add",  gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)));
}

static void
_filebrowser_save_path (GtkButton *button, GtkFileChooser *chooser) {
    xcs_key_set_string (xcs, "bmp", "file-chooser-path", gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER(chooser)));
}

void
run_filebrowser(gboolean play_button)
{
    static GladeXML *xml = NULL;
    static GtkWidget *dialog = NULL;
    static GtkWidget *chooser = NULL;

    static GtkWidget *button_add, *button_close;
    static GtkWidget *button_select_all, *button_deselect_all;
    static GtkWidget *toggle;

    static GtkFileFilter *file_filter, *audio_filter;

    static gulong handlerid, handlerid_check, handlerid_do;
    static gulong handlerid_activate, handlerid_do_activate;

    if (!xml)
      {

        GtkWidget *alignment;
	GList *containers = NULL;

        xml =
	    glade_xml_new_or_die(Q_("Add/Open Files dialog"),
                                   DATA_DIR "/glade/dialog_add_files.glade",
                                   NULL, NULL);
        glade_xml_signal_autoconnect(xml);

        dialog =
	    glade_xml_get_widget(xml, "add_files_dialog");

        chooser =
	    gtk_file_chooser_widget_new(GTK_FILE_CHOOSER_ACTION_OPEN);
        gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(chooser), TRUE);

	/* Set Path */
	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(chooser), xcs_key_fetch (xcs, "bmp", "file-chooser-path"));

	/* Add File Filters */
	file_filter = gtk_file_filter_new();
	gtk_file_filter_set_name (file_filter, "All Files");
	gtk_file_filter_add_pattern(file_filter, "*");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	audio_filter = gtk_file_filter_new();
	gtk_file_filter_set_name (audio_filter, "Audio Files");
        gtk_file_filter_add_custom (audio_filter, GTK_FILE_FILTER_URI, audio_files_filter, NULL, NULL);
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), audio_filter);

	containers = 
	    plugins_find_containers (); 

        for ( ; containers; containers=containers->next)
	  {
	    ContainerHTuple *tuple = (ContainerHTuple*)containers->data;
	    gchar *glob, *glob_suffix;

	    glob = 
		create_glob (tuple->suffix);
	    glob_suffix = 
		g_strconcat ("*.", glob, NULL);
	    g_free (glob);

	    file_filter = gtk_file_filter_new();
	    gtk_file_filter_set_name (file_filter, tuple->desc);
	    gtk_file_filter_add_pattern(file_filter, glob_suffix); 
	    gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	    g_free (glob_suffix);
	  }

	g_list_foreach (g_list_first (containers), (GFunc)container_h_tuple_free, NULL);

	file_filter = gtk_file_filter_new();
	gtk_file_filter_set_name (file_filter, "FLAC");
	gtk_file_filter_add_pattern(file_filter, "*.[fF][lL][aA][cC]");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	file_filter = gtk_file_filter_new();
	gtk_file_filter_set_name (file_filter, "M4A");
	gtk_file_filter_add_pattern(file_filter, "*.[mM]4[aA]");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	file_filter = gtk_file_filter_new();
	gtk_file_filter_set_name (file_filter, "MP3 Audio");
	gtk_file_filter_add_pattern(file_filter, "*.[mM][pP]3");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	file_filter = gtk_file_filter_new();
	gtk_file_filter_set_name (file_filter, "MPC");
	gtk_file_filter_add_pattern(file_filter, "*.[mM][pP][cC]");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	file_filter = gtk_file_filter_new();
	gtk_file_filter_set_name (file_filter, "Ogg Vorbis");
	gtk_file_filter_add_pattern(file_filter, "*.[oO][gG][gG]");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	file_filter = gtk_file_filter_new();
	gtk_file_filter_set_name (file_filter, "WMA");
	gtk_file_filter_add_pattern(file_filter, "*.[wW][mM][aA]");
	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER(chooser), file_filter);

	gtk_file_chooser_set_filter (GTK_FILE_CHOOSER(chooser), audio_filter);


        alignment = glade_xml_get_widget(xml, "alignment2");
        gtk_container_add(GTK_CONTAINER(alignment), chooser);

        toggle =
	    glade_xml_get_widget(xml, "close_on_action");
        button_select_all =
	    glade_xml_get_widget(xml, "select_all");
        button_deselect_all =
	    glade_xml_get_widget(xml, "deselect_all");
        button_add =
	    glade_xml_get_widget(xml, "action");
	button_close =
	    glade_xml_get_widget(xml, "close");

        g_signal_connect_swapped(button_select_all, "clicked",
                                 G_CALLBACK(gtk_file_chooser_select_all),
                                 chooser);

        g_signal_connect_swapped(button_deselect_all, "clicked",
                                 G_CALLBACK(gtk_file_chooser_unselect_all),
                                 chooser);

        g_signal_connect(button_close, "clicked",
                                 G_CALLBACK(_filebrowser_save_path),
                                 chooser);

        g_signal_connect_swapped(button_close, "clicked",
                                 G_CALLBACK(gtk_widget_hide),
                                 dialog);


#if 0
        g_signal_connect(dialog, "key_press_event",
                         G_CALLBACK(filebrowser_on_keypress),
                         NULL);
#endif

        gtk_widget_show_all(dialog);
      } /* !xml */
    else
      {
        g_signal_handler_disconnect(button_add, handlerid);
        g_signal_handler_disconnect(toggle, handlerid_check);
        g_signal_handler_disconnect(chooser, handlerid_activate);
        g_signal_handler_disconnect(button_add, handlerid_do);
        g_signal_handler_disconnect(chooser, handlerid_do_activate);

	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(chooser), xcs_key_fetch (xcs, "bmp", "file-chooser-path"));
      }

    if (play_button)
      {
        gtk_window_set_title(GTK_WINDOW(dialog), Q_("Open Files"));
        gtk_button_set_label(GTK_BUTTON(button_add), GTK_STOCK_OPEN);
        gtk_button_set_label(GTK_BUTTON(toggle), Q_("Close dialog on Open"));

        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(toggle),
	    xcs_key_fetch_boolean(xcs, "bmp", "file-chooser-close-on-open"));

        handlerid_check =
	    g_signal_connect(toggle,
			"toggled",
			G_CALLBACK(_filebrowser_check_hide_open),
			NULL);

        handlerid =
	    g_signal_connect(button_add,
			"clicked",
			G_CALLBACK(_filebrowser_play),
			chooser);

        handlerid_do =
	    g_signal_connect_after(button_add,
			"clicked",
			G_CALLBACK(_filebrowser_do_hide_open),
			dialog);

        handlerid_activate =
	    g_signal_connect(chooser,
			"file-activated",
			G_CALLBACK(_filebrowser_play),
			chooser);

        handlerid_do_activate =
	    g_signal_connect_after(chooser,
			"file-activated",
			G_CALLBACK(_filebrowser_do_hide_open),
			dialog);

      }
    else
      {
        gtk_window_set_title(GTK_WINDOW(dialog), Q_("Add Files"));
        gtk_button_set_label(GTK_BUTTON(button_add), GTK_STOCK_ADD);
        gtk_button_set_label(GTK_BUTTON(toggle), Q_("Close dialog on Add"));

        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(toggle),
	    xcs_key_fetch_boolean(xcs, "bmp", "file-chooser-close-on-add"));

        handlerid_check =
	    g_signal_connect(toggle,
			"toggled",
			G_CALLBACK(_filebrowser_check_hide_add),
			NULL);

        handlerid =
	    g_signal_connect(button_add,
			"clicked",
			G_CALLBACK(_filebrowser_add),
			chooser);

        handlerid_do =
	    g_signal_connect_after(button_add,
			"clicked",
			G_CALLBACK(_filebrowser_do_hide_add),
			dialog);

        handlerid_activate =
	    g_signal_connect(chooser,
			"file-activated",
			G_CALLBACK(_filebrowser_add),
			chooser);

        handlerid_do_activate =
	    g_signal_connect_after(chooser,
			"file-activated",
			G_CALLBACK(_filebrowser_do_hide_add),
			dialog);

      }

    gtk_window_present(GTK_WINDOW(dialog));
}

void
bmp_ui_callback_add_url (GtkAction *action, gpointer data)
{
    const gchar *glade_file = DATA_DIR "/glade/dialog_add_url.glade";
    static GladeXML *xml = NULL;
    static GtkWidget *add_url_dialog;
    gint response, new_ctr = 0;

    if (!xml)
      {
        xml =
	    glade_xml_new_or_die(Q_("Dialogs"), glade_file, NULL, NULL);
        add_url_dialog =
		glade_xml_get_widget_warn (xml, "dialog_add_url");
      }

    gtk_window_set_position (GTK_WINDOW(add_url_dialog), GTK_WIN_POS_CENTER);
    gtk_widget_set_size_request (GTK_WIDGET(add_url_dialog), 400, -1);

    response =
	gtk_dialog_run (GTK_DIALOG(add_url_dialog));

    switch (response)
      {

	case GTK_RESPONSE_OK:
          {

	    const gchar *uri_list[2];

	    uri_list[0] = gtk_entry_get_text (GTK_ENTRY(glade_xml_get_widget_warn (xml, "entry_url")));
	    uri_list[1] = NULL;

	    bmp_system_control_add_uri_list (bmp_system_control, (const gchar**)uri_list, -1, -1, &new_ctr, NULL);

	  }
        break;

	default: break;

      }

    gtk_entry_set_text (GTK_ENTRY(glade_xml_get_widget_warn (xml, "entry_url")), "");
    gtk_widget_hide (add_url_dialog);
}

void
bmp_ui_callback_add_files (GtkAction *action, gpointer data)
{
    run_filebrowser (FALSE);
}

void
bmp_ui_callback_open_files (GtkAction *action, gpointer data)
{
    run_filebrowser (TRUE);
}

#ifdef HAVE_XINE
void
bmp_ui_callback_add_cd (GtkAction *action, gpointer data)
{
    gchar **mrls, **uri_strv, *disc_id;
    gchar **fragments;
    gchar  *full_uri;
    gint n, n_mrls, new_ctr = 0;

    mrls = bmp_play_get_cd (bmp_play, &n_mrls, &disc_id);
    if ( (! mrls) || (!n_mrls)) return;

    uri_strv = g_new (gchar *, n_mrls+1);
    for (n = 0; n < n_mrls; n++)
      {
	fragments = 
	    g_strsplit (mrls[n],"/",-1);	
	full_uri = 
	    g_strconcat (fragments[0],"//",disc_id,"/",fragments[1],NULL);
	uri_strv[n] = g_strdup (full_uri); 
	g_free (full_uri);
	g_strfreev (fragments);
      }
    uri_strv[n] = NULL;

    bmp_system_control_add_uri_list (bmp_system_control, (const gchar**)uri_strv, -1, -1, &new_ctr, NULL);
    g_strfreev (uri_strv);
}

#else
void
bmp_ui_callback_add_cd (GtkAction *action, gpointer data)
{
}
#endif

void
bmp_ui_callback_del_files (GtkAction *action, gpointer data)
{
    BmpWindowPlaylist *window_playlist;
    ChromaList *list;

    window_playlist = RM_GET_ITEM("windows", "window-playlist", NOLOCK);
    list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);

    bmp_playlist_tracklist_remove_selected (bmp_playlist, g_object_get_int (bmp_system_control, "current_tracklist"));
}

void
bmp_ui_callback_crop_files (GtkAction *action, gpointer data)
{
    BmpWindowPlaylist *window_playlist;
    ChromaList *list;

    window_playlist = RM_GET_ITEM("windows", "window-playlist", NOLOCK);
    list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);

    bmp_playlist_tracklist_remove_unselected (bmp_playlist, g_object_get_int (bmp_system_control, "current_tracklist"));
}

void
bmp_ui_callback_del_files_all (GtkAction *action, gpointer data)
{
    BmpWindowPlaylist *window_playlist;
    ChromaList *list;

    window_playlist = RM_GET_ITEM("windows", "window-playlist", NOLOCK);
    list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);

    bmp_playlist_tracklist_remove_rows_all (bmp_playlist, g_object_get_int (bmp_system_control, "current_tracklist"));
    gtk_action_set_sensitive (gtk_action_group_get_action(bmp_actions, BMP_ACTION_CLEAR_PLAYBACK_HISTORY), FALSE);
}

/* Well, WTF you will be asking here. All the above functions operate on the list itself, while this operates with ChromaList.
   The explanation is like this: The above functions perform stuff directly related to the list, AFTER a selection has been made.
   After a selection has been made, a certain operation on the list items happens, but THIS function does something that is a pure
   UI-thing. It selects all items. Same with unselect all etc */

void
bmp_ui_callback_select_all_files (GtkAction *action, gpointer data)
{
	ChromaList *list;

	list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);
	chroma_list_select_all (list);
}

void
bmp_ui_callback_unselect_all_files (GtkAction *action, gpointer data)
{
	ChromaList *list;

	list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);
	chroma_list_unselect_all (list);
}

void
bmp_ui_callback_invert_selection (GtkAction *action, gpointer data)
{
	ChromaList *list;

	list = CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget);
	chroma_list_invert_selection (CHROMA_LIST(list));
}

void
bmp_ui_callback_sort_current_list (GtkAction *action, gpointer data)
{
	bmp_playlist_sort_current_list (bmp_playlist);
}

void
bmp_ui_callback_sort_current_list_artist (GtkAction *action, gpointer data)
{
	bmp_playlist_sort_current_list_by_tag (bmp_playlist, BMP_DATUM_ARTIST);
}

void
bmp_ui_callback_sort_current_list_album (GtkAction *action, gpointer data)
{
	bmp_playlist_sort_current_list_by_tag (bmp_playlist, BMP_DATUM_ALBUM);
}

void
bmp_ui_callback_sort_current_list_title (GtkAction *action, gpointer data)
{
	bmp_playlist_sort_current_list_by_tag (bmp_playlist, BMP_DATUM_TITLE);
}

void
bmp_ui_callback_sort_current_list_genre (GtkAction *action, gpointer data)
{
	bmp_playlist_sort_current_list_by_tag (bmp_playlist, BMP_DATUM_GENRE);
}

void
bmp_ui_callback_sort_current_list_track (GtkAction *action, gpointer data)
{
	bmp_playlist_sort_current_list_by_tag (bmp_playlist, BMP_DATUM_TRACK);
}

void
bmp_ui_callback_sort_current_list_location (GtkAction *action, gpointer data)
{
	bmp_playlist_sort_current_list_by_tag (bmp_playlist, BMP_DATUM_LOCATION);
}

static void
_entry_set_text (BmpFileItem *item, GladeXML *xml, const gchar *widget, BmpMetadatumId tag_id)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkWidget *entry; 
    gchar *text;
	
    entry = 
      glade_xml_get_widget (xml, widget);
    tuple = 
      bmp_metadata_cache_get_metadata (bmp_metadata_cache, item->uri);
    elem  = 
      h_tuple_get (tuple, bmp_metadata_get_id_static (tag_id));
    
    if (elem->value && (elem->type == G_TYPE_INT)) 
      {
	text = g_strdup_printf("%d", GPOINTER_TO_INT(elem->value));
      }
    else if (elem->value && (elem->type == G_TYPE_STRING)) 
      {
	text = g_strdup (elem->value); 
      }
    else
	text = NULL;

    if (text) 
      {
	gtk_widget_set_sensitive (entry, TRUE); 
      }
    else
      { 
	gtk_widget_set_sensitive (entry, FALSE); 
	text = g_strdup("");
      }

    if (tag_id == BMP_DATUM_LOCATION) 
      { 
	GURI *uri; 
	gchar *text_uri;

	uri = bmp_uri_new (text); 
	bmp_uri_unescape (uri);

	text_uri = bmp_uri_get_string(uri); 

	gtk_entry_set_text (GTK_ENTRY(entry), text_uri);

	g_free (text_uri); 
	bmp_uri_delete (uri);
      } 
    else 
      {
	gtk_entry_set_text (GTK_ENTRY(entry), text); 
      } 

  g_free (text);
}

static void
_label_set_text (BmpFileItem *item, GladeXML *xml, const gchar *widget, BmpMetadatumId tag_id, const gchar *append)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkWidget *label; 
    gchar *text, *text_fq;
	
    label = glade_xml_get_widget (xml, widget);
    tuple = bmp_metadata_cache_get_metadata (bmp_metadata_cache, item->uri);
    elem  = h_tuple_get (tuple, bmp_metadata_get_id_static (tag_id));
    
    if (elem->value && (elem->type == G_TYPE_INT)) 
      {
	text = g_strdup_printf("%d", GPOINTER_TO_INT(elem->value));
      }
    else if (elem->value && (elem->type == G_TYPE_STRING)) 
      {
	text = g_strdup (elem->value); 
      }
    else
	text = NULL;

    if (text) 
      {
	gtk_widget_set_sensitive (label, TRUE); 
      }
    else
      { 
	gtk_widget_set_sensitive (label, FALSE); 
	text = g_strdup("");
      }

  text_fq = g_strdup_printf("%s %s", text, append);
  gtk_label_set_text (GTK_LABEL(label), text_fq);

  g_free (text);
  g_free (text_fq); 
}

static void
run_predicated_filtering (gboolean keep, BmpMetadatumId tag_id)
{
    BmpFileItem *item;
    HTuple      *tuple;
    gint	 _idx_item, _idx_list;
    GList	*paths = NULL;

    paths = chroma_list_get_selected_paths (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));

    if (!paths) return;
    if (g_list_length (paths) > 1)
      {
	gtk_tree_path_free (paths->data);
	return;
      }

    _idx_item = 
	gtk_tree_path_get_indices(paths->data)[0];
    _idx_list =
	g_object_get_int (bmp_system_control, "current_tracklist");
    item = 
	bmp_playlist_tracklist_get_item (bmp_playlist, _idx_list, _idx_item);

    tuple = bmp_metadata_cache_get_metadata (bmp_metadata_cache, item->uri);

    if (keep)
      bmp_playlist_tracklist_keep_predicated (bmp_playlist,
					    _idx_list, 
					    tuple, 
					    tag_id); 
    else
      bmp_playlist_tracklist_remove_predicated (bmp_playlist,
					    _idx_list, 
					    tuple, 
					    tag_id); 

}

void
bmp_ui_callback_remove_artist (GtkAction *action, gpointer data)
{
    run_predicated_filtering (FALSE, BMP_DATUM_ARTIST);
}

void
bmp_ui_callback_keep_artist (GtkAction *action, gpointer data)
{
     run_predicated_filtering (TRUE, BMP_DATUM_ARTIST);
}

void
bmp_ui_callback_remove_album (GtkAction *action, gpointer data)
{
    run_predicated_filtering (FALSE, BMP_DATUM_ALBUM);
}

void
bmp_ui_callback_keep_album (GtkAction *action, gpointer data)
{
     run_predicated_filtering (TRUE, BMP_DATUM_ALBUM);
}
   
static void
fileinfo_update (void)
{

    const gchar *glade_file = DATA_DIR "/glade/dialog_file_info.glade";
    GtkWidget *fi_image;
    static GladeXML *xml;
    BmpListType *list;
    BmpFileItem *item;
    GList *paths;
    gint idx;

    paths = chroma_list_get_selected_paths (CHROMA_LIST(chroma_widgets[BMP_CHROMA_LIST_TRACKLIST]->widget));

    if (! paths) return;
    if (g_list_length (paths) > 1)
    {
	gtk_tree_path_free (paths->data);
	return;
    }

    idx = gtk_tree_path_get_indices(paths->data)[0];

    if (! fi_dialog)
    {
        xml = glade_xml_new_or_die(Q_("Dialogs"), glade_file, NULL, NULL);
	fi_dialog = glade_xml_get_widget(xml, "dialog_file_information");
	gtk_widget_realize (fi_dialog);
	gtk_window_set_position (GTK_WINDOW(fi_dialog), GTK_WIN_POS_CENTER_ALWAYS);
	g_signal_connect_swapped (G_OBJECT(glade_xml_get_widget(xml,"fi_ok")), "clicked", G_CALLBACK(gtk_widget_hide), fi_dialog);
	fi_image = glade_xml_get_widget(xml, "fi_image");
	gtk_image_set_from_file (GTK_IMAGE(fi_image), DATA_DIR "/images/audio.png");
    }

    list = bmp_playlist_tracklist_get_i (bmp_playlist, g_object_get_int (bmp_system_control, "current_tracklist"));
    item = g_list_nth (list->items, idx)->data;

    _entry_set_text(item, xml, "entry_file",	BMP_DATUM_LOCATION);
    _entry_set_text(item, xml, "entry_artist",  BMP_DATUM_ARTIST);
    _entry_set_text(item, xml, "entry_album",   BMP_DATUM_ALBUM);
    _entry_set_text(item, xml, "entry_title",   BMP_DATUM_TITLE);
    _entry_set_text(item, xml, "entry_genre",   BMP_DATUM_GENRE);
    _entry_set_text(item, xml, "entry_date",    BMP_DATUM_DATE);
    _entry_set_text(item, xml, "entry_comment", BMP_DATUM_COMMENT);
    _entry_set_text(item, xml, "entry_track",   BMP_DATUM_TRACK);

    _label_set_text(item, xml, "label_bitrate", BMP_DATUM_BITRATE, "kbit/s");

    gtk_tree_path_free (paths->data);
}

void
bmp_ui_callback_fileinfo_update (void)
{
    if (! fi_dialog) return;
    if (! GTK_WIDGET_VISIBLE(fi_dialog)) return;
    fileinfo_update ();
    gtk_widget_set_size_request (GTK_WIDGET(fi_dialog), 600, -1);
}

void
bmp_ui_callback_fileinfo (GtkAction *action, gpointer data)
{
    fileinfo_update ();
    gtk_widget_set_size_request (GTK_WIDGET(fi_dialog), 600, -1);

    gtk_window_present (GTK_WINDOW(fi_dialog));
}

void
bmp_ui_callback_stop_after_current_track (GtkToggleAction *action, gpointer data)
{
    gboolean active;

    active = gtk_toggle_action_get_active (action);

    bmp_system_control_stop_after_current_set (bmp_system_control, active, NULL);
}

static gboolean
export_run_fc (GtkWidget *button, gpointer data)
{
    GladeXML *xml = (GladeXML*)data;
    GtkFileChooserDialog *dialog = GTK_FILE_CHOOSER_DIALOG(glade_xml_get_widget_warn (xml, "fc_dialog")); 
    GtkWidget *widget;
    gint rv = 0;

    gtk_widget_realize (GTK_WIDGET(dialog));

    widget =
	glade_xml_get_widget_warn (xml, "entry_filename");

    if (strlen(gtk_entry_get_text (GTK_ENTRY(widget)))>0)
      {
	gchar *basename, *folder;

	basename =
	  g_path_get_basename (gtk_entry_get_text (GTK_ENTRY(widget)));
	folder =
	  g_path_get_dirname (gtk_entry_get_text (GTK_ENTRY(widget)));

	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog), folder); 
	gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER(dialog), basename);
      
	g_free (basename);
	g_free (folder);
      }
  
    rv = gtk_dialog_run (GTK_DIALOG(dialog));

    switch (rv)
      {
	case GTK_RESPONSE_CANCEL:
	  {
	  }
	break;

	case GTK_RESPONSE_OK:
	  {
	    gchar *filename;
	    filename =
		gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
	    gtk_entry_set_text (GTK_ENTRY(widget), filename); 
	    g_free (filename);
	  }
      }

    gtk_widget_hide (GTK_WIDGET(dialog));

    return FALSE;
}

static void
export_tree_view_cursor_changed (GtkTreeView *tree_view,
				 gpointer user_data)
{
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GtkWidget *entry;
  GString *result;
  GladeXML *xml = (GladeXML*)user_data;
  gchar *composited = NULL, *suffix = NULL;
  gint position;

  selection = 
      gtk_tree_view_get_selection (tree_view);
  entry =
      glade_xml_get_widget_warn (xml, "entry_filename");
  result =
      g_string_new (gtk_entry_get_text (GTK_ENTRY(entry)));
  
  position = strlen (result->str)-1;

  while ( (result->str[position] != '.') && position >= 0)
    {
      position--;
    }

  if (position != 0)
      g_string_truncate (result, position);
  
  gtk_tree_selection_get_selected (selection, &model, &iter);
  gtk_tree_model_get (model, &iter, 1, &suffix, -1);
  composited = g_strconcat (result->str, ".", suffix, NULL); 

  gtk_entry_set_text (GTK_ENTRY(entry), composited);
}

void
bmp_ui_callback_export_tracklist (GtkAction *action, gpointer data)
{
  static GladeXML *xml = NULL;
  static GtkWidget *dialog_window = NULL;
  static GtkWidget *tree_view = NULL;

  const gchar *glade_file = DATA_DIR "/glade/dialog_export_tracklist.glade";

  GtkTreeSelection *selection;
  GtkTreeModel *store;
  GtkTreeIter iter;
  gchar *listname = NULL, *suffix = NULL, *suffix_fq = NULL;
  gint rv = 0;

  if (! dialog_window)
    {
	GtkTreeIter iter;
	GtkListStore *store = NULL;
	GtkTreeViewColumn *column = NULL;
	GtkCellRenderer *renderer = NULL;
	GList *containers = NULL;

	xml =
	    glade_xml_new_or_die(Q_("Dialog"), glade_file, NULL, NULL);
	dialog_window =
	    glade_xml_get_widget(xml, "export_dialog");
	tree_view =
	    glade_xml_get_widget(xml, "treeview_plugins");
	column =
	    gtk_tree_view_column_new ();
	renderer =
	    gtk_cell_renderer_text_new ();

	gtk_tree_view_column_pack_start
		(column,
		 renderer,
		 FALSE);

	gtk_tree_view_column_set_attributes
		(column,
		 renderer, 
		 "text",
		  0,
		  NULL);

	gtk_tree_view_column_set_expand
		(column,
		 FALSE);

	gtk_tree_view_column_set_title
		(column,
		 N_("Export Plugin")); 

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (tree_view),
		 GTK_TREE_VIEW_COLUMN (column));

	containers = 
	    plugins_find_containers (); 
	store = 
	    gtk_list_store_new (3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	
	for ( ; containers; containers=containers->next)
	  {
	    ContainerHTuple *tuple = (ContainerHTuple*)containers->data;

	    gtk_list_store_append (GTK_LIST_STORE(store), &iter);
	    gtk_list_store_set (GTK_LIST_STORE(store), &iter, 0, tuple->desc, 1, tuple->suffix, 2, tuple->name, -1);
	  }

	gtk_tree_view_set_model (GTK_TREE_VIEW(tree_view), GTK_TREE_MODEL(store));

	gtk_widget_realize (dialog_window);

	g_object_connect (G_OBJECT(glade_xml_get_widget_warn(xml, "button_browse")), "signal::clicked", export_run_fc, xml, NULL);
	g_object_connect (G_OBJECT(tree_view), "signal::cursor-changed", export_tree_view_cursor_changed, xml, NULL);
    }
  else
    {
      store = 
	  gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
      selection =
	  gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
      gtk_tree_model_get_iter_first (store, &iter);
      gtk_tree_selection_select_iter (selection, &iter);
    }

  store = 
      gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));

  gtk_tree_model_get_iter_first (store, &iter);
  gtk_tree_model_get (store, &iter, 1, &suffix, -1);

  suffix_fq = g_strconcat (".", suffix, NULL);

  listname = g_strconcat (g_get_home_dir(), 
		     G_DIR_SEPARATOR_S,
		     bmp_playlist_get_nth_tracklist_name (bmp_playlist, 
							  g_object_get_int (bmp_system_control, "current_tracklist")),
		     suffix_fq,
		     NULL);
  g_free (suffix_fq);
  gtk_entry_set_text (GTK_ENTRY(glade_xml_get_widget_warn (xml, "entry_filename")), listname);

  rv = gtk_dialog_run (GTK_DIALOG(dialog_window)); 

  switch (rv)
    {
	case GTK_RESPONSE_CANCEL: break;

	case GTK_RESPONSE_OK:
	  {
	    GtkTreeModel *store;
	    GtkTreeIter iter;
	    gchar *container;
	    gboolean rv;

	    gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW(glade_xml_get_widget_warn (xml, "treeview_plugins"))), &store, &iter);	    
	    gtk_tree_model_get (GTK_TREE_MODEL(store), &iter, 2, &container, -1);

	    rv = bmp_system_control_export_tracklist (bmp_system_control,
						      container,
						      g_object_get_int (bmp_system_control, "current_tracklist"),
						      !gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(glade_xml_get_widget_warn(xml, "rb_options1"))),
						      gtk_entry_get_text (GTK_ENTRY(glade_xml_get_widget_warn (xml, "entry_filename"))),
						      NULL);  

	    if (!rv)
	      {
		// ERROR HERE	
	      }
	  }
    }

  gtk_widget_hide (dialog_window);
}

void
bmp_ui_callback_library (GtkAction *action, gpointer data)
{
    invoke_dialog_library ();
}
