#include <stdlib.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>
#include <string.h>

#include "main.h"
#include "system_control.h"
#include "resource_manager.h"

#include "ui.h"
#include "ui_util.h"

#include "bmpx/util.h"
#include "bmpx/uri.h"
#include "bmpx/playlist.h"
#include "bmpx/metadata.h"

#include "xcs/xml.h"
#include "xcs/xcs.h"
#include "goa/libgoa.h"

#include "glade.h"
#include "plugin_priv.h"

#include "embedded-images/media-pixbufs.h"


// FIXME: This code currently assumes all the browse_level metadata ids are of G_TYPE_STRING
static GtkTreeStore*	  tree_store_browse    = NULL; 
static GHashTable*	  map_root = NULL;

static GtkWidget *dialog_window = NULL;

enum {
      BROWSE_COL_IMG,
      BROWSE_COL_KEY,
      BROWSE_COL_TUPLE,
      BROWSE_COL_MAP,
      BROWSE_COL_TRACK
};


static gint browse_levels[] = {
    BMP_DATUM_ARTIST,
    BMP_DATUM_ALBUM,
    BMP_DATUM_TITLE
};

static void
update_browse_view (BmpMetadataCache *cache, gpointer _tuple, gpointer user_data)
{
    HTuple *tuple = (HTuple*)_tuple;
    HTupleElem *elem;

#define N_LEVELS 3

    GdkPixbuf*	  pb[N_LEVELS];
    GtkTreeIter	  iter[N_LEVELS];
    GHashTable*	  map[N_LEVELS];
    GHashTable*	  map_c;
    gint	  n;
    gchar*	  key;

    elem = h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));

    if (g_ascii_strncasecmp (elem->value, "file:", 5))
      return;

    map[0] = map_root;

    pb[0]  = gdk_pixbuf_new_from_inline (-1, icon_artist, FALSE, NULL);
    pb[1]  = gdk_pixbuf_new_from_inline (-1, icon_album,  FALSE, NULL);
    pb[2]  = gdk_pixbuf_new_from_inline (-1, icon_title,  FALSE, NULL);

    // FIXME: This can be written much more optimized probably, and we have to get rid of the semi-hardcoding
  
    map_c = map[0];

    for (n = 0; n < N_LEVELS; n++)
      {
	elem = h_tuple_get (tuple, bmp_metadata_get_id_static (browse_levels[n])); 
	key = g_strdup (elem->value);
	g_strstrip (key);

	if ((!key) || (!strlen(key)))
	  {
	    g_object_unref (pb[0]);
	    g_object_unref (pb[1]);
	    g_object_unref (pb[2]);
	    g_free (key);
	    return;
	  }

	if (!attr_row_map_mapping_exists (map_c, key))
          {	
	      if (n == 0)
		  gtk_tree_store_append (tree_store_browse, &iter[n], NULL); 
	      else
		  gtk_tree_store_append (tree_store_browse, &iter[n], &iter[n-1]); 

	      gtk_tree_store_set (tree_store_browse, 
			     &iter[n], 
			     BROWSE_COL_IMG, 
			     pb[n], 
			     BROWSE_COL_KEY, 
			     key, 
			     BROWSE_COL_TUPLE, 
			     h_tuple_clone (tuple), 
			     BROWSE_COL_MAP, 
			     attr_row_map_new (),
			     -1); 

	      attr_row_map_mapping_insert (map_c, key, GTK_TREE_MODEL(tree_store_browse), &iter[n]);
	  }
	else if (!attr_row_map_mapping_get (map_c, key, GTK_TREE_MODEL(tree_store_browse), &iter[n])) 
	  {
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s: Mapping for key '%p' disappeared", G_STRLOC, key); 
	    g_free (key);
	    break;
	  }

	g_free (key);
	gtk_tree_model_get (GTK_TREE_MODEL(tree_store_browse), &iter[n], BROWSE_COL_MAP, &map_c, -1); 
    }

    g_object_unref (pb[0]);
    g_object_unref (pb[1]);
    g_object_unref (pb[2]);
}

static gint
tree_store_browse_sort_func (GtkTreeModel *model,
			     GtkTreeIter  *a,
			     GtkTreeIter  *b,
			     gpointer      user_data)
{
    HTuple *tuple_a, *tuple_b;
    GtkTreePath *path_a, *path_b;
    gchar *key_a, *_key_a;
    gchar *key_b, *_key_b;
    gint cmp_result = 0;

    gtk_tree_model_get (model, a, BROWSE_COL_KEY, &_key_a, BROWSE_COL_TUPLE, &tuple_a, -1);
    gtk_tree_model_get (model, b, BROWSE_COL_KEY, &_key_b, BROWSE_COL_TUPLE, &tuple_b, -1);

    key_a = g_utf8_casefold (_key_a, strlen (_key_a));
    key_b = g_utf8_casefold (_key_b, strlen (_key_b));

    path_a =
      gtk_tree_model_get_path (model, a);

    path_b =
      gtk_tree_model_get_path (model, b);

    if (path_b && (gtk_tree_path_get_depth (path_b) == 3))
      {
	HTupleElem *elem_a, *elem_b;
	gint track_a, track_b;

#if 0 
	elem_a = h_tuple_get (tuple_a, bmp_metadata_get_id_static (BMP_DATUM_LOCATION));
	elem_b = h_tuple_get (tuple_b, bmp_metadata_get_id_static (BMP_DATUM_LOCATION));
	g_message ("URI a: %s", elem_a->value);
	g_message ("URI b: %s", elem_b->value);
#endif

	if ((!tuple_a) || (!tuple_b))
	  return 0;

	elem_a = h_tuple_get (tuple_a, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	elem_b = h_tuple_get (tuple_b, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	track_a = GPOINTER_TO_INT (elem_a->value);
	track_b = GPOINTER_TO_INT (elem_b->value);
#if 0
	g_message ("Track a: %d", track_a); 
	g_message ("Track b: %d", track_b); 
#endif

	gtk_tree_path_free (path_a);
	gtk_tree_path_free (path_b);

	if (track_a  < track_b)
	  cmp_result = -1;
  
	if (track_a == track_b)
	  cmp_result =  0;

	if (track_a >  track_b)
	  cmp_result =  1;

//	if ((track_a != 0) && (track_b != 0))
	  goto done;
      }
    else 
      {
	if (path_a)
	  gtk_tree_path_free (path_a);

	if (path_b)
	  gtk_tree_path_free (path_b);
      }

    if ((key_a != NULL) && (key_b == NULL))
      {
	cmp_result = 1;
	goto done;
      }

    if ((key_a == NULL) && (key_b == NULL))
      {
	cmp_result = 0;
	goto done;
      }

    if ((key_a == NULL) && (key_b != NULL))
      {
	cmp_result = -1;
	goto done;
      }
    
    cmp_result = g_utf8_collate (key_a, key_b);

    done:

#if 0
    g_free (_key_a);
    g_free (_key_b);

    g_free (key_a);
    g_free (key_b);
#endif

    return cmp_result;
}

void
library_metadata_hookup (BmpMetadataCache *cache)
{
    tree_store_browse = gtk_tree_store_new (5, 
					    GDK_TYPE_PIXBUF, 
					    G_TYPE_STRING, 
					    G_TYPE_POINTER, 
					    G_TYPE_POINTER,
					    G_TYPE_INT);

    gtk_tree_sortable_set_sort_func
                                            (GTK_TREE_SORTABLE (tree_store_browse),
					     1,
                                             (GtkTreeIterCompareFunc) tree_store_browse_sort_func,
                                             NULL, 
                                             NULL); 

    gtk_tree_sortable_set_sort_column_id 
					    (GTK_TREE_SORTABLE (tree_store_browse),
					     1, 
					     GTK_SORT_ASCENDING);

    map_root = attr_row_map_new ();

    g_object_connect (G_OBJECT (cache), 
		      "signal::file-added",
		      G_CALLBACK (update_browse_view),
		      NULL,
		      NULL);		
}

static void
on_library_add_path_clicked (GtkButton *button, gpointer data) 
{
    GtkWidget *dialog, *notebook_library_add, *window;
    GList *items = NULL;
    GError *error = NULL;
    gchar *folder;
    gint rv;

    g_return_if_fail (GTK_IS_BUTTON(button));

    window = glade_xml_get_widget_warn (GLADE_XML(data), "dialog_library");

    dialog = gtk_file_chooser_dialog_new (Q_("Select Path To Add"),
                                          GTK_WINDOW (glade_xml_get_widget_warn (GLADE_XML(data), 
					  "dialog_library")),
                                          GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                          GTK_STOCK_CANCEL,
                                          GTK_RESPONSE_CANCEL,
                                          GTK_STOCK_ADD,
                                          GTK_RESPONSE_OK,
                                          NULL);

    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER(dialog),  
					 xcs_key_fetch (xcs, 
					 "bmp", 
					 "file-chooser-path"));

    rv = gtk_dialog_run (GTK_DIALOG (dialog));

    if (rv == GTK_RESPONSE_CANCEL)
      {
	gtk_widget_destroy (dialog);
	return;
      }

    folder = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(dialog));
    gtk_widget_destroy (dialog);

    notebook_library_add = glade_xml_get_widget_warn (GLADE_XML(data), "notebook_library_add");
    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_library_add), 1);
    bmp_window_set_busy (window);

    dir_foreach(folder, (DirForeachFunc)collect_path,
                &items, &error);

    for ( ; items ; items = items->next)
      {
	gchar *uri;

	uri = g_strconcat ("file://", items->data, NULL);
	bmp_metadata_cache_get_metadata (bmp_metadata_cache, uri);
	g_free (uri);

	while (gtk_events_pending()) gtk_main_iteration();
      }

    g_list_foreach (g_list_first(items), (GFunc)g_free, NULL);
    g_free (folder);

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_library_add), 0);
    bmp_window_set_idle (window);
}

/* Library search dialog */
static GtkTreeViewColumn* library_column_search[4];
static GtkCellRenderer*   library_renderer_search[4];

static GtkTreeViewColumn* library_column_browse[1];
static GtkCellRenderer*   library_renderer_browse[3];

static gint search_store_ids[] = {
	BMP_DATUM_ARTIST,
	BMP_DATUM_ALBUM,
	BMP_DATUM_TRACK,
	BMP_DATUM_TITLE,
	BMP_DATUM_LOCATION,
};

static GType search_store_types[] = {
	  G_TYPE_STRING,
	  G_TYPE_STRING,
	  G_TYPE_INT,
	  G_TYPE_STRING,
	  G_TYPE_STRING	     
};

static void
on_button_enqueue_clicked (GtkButton *button, GladeXML *xml)
{
    GtkWidget*	      window;
    GtkNotebook*      library_notebook;
    GtkWidget*	      tree_view;
    GtkTreeModel*     store;
    GtkTreeSelection* selection;
    GList*	      paths = NULL;
    GList*	      uris = NULL;
    GtkTreeIter	      iter, iter_children, iter_children_children;
    gchar**	      uri_list;
    gchar*	      uri;
    gint	      new_ctr = 0, current_page = -1;

    window = glade_xml_get_widget_warn (xml, "dialog_library");
    bmp_window_set_busy (window);

    library_notebook = GTK_NOTEBOOK (glade_xml_get_widget_warn (xml, "library_notebook"));
    current_page = gtk_notebook_get_current_page (library_notebook);

    switch (current_page)
      {
	case 3:
	  {
	    tree_view =
		glade_xml_get_widget_warn (xml, "tree_view_icecast");
	    store =
		gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
	    selection =
		gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
	    paths =
		gtk_tree_selection_get_selected_rows (selection, NULL);

	    if (!paths)
	      {
		bmp_window_set_idle (window);
		return;
	      }

	    while (paths)
	      {
		gtk_tree_model_get_iter (store, &iter, paths->data);
		gtk_tree_model_get (store, &iter, 4, &uri, -1);

		uris = g_list_append (uris, uri);
		paths = g_list_next (paths);
	      }

	      uri_list = glist_to_strv (uris);

	      bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    -1,
					    &new_ctr,
					    NULL);

	      g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
	      g_list_free (paths);
	  }
	  break;

	case 2:
	  {
	    tree_view =
		glade_xml_get_widget_warn (xml, "tree_view_shoutcast");
	    store =
		gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
	    selection =
		gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
	    paths =
		gtk_tree_selection_get_selected_rows (selection, NULL);

	    if (!paths)
	      {
		bmp_window_set_idle (window);
		return;
	      }

	    while (paths)
	      {
		gtk_tree_model_get_iter (store, &iter, paths->data);
		gtk_tree_model_get (store, &iter, 5, &uri, -1);

		uris = g_list_append (uris, uri);
		paths = g_list_next (paths);
	      }

	      uri_list = glist_to_strv (uris);

	      bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    -1,
					    &new_ctr,
					    NULL);

	      g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
	      g_list_free (paths);
	  }
	  break;

      case 1: 
	{
	  tree_view =
      	    glade_xml_get_widget_warn (xml, "tree_view_search");
	  store =
	    gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
	  selection =
	    gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
	  paths =
	    gtk_tree_selection_get_selected_rows (selection, NULL);

	  if (!paths)
	    {
	      gtk_tree_selection_select_all (selection);
	      paths =
		gtk_tree_selection_get_selected_rows (selection, NULL);
	    }

	  while (paths)
	    {
	      gtk_tree_model_get_iter (store, &iter, paths->data);
	      gtk_tree_model_get (store, &iter, 4, &uri, -1);

	      uris = g_list_append (uris, uri);
	      paths = g_list_next (paths);
	    }

	    uri_list = glist_to_strv (uris);

	    bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    -1,
					    &new_ctr,
					    NULL);

	    g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
	    g_list_free (paths);

	    gtk_tree_selection_unselect_all (selection);
	}
	break;

    case 0: 
      {
        tree_view =
      	    glade_xml_get_widget_warn (xml, "tree_view_browse");
        store =
	    gtk_tree_view_get_model (GTK_TREE_VIEW(tree_view));
	selection =
	    gtk_tree_view_get_selection (GTK_TREE_VIEW(tree_view));
        paths =
	    gtk_tree_selection_get_selected_rows (selection, NULL);

	if (!paths)
	  {
	    gtk_tree_selection_select_all (selection);
	    paths =
		gtk_tree_selection_get_selected_rows (selection, NULL);
	  }

        while (paths)
	  {
	      gtk_tree_model_get_iter (store, &iter, paths->data);

	      switch (gtk_tree_path_get_depth (paths->data))
		  {
			// FIXME: Yay wow man is this efficient, can't hold onto my chair 
			case 1:
			  {
			    gboolean run_1 = TRUE, run_2 = TRUE;

			    run_1 = 
			      gtk_tree_model_iter_children (store, &iter_children, &iter);

			    while (run_1)
			      {
				run_2 = 
				  gtk_tree_model_iter_children (store, &iter_children_children, &iter_children);

				while (run_2)
				  {
				    HTuple *tuple;
				    HTupleElem *elem;

				    gtk_tree_model_get (store, &iter_children_children, BROWSE_COL_TUPLE, &tuple, -1); 

				    elem = 
				      h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));
				    uris = 
				      g_list_append (uris, g_strdup (elem->value));
				    run_2 = 
				      gtk_tree_model_iter_next (store, &iter_children_children);
				  }

				  run_1 = 
				    gtk_tree_model_iter_next (store, &iter_children);
			      }

			  } break;
  
			case 2:
			  {
			    gboolean run = TRUE;

			    run = 
			      gtk_tree_model_iter_children (store, &iter_children, &iter);

			    while (run)
			      {
				HTuple *tuple;
				HTupleElem *elem;

				gtk_tree_model_get (store, &iter_children, BROWSE_COL_TUPLE, &tuple, -1); 

				elem = 
				  h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));
				uris = 
				  g_list_append (uris, g_strdup (elem->value));
				run = 
				  gtk_tree_model_iter_next (store, &iter_children);
			      }

			  } break;
	
			case 3:
			  {
			    HTuple *tuple;
			    HTupleElem *elem;

			    gtk_tree_model_get (store, &iter, BROWSE_COL_TUPLE, &tuple, -1); 

			    elem = 
			      h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_METADATA_PKEY));
			    uris = 
			      g_list_append (uris, g_strdup (elem->value));
			  } break;
			
			default: /* Uhm... */ break; 
		  }

	      paths = paths->next;
	  }

	uri_list = glist_to_strv (uris);

	bmp_system_control_add_uri_list (bmp_system_control,
					    (const gchar**)uri_list,
					    -1,
					    -1,
					    &new_ctr,
					    NULL);

	g_list_foreach (uris, (GFunc)g_free, NULL);
	g_list_foreach (paths, (GFunc)gtk_tree_path_free, NULL);
	g_list_free (paths);
	g_list_free (uris);

	gtk_tree_selection_unselect_all (selection);
    }
    break;

    default: break;
  }

  bmp_window_set_idle (window);
}

static void
on_button_play_clicked (GtkButton *button, GladeXML *xml)
{
    gint tracklist;

    tracklist = g_object_get_int (bmp_system_control, "current_tracklist");

    bmp_playlist_tracklist_remove_rows_all (bmp_playlist, tracklist);
    on_button_enqueue_clicked (GTK_BUTTON (glade_xml_get_widget_warn (xml, "button_enqueue")), xml);
    bmp_system_control_play_list_track (bmp_system_control, tracklist, 0);
}

static void
on_tree_view_search_row_activated (GtkTreeView *treeview,
                                   GtkTreePath *path,
                                   GtkTreeViewColumn *column,
                                   GladeXML *xml)
{
    GtkTreeModel *store;
    GtkTreeIter iter;
    gchar *uri_list[2];
    gchar *uri;
    gint new_ctr = 0;

    store =
	gtk_tree_view_get_model (treeview);

    gtk_tree_model_get_iter (store, &iter, path);
    gtk_tree_model_get (store, &iter, 4, &uri, -1);

    uri_list[0] =
	    uri;
    uri_list[1] =
	    NULL;

    bmp_system_control_add_uri_list (bmp_system_control,
				(const gchar**)uri_list,
				-1,
				-1,
				&new_ctr,
				NULL);
}

static void
on_cbox_tag_id_changed (GtkComboBox *cbox, GladeXML *xml) {

    GtkWidget *editable;

    g_return_if_fail (GTK_IS_COMBO_BOX (cbox));

    editable = glade_xml_get_widget_warn (xml, "entry_value");
    g_signal_emit_by_name (editable, "changed");
}

static void
on_cb_exact_changed (GtkToggleButton *button, GladeXML *xml) {

    GtkWidget *editable;

    g_return_if_fail (GTK_TOGGLE_BUTTON (button));

    editable = glade_xml_get_widget_warn (xml, "entry_value");
    g_signal_emit_by_name (editable, "changed");
}

void 
_liststore_fill (HTuple *tuple, gpointer user_data)
{
    GtkListStore *list_store = GTK_LIST_STORE(user_data);
    GtkTreeIter   iter;
    HTupleElem	 *elem;
    gint n;

    gtk_list_store_append (list_store, &iter);

    for (n = 0; n < G_N_ELEMENTS(search_store_ids); n++) 
      {
	elem = h_tuple_get (tuple, bmp_metadata_get_id_static (search_store_ids[n]));

	switch (elem->type)
	  {
	    case G_TYPE_INT:
	      {
		  gtk_list_store_set (list_store, &iter, n, GPOINTER_TO_INT(elem->value), -1);
	      }
	      break;

	    case G_TYPE_STRING:
	      {
		  gtk_list_store_set (list_store, &iter, n, (gchar*)elem->value, -1);
	      }
	      break;
	  }
      }
}

static void
on_entry_value_changed (GtkEditable *editable, GladeXML *xml)
{
    GtkWidget	  *tree_view;
    GtkListStore  *list_store;
    HRelation	  *result;

    Query	  *query = NULL;

    gint	   tag_id;

    gchar	  *match_value_s;
    gint	   match_value_i;
    
    gboolean	   exact_match;

    gchar	  *text;

    g_return_if_fail (GTK_IS_EDITABLE(editable));

    tree_view =
	glade_xml_get_widget_warn (xml, "tree_view_search");
    list_store =
	gtk_list_store_newv (G_N_ELEMENTS(search_store_types), search_store_types);
    tag_id =
	gtk_combo_box_get_active (GTK_COMBO_BOX (glade_xml_get_widget_warn (xml, "cbox_tag_id")));
    text =
	gtk_editable_get_chars (editable, 0, -1);
    exact_match =
	gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (glade_xml_get_widget_warn (xml, "cb_exact")));

    if (bmp_metadata_get_type (tag_id) == G_TYPE_INT)
      {
	match_value_i = 
	  strtol (text, NULL, 10);
        query =
	  bmp_metadata_cache_query_construct (bmp_metadata_cache,
					      1,
					      tag_id,
					      exact_match,
					      match_value_i);

      }
    else if (bmp_metadata_get_type (tag_id) == G_TYPE_STRING)
      {
	match_value_s = text;
        query =
	  bmp_metadata_cache_query_construct (bmp_metadata_cache,
					      1,
					      tag_id,
					      exact_match,
					      match_value_s);
      }
    
    if (!query)
      {
	g_message ("%s: No query", G_STRLOC);
	return;
      }
    
    result = h_relation_sort (
			       h_relation_sort (
						 h_relation_sort (bmp_metadata_cache_query_run (bmp_metadata_cache, query),
								 (GCompareDataFunc) h_tuple_compare_elem, 
								 (gpointer)bmp_metadata_get_id_static (BMP_DATUM_TRACK)),

					         (GCompareDataFunc) h_tuple_compare_elem, 
					         (gpointer)bmp_metadata_get_id_static (BMP_DATUM_ARTIST)),

			      (GCompareDataFunc) h_tuple_compare_elem, 
			      (gpointer)bmp_metadata_get_id_static (BMP_DATUM_ALBUM));

    if (result)
      {
	h_relation_for_each_tuple (result, (HTupleFunc)_liststore_fill, list_store); 
	h_relation_unref (result);
	gtk_tree_view_set_model (GTK_TREE_VIEW(tree_view), GTK_TREE_MODEL(list_store));
      }
    else
      {
	gtk_tree_view_set_model (GTK_TREE_VIEW(tree_view), NULL); 
      }

    gtk_tree_view_columns_autosize (GTK_TREE_VIEW(tree_view));
}

static gboolean
on_library_dialog_key_press (GtkWidget *widget,
		    GdkEventKey *event,
		    gpointer user_data)
{
    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    if (event->keyval == GDK_Escape)
	gtk_widget_hide (widget);

    return FALSE;
}

void
tracknumber_renderer_func          (GtkTreeViewColumn *tree_column,
                                             GtkCellRenderer *cell,
                                             GtkTreeModel *tree_model,
                                             GtkTreeIter *iter,
                                             gpointer data)
{
    HTuple *tuple;
    HTupleElem *elem;
    GtkTreePath *path;
    gchar *cell_data;

    path = gtk_tree_model_get_path (tree_model, iter);

    if (gtk_tree_path_get_depth (path) < 3)
      {
	g_object_set (G_OBJECT(cell), "visible", FALSE, NULL);
	g_object_set (G_OBJECT(cell), "text", "0", NULL);
      }
    else
      {
	g_object_set (G_OBJECT(cell), "visible", TRUE,  NULL);

	gtk_tree_model_get (tree_model, iter, BROWSE_COL_TUPLE, &tuple, -1);

	elem = 
	  h_tuple_get (tuple, bmp_metadata_get_id_static (BMP_DATUM_TRACK));
	cell_data = 
	  g_strdup_printf ("%d.", GPOINTER_TO_INT(elem->value));

	g_object_set (G_OBJECT(cell), "text", cell_data, NULL);
	g_free (cell_data);
      }
}

//----------- STREAM LISTERS

static gboolean
model_filter_func (GtkTreeModel *model,
		   GtkTreeIter *iter,
                   gpointer data)
{
    GtkEntry *entry = GTK_ENTRY(data);
    gchar    *_name,  *name,
	     *_genre, *genre,
	     *entry_text_lc;

    const gchar *entry_text;

    gtk_tree_model_get (model, iter, 0, &_genre, 1, &_name, -1);
    entry_text = gtk_entry_get_text (entry);

    if ((!entry_text) || (strlen(entry_text)==0))
      return TRUE;

    entry_text_lc = g_utf8_strdown (entry_text, -1);
  
    name = g_utf8_strdown (_name, -1); 
    genre = g_utf8_strdown (_genre, -1); 

    if ((match_keys (name, entry_text_lc) != NULL) ||
        (match_keys (genre, entry_text_lc) != NULL)) 
      {
	g_free (entry_text_lc);
	g_free (_name);
	g_free (name);
	g_free (_genre);
	g_free (genre);
	return TRUE;
      }

  g_free (entry_text_lc);
  g_free (_name);
  g_free (name);
  g_free (_genre);
  g_free (genre);

  return FALSE;  
}

typedef struct _SortDataTuple SortDataTuple;
struct _SortDataTuple {
      GtkTreeView   *tree_view;
      gint	     column_id;
};

void
column_clicked (GtkTreeViewColumn *tree_view_column,
		gpointer user_data)
{
    GtkTreeViewColumn *column;
    GtkTreeModel      *model_filter;
    GtkTreeModel      *model_real;
    GtkTreeView	      *tree_view;
  
    GtkSortType	       sort_type, sort_type_new;
    gint	       sort_id;

    SortDataTuple *sort_data_tuple = (SortDataTuple*)user_data;    

    tree_view = sort_data_tuple->tree_view; 
    model_filter = gtk_tree_view_get_model (tree_view);
    model_real = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (model_filter));

    gtk_tree_sortable_get_sort_column_id (GTK_TREE_SORTABLE(model_real), &sort_id, &sort_type);
    if ((sort_id >= 0) && (sort_id != sort_data_tuple->column_id))
      {
	column = gtk_tree_view_get_column (tree_view, sort_id);
	gtk_tree_view_column_set_sort_indicator (column, FALSE);
      }

    if (sort_id >= 0)
      sort_type_new = (sort_type == GTK_SORT_ASCENDING) ? GTK_SORT_DESCENDING : GTK_SORT_ASCENDING;
    else
      sort_type_new = GTK_SORT_ASCENDING;

    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE(model_real), sort_data_tuple->column_id, sort_type_new);
    gtk_tree_view_column_set_sort_indicator (tree_view_column, TRUE);
    gtk_tree_view_column_set_sort_order (tree_view_column, sort_type_new);
}

static gchar* shoutcast_headers[] = {
      N_("Stream Genre"),
      N_("Stream Name"),
      N_("Now Playing"),
      N_("Listeners"),
      N_("Bitrate")
};

static gchar* icecast_headers[] = {
      N_("Stream Genre"),
      N_("Stream Name"),
      N_("Now Playing"),
      N_("Bitrate")
};

static gchar* peercast_headers[] = {
      N_("Stream Genre"),
      N_("Stream Name"),
      N_("Now Playing"),
      N_("Bitrate")
};

#if 0
static void
on_refresh_peercast_clicked (GtkWidget	  *widget,
			     gpointer	   data)
{
  xmlDocPtr	      doc;
  xmlXPathObjectPtr   xpathobj;
  xmlNodeSetPtr	      nv;

  GtkTreeModelFilter *list_store_peercast_filter;
  GtkListStore	     *list_store_peercast;
  GtkTreeIter	      iter;

  VFSHandle	     *handle;
  gint		      n;

  typedef struct _PeerCastEntry PeerCastEntry;
  struct _PeerCastEntry { 
      gchar   *Name;
      gchar   *Genre;
      gchar   *Nowplaying;

      gint     Bitrate;

      gchar   *URI;
  };

#define PEERCAST_LIST_URI "http://yp.peercast.org/?xml"

#define PN_URI	     "host"
#define PP_URI	     "ip"

#define PN_NP	     "track"
#define PP_NP	     "title"

#define PP_NAME	     "name"
#define PP_GENRE     "genre"
#define PP_BITRATE   "bitrate"
  
  handle = vfs_handle_new ();  
  handle->uri = g_strdup (PEERCAST_LIST_URI);
  handle->need_container = FALSE;
  if (!vfs_handle_read (handle))
    return;

  doc = xmlParseMemory (handle->buffer, strlen(handle->buffer));
  if (!doc)
    return;

  vfs_handle_free (handle);

  xpathobj = xml_execute_xpath_expression (doc, "//channel", NULL);
  nv = xpathobj->nodesetval;

  list_store_peercast = gtk_list_store_new (5,
					     G_TYPE_STRING,	// Name
					     G_TYPE_STRING,	// Genre
					     G_TYPE_STRING,	// Nowplaying
					     G_TYPE_INT,	// Bitrate
					     G_TYPE_STRING);	// URI

  for (n = 0; n < nv->nodeNr; n++) 
    {
      PeerCastEntry	*entry;
      gchar		*id, *ip, *title, *xpath, *aux;
      xmlXPathObjectPtr  xpathobj_aux;
      xmlNodePtr	 node, node_child;

      node = 
	nv->nodeTab[n];

      entry = g_new0 (PeerCastEntry,1);

      entry->Name    = xmlGetProp (node, PP_NAME);
      entry->Genre   = xmlGetProp (node, PP_GENRE);
      entry->Bitrate = xmlGetProp (node, PP_BITRATE);

      id	     = xmlGetProp (node, "id");

      // Get URI
      xpath	     = g_strdup_printf("//channel[@id='%s']//%s", id, PN_URI);
      xpathobj_aux   = xml_execute_xpath_expression (doc, xpath, NULL);
      g_free (xpath);

      ip	    = xmlGetProp (xpathobj_aux->nodesetval->nodeTab[0], PP_URI); 
      entry->URI    = g_strdup_printf ("http://%s", ip);
      g_free (ip);
      xmlXPathFreeObject (xpathobj_aux);

      // Get Title
      xpath	     = g_strdup_printf("//channel[@id='%s']//%s", id, PN_NP);
      xpathobj_aux   = xml_execute_xpath_expression (doc, xpath, NULL);
      g_free (xpath);
      entry->Nowplaying = xmlGetProp (xpathobj_aux->nodesetval->nodeTab[0], PP_NP); 
      xmlXPathFreeObject (xpathobj_aux);

      gtk_list_store_append (list_store_peercast, &iter);
      gtk_list_store_set (list_store_peercast, &iter,
			  0,
			  entry->Genre,
			  1,
			  entry->Name,
			  2,
			  entry->Nowplaying,
			  3,
			  entry->Bitrate,
			  4,
			  entry->URI,
			  -1);

      {
	g_free (entry->Genre);
	g_free (entry->Name);
	g_free (entry->Nowplaying);
	g_free (entry->URI);
	g_free (entry);
      }

    } 

    xmlXPathFreeObject (xpathobj);

    list_store_peercast_filter = 
	gtk_tree_model_filter_new (GTK_TREE_MODEL (list_store_peercast), NULL); 

    gtk_tree_model_filter_set_visible_func (list_store_peercast_filter,
					    model_filter_func,
					    glade_xml_get_widget_warn (GLADE_XML(data), "peercast_filter"),
					    NULL);

    g_object_connect (G_OBJECT(glade_xml_get_widget_warn (GLADE_XML(data), "peercast_filter")),
		      "swapped-signal::changed",
		      G_CALLBACK (gtk_tree_model_filter_refilter),
		      list_store_peercast_filter,
		      NULL);

    gtk_tree_view_set_model (GTK_TREE_VIEW(glade_xml_get_widget_warn (GLADE_XML(data), "tree_view_peercast")), 
			     GTK_TREE_MODEL(list_store_peercast_filter));
}
#endif

static void
on_refresh_icecast_clicked (GtkWidget	  *widget,
			    gpointer	   data)
{
  xmlDocPtr	      doc;
  xmlXPathObjectPtr   xpathobj;
  xmlNodeSetPtr	      nv;
  xmlNodePtr	      node;

  GtkWidget          *refresh_shoutcast;
  GtkWidget          *notebook_icecast_inner;
  GtkTreeModelFilter *list_store_icecast_filter;
  GtkListStore	     *list_store_icecast;
  GtkTreeIter	      iter;

  VFSHandle	     *handle;
  gint		      n;

  typedef struct _IceCastEntry IceCastEntry;
  struct _IceCastEntry { 
      gchar   *Name;
      gchar   *Genre;
      gchar   *Nowplaying;

      gint     Bitrate;

      gchar   *URI;
  };

#define ICECAST_LIST_URI "http://dir.xiph.org/yp.xml"

#define IN_URI      "listen_url"
#define IN_NAME	    "server_name"
#define IN_GENRE    "genre"
#define IN_NP	    "current_song"
#define IN_BITRATE  "bitrate"
  
  handle = vfs_handle_new ();  
  handle->uri = g_strdup (ICECAST_LIST_URI);
  handle->need_container = FALSE;

  refresh_shoutcast = glade_xml_get_widget_warn (GLADE_XML(data),
							"refresh_shoutcast");
  notebook_icecast_inner = glade_xml_get_widget_warn (GLADE_XML(data),
						      "notebook_icecast_inner");
  gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_icecast_inner), 1);
  gtk_widget_set_sensitive (refresh_shoutcast, FALSE);
  bmp_window_set_busy (glade_xml_get_widget(GLADE_XML(data), "dialog_library"));
  while (gtk_events_pending ()) gtk_main_iteration();

  if (!vfs_handle_read (handle))
    {
      gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_icecast_inner), 0);
      gtk_widget_set_sensitive (refresh_shoutcast, FALSE);
      bmp_window_set_idle (glade_xml_get_widget(GLADE_XML(data), "dialog_library"));
      vfs_handle_free (handle);
      return;
    }

  doc = xmlParseMemory (handle->buffer, strlen(handle->buffer));
  if (!doc)
    {
      gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_icecast_inner), 0);
      gtk_widget_set_sensitive (refresh_shoutcast, FALSE);
      bmp_window_set_idle (glade_xml_get_widget(GLADE_XML(data), "dialog_library"));
      vfs_handle_free (handle);
      return;
    } 

  vfs_handle_free (handle);

  xpathobj = xml_execute_xpath_expression (doc, "//entry", NULL);
  nv = xpathobj->nodesetval;

  list_store_icecast = gtk_list_store_new (5,
					     G_TYPE_STRING,	// Name
					     G_TYPE_STRING,	// Genre
					     G_TYPE_STRING,	// Nowplaying
					     G_TYPE_INT,	// Bitrate
					     G_TYPE_STRING);	// URI

  for (n = 0; n < nv->nodeNr; n++) 
    {
      IceCastEntry  *entry;
      xmlNodePtr     node_sibling;

      node = 
	nv->nodeTab[n];
      node_sibling = 
	node->children;

      entry = g_new0 (IceCastEntry,1);

      while (node_sibling)
	{
	    if (node_sibling->type != XML_TEXT_NODE)
	      {
		const unsigned char *name = node_sibling->name;

		if (!strcmp (name, IN_NAME)) 
		  {
		    if (node_sibling->children)
		      {
			gchar *_data = XML_GET_CONTENT (node_sibling->children);
			entry->Name = _data; 
		      }
		    else
			entry->Name = g_strdup ("");
		  }
		else if (!strcmp (name, IN_URI)) 
		  {
		    if (node_sibling->children)
		      {
			gchar *_data = XML_GET_CONTENT (node_sibling->children);
			entry->URI = _data; 
		      }
		    else
			entry->Name = g_strdup ("");
		  }
		else if (!strcmp (name, IN_GENRE))
		  {
		    if (node_sibling->children)
		      {
			gchar *_data = XML_GET_CONTENT (node_sibling->children);
			entry->Genre = _data; 
		      }
		    else
			entry->Genre = g_strdup ("");
		  }
		else if (!strcmp (name, IN_NP))
		  {
		    if (node_sibling->children)
		      {
			gchar *_data = XML_GET_CONTENT (node_sibling->children);
			entry->Nowplaying = _data; 
		      }
		    else
			entry->Nowplaying = g_strdup ("");
		  }
		else if (!strcmp (name, IN_BITRATE))
		  {
		    gchar *_data =  XML_GET_CONTENT (node_sibling->children);
		    entry->Bitrate = strtol (_data, NULL, 10);
		    g_free (_data);
		  }
	      }
	    node_sibling = node_sibling->next;
	    while (gtk_events_pending()) gtk_main_iteration ();
	}

      gtk_list_store_append (list_store_icecast, &iter);
      gtk_list_store_set (list_store_icecast, &iter,
			  0,
			  entry->Genre,
			  1,
			  entry->Name,
			  2,
			  entry->Nowplaying,
			  3,
			  entry->Bitrate,
			  4,
			  entry->URI,
			  -1);

      {
	g_free (entry->Genre);
	g_free (entry->Name);
	g_free (entry->Nowplaying);
	g_free (entry->URI);
	g_free (entry);
      }

    } 

    xmlXPathFreeObject (xpathobj);

    list_store_icecast_filter = GTK_TREE_MODEL_FILTER
	(gtk_tree_model_filter_new (GTK_TREE_MODEL (list_store_icecast), NULL));

    gtk_tree_model_filter_set_visible_func (list_store_icecast_filter,
					    model_filter_func,
					    glade_xml_get_widget_warn (GLADE_XML(data), "icecast_filter"),
					    NULL);

    g_object_connect (G_OBJECT(glade_xml_get_widget_warn (GLADE_XML(data), "icecast_filter")),
		      "swapped-signal::changed",
		      G_CALLBACK (gtk_tree_model_filter_refilter),
		      list_store_icecast_filter,
		      NULL);

    gtk_tree_view_set_model (GTK_TREE_VIEW(glade_xml_get_widget_warn (GLADE_XML(data), "tree_view_icecast")), 
			     GTK_TREE_MODEL(list_store_icecast_filter));

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_icecast_inner), 0);
    gtk_widget_set_sensitive (refresh_shoutcast, TRUE);
    bmp_window_set_idle (glade_xml_get_widget(GLADE_XML(data), "dialog_library"));
    while (gtk_events_pending ()) gtk_main_iteration();

    gtk_tree_view_columns_autosize (GTK_TREE_VIEW(glade_xml_get_widget_warn (GLADE_XML(data), "tree_view_icecast"))); 
}

static void
on_refresh_shoutcast_clicked (GtkWidget	  *widget,
			      gpointer	   data)
{
  xmlDocPtr	      doc;
  xmlXPathObjectPtr   xpathobj;
  xmlNodeSetPtr	      nv;
  xmlNodePtr	      node;

  GtkWidget          *refresh_icecast;
  GtkWidget          *notebook_shoutcast_inner;
  GtkTreeModelFilter *list_store_shoutcast_filter;
  GtkListStore	     *list_store_shoutcast;
  GtkTreeIter	      iter;

  VFSHandle	     *handle;
  gint		      n;

  typedef struct _ShoutCastEntry ShoutCastEntry;
  struct _ShoutCastEntry { 
      gchar   *Name;
      gchar   *Genre;
      gchar   *Nowplaying;

      gint     Listeners;
      gint     Bitrate;

      gchar   *URI;
  };

#define SHOUTCAST_LIST_URI "http://www.shoutcast.com/sbin/xmllister.phtml?service=winamp2&limit=1500&no_compress=1"

#define P_URI	    "Playstring"

#define N_NAME	    "Name"
#define N_GENRE	    "Genre"
#define N_NP	    "Nowplaying"
#define N_LISTENERS "Listeners"
#define N_BITRATE   "Bitrate"
  
  handle = vfs_handle_new ();  
  handle->uri = g_strdup (SHOUTCAST_LIST_URI);
  handle->need_container = FALSE;
  
  refresh_icecast = glade_xml_get_widget_warn (GLADE_XML(data),
						      "refresh_icecast");
  notebook_shoutcast_inner = glade_xml_get_widget_warn (GLADE_XML(data),
							"notebook_shoutcast_inner");
  gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_shoutcast_inner), 1);
  gtk_widget_set_sensitive (refresh_icecast, FALSE);
  bmp_window_set_busy (glade_xml_get_widget(GLADE_XML(data), "dialog_library"));
  while (gtk_events_pending ()) gtk_main_iteration();

  if (!vfs_handle_read (handle))
    {
      gtk_widget_set_sensitive (refresh_icecast, TRUE);
      bmp_window_set_idle (glade_xml_get_widget(GLADE_XML(data), "dialog_library"));
      gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_shoutcast_inner), 0);
      vfs_handle_free (handle);
      return;
    }

  doc = xmlParseMemory (handle->buffer, strlen(handle->buffer));
  if (!doc)
    {
      gtk_widget_set_sensitive (refresh_icecast, TRUE);
      bmp_window_set_idle (glade_xml_get_widget(GLADE_XML(data), "dialog_library"));
      gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_shoutcast_inner), 0);
      vfs_handle_free (handle);
      return;
    }

  vfs_handle_free (handle);

  xpathobj = xml_execute_xpath_expression (doc, "//entry", NULL);
  nv = xpathobj->nodesetval;

  list_store_shoutcast = gtk_list_store_new (6,
					     G_TYPE_STRING,	// Name
					     G_TYPE_STRING,	// Genre
					     G_TYPE_STRING,	// Nowplaying
					     G_TYPE_INT,	// Listeners
					     G_TYPE_INT,	// Bitrate
					     G_TYPE_STRING);	// URI

  for (n = 0; n < nv->nodeNr; n++) 
    {
      ShoutCastEntry  *entry;
      xmlNodePtr       node_sibling;

      node = 
	nv->nodeTab[n];
      node_sibling = 
	node->children;

      entry = g_new0 (ShoutCastEntry,1);
      entry->URI = xmlGetProp (node, P_URI); 

      while (node_sibling)
	{
	    if (node_sibling->type != XML_TEXT_NODE)
	      {
		const unsigned char *name = node_sibling->name;

		if (!strcmp (name, N_NAME)) 
		  {
		    if (node_sibling->children)
		      {
			gchar *_data = XML_GET_CONTENT (node_sibling->children);
			entry->Name = _data; 
		      }
		    else
			entry->Name = g_strdup ("");
		  }
		else if (!strcmp (name, N_GENRE))
		  {
		    if (node_sibling->children)
		      {
			gchar *_data = XML_GET_CONTENT (node_sibling->children);
			entry->Genre = _data; 
		      }
		    else
			entry->Genre = g_strdup ("");
		  }
		else if (!strcmp (name, N_NP))
		  {
		    if (node_sibling->children)
		      {
			gchar *_data = XML_GET_CONTENT (node_sibling->children);
			entry->Nowplaying = _data; 
		      }
		    else
			entry->Nowplaying = g_strdup ("");
		  }
		else if (!strcmp (name, N_LISTENERS))
		  {
		    gchar *_data =  XML_GET_CONTENT (node_sibling->children);
		    entry->Listeners = strtol (_data, NULL, 10);
		    g_free (_data);
		  }
		else if (!strcmp (name, N_BITRATE))
		  {
		    gchar *_data =  XML_GET_CONTENT (node_sibling->children);
		    entry->Bitrate = strtol (_data, NULL, 10);
		    g_free (_data);
		  }
	      }
	    node_sibling = node_sibling->next;
	    while (gtk_events_pending()) gtk_main_iteration ();
	}

      gtk_list_store_append (list_store_shoutcast, &iter);
      gtk_list_store_set (list_store_shoutcast, &iter,
			  0,
			  entry->Genre,
			  1,
			  entry->Name,
			  2,
			  entry->Nowplaying,
			  3,
			  entry->Listeners,
			  4,
			  entry->Bitrate,
			  5,
			  entry->URI,
			  -1);

      {
	g_free (entry->Genre);
	g_free (entry->Name);
	g_free (entry->Nowplaying);
	g_free (entry->URI);
	g_free (entry);
      }

    } 

    xmlXPathFreeObject (xpathobj);

    list_store_shoutcast_filter = GTK_TREE_MODEL_FILTER
	(gtk_tree_model_filter_new (GTK_TREE_MODEL (list_store_shoutcast), NULL));

    gtk_tree_model_filter_set_visible_func (list_store_shoutcast_filter,
					    model_filter_func,
					    glade_xml_get_widget_warn (GLADE_XML(data), "shoutcast_filter"),
					    NULL);

    g_object_connect (G_OBJECT(glade_xml_get_widget_warn (GLADE_XML(data), "shoutcast_filter")),
		      "swapped-signal::changed",
		      G_CALLBACK (gtk_tree_model_filter_refilter),
		      list_store_shoutcast_filter,
		      NULL);

    gtk_tree_view_set_model (GTK_TREE_VIEW(glade_xml_get_widget_warn (GLADE_XML(data), "tree_view_shoutcast")), 
			     GTK_TREE_MODEL(list_store_shoutcast_filter));

    gtk_notebook_set_current_page (GTK_NOTEBOOK(notebook_shoutcast_inner), 0);
    gtk_widget_set_sensitive (refresh_icecast, TRUE);
    bmp_window_set_idle (glade_xml_get_widget(GLADE_XML(data), "dialog_library"));
    while (gtk_events_pending ()) gtk_main_iteration();

    gtk_tree_view_columns_autosize (GTK_TREE_VIEW(glade_xml_get_widget_warn (GLADE_XML(data), "tree_view_icecast"))); 
}

static gboolean
on_dialog_library_delete (GtkWidget *widget,
			  GdkEvent *event,
                          gpointer user_data)
{

  gtk_widget_hide (widget);

  return TRUE;
}

void
invoke_dialog_library (void)
{
    GtkWidget	     *widget;
    static GladeXML  *xml;
    const gchar	     *glade_file = DATA_DIR "/glade/dialog_library.glade";
    gint	      n;

    if (! dialog_window)
    {
	xml = glade_xml_new_or_die(Q_("Dialogs"), glade_file, NULL, NULL);
	glade_xml_signal_autoconnect (xml);

	dialog_window = glade_xml_get_widget_warn (xml, "dialog_library");
	gtk_widget_realize (dialog_window);
	g_object_connect (G_OBJECT(dialog_window), 
			  "signal::delete-event", 
			  G_CALLBACK(on_dialog_library_delete),
			  NULL,
			  NULL);

	bmp_window_set_icon_list (GTK_WIDGET(dialog_window), "library");

	gtk_window_set_position (GTK_WINDOW(dialog_window), GTK_WIN_POS_CENTER);
	g_object_connect (G_OBJECT(dialog_window), 
			  "signal::key-press-event", 
			  G_CALLBACK(on_library_dialog_key_press), 
			  NULL, 
			  NULL);
    
	/* Generic stuff */
   	widget = glade_xml_get_widget_warn (xml, "image_library");
	gtk_image_set_from_file (GTK_IMAGE (widget), DATA_DIR "/images/library-headerlogo.png");

	widget = glade_xml_get_widget_warn (xml, "image_wait1");
	gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");
	widget = glade_xml_get_widget_warn (xml, "image_wait2");
	gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");
	widget = glade_xml_get_widget_warn (xml, "image_wait3");
	gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/wait.gif");


	widget = glade_xml_get_widget_warn (xml, "library_add_path");	
	g_object_connect (G_OBJECT(widget), 
			  "signal::clicked", 
			  G_CALLBACK(on_library_add_path_clicked), 
			  xml, 
			  NULL);

	/* Setup SEARCH Tab */
   	widget = glade_xml_get_widget_warn (xml, "cbox_tag_id");
	gtk_combo_box_remove_text (GTK_COMBO_BOX(widget), 0);
	for (n = 0; n < BMP_DATUM_NONE; 
	    gtk_combo_box_append_text (GTK_COMBO_BOX(widget), bmp_metadata_get_title(n++)));

	widget = glade_xml_get_widget_warn (xml, "tree_view_search");
	g_object_connect(G_OBJECT(widget), 
			 "signal::row-activated", 
			 G_CALLBACK(on_tree_view_search_row_activated), 
			 xml, 
			 NULL);

	gtk_tree_selection_set_mode
	    (gtk_tree_view_get_selection (GTK_TREE_VIEW(widget)),
	     GTK_SELECTION_MULTIPLE);

	for (n = 0; n < 4; n++)
	  {

	    library_column_search[n] =
		gtk_tree_view_column_new ();

	    library_renderer_search[n] =
		gtk_cell_renderer_text_new ();

	    gtk_tree_view_column_pack_start
		(library_column_search[n],
		 library_renderer_search[n],
		 FALSE);

	    gtk_tree_view_column_set_attributes
		(library_column_search[n],
		 library_renderer_search[n],
		 "text",
		  n,
		  NULL);

	    gtk_tree_view_column_set_sort_column_id (library_column_search[n], n);

	    gtk_tree_view_column_set_expand
		(library_column_search[n],
		 FALSE);

	    gtk_tree_view_column_set_title
		(library_column_search[n],
		 bmp_metadata_get_title (search_store_ids[n]));

	    gtk_tree_view_column_set_sizing
		(library_column_search[n],
		 GTK_TREE_VIEW_COLUMN_GROW_ONLY);

	    gtk_tree_view_column_set_resizable
		(library_column_search[n],
		 TRUE);

	    gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (library_column_search[n]));

	  }

	widget = glade_xml_get_widget_warn (xml, "entry_value");
	g_object_connect(G_OBJECT(widget), "signal::changed", G_CALLBACK(on_entry_value_changed), xml, NULL);

   	widget = glade_xml_get_widget_warn (xml, "cbox_tag_id");
	gtk_combo_box_set_active (GTK_COMBO_BOX(widget), BMP_DATUM_ARTIST);

	widget = glade_xml_get_widget_warn (xml, "cbox_tag_id");
	g_object_connect(G_OBJECT(widget), "signal::changed", G_CALLBACK(on_cbox_tag_id_changed), xml, NULL);

	widget = glade_xml_get_widget_warn (xml, "cb_exact");
	g_object_connect(G_OBJECT(widget), "signal::toggled", G_CALLBACK(on_cb_exact_changed), xml, NULL);

	/* Setup BROWSE Tab */
	widget = glade_xml_get_widget_warn (xml, "tree_view_browse");
	g_object_connect(G_OBJECT(widget), 
			 "swapped-signal::row-activated", 
			 G_CALLBACK(gtk_window_activate_default), 
			 GTK_WINDOW(dialog_window), 
			 NULL);

	gtk_tree_selection_set_mode
	    (gtk_tree_view_get_selection (GTK_TREE_VIEW(widget)),
	     GTK_SELECTION_MULTIPLE);

	/* Two columns, one for artist/album/title and a 2nd seperate column for track # */
	library_column_browse[0] =
		gtk_tree_view_column_new ();

	/* The cell renderer for the artist/album/title icon */
	library_renderer_browse[0] =
		gtk_cell_renderer_pixbuf_new ();

	g_object_set (library_renderer_browse[0], "width", 16, NULL);
	g_object_set (library_renderer_browse[0], "height", 16, NULL);

	gtk_tree_view_column_pack_start
		(library_column_browse[0],
		 library_renderer_browse[0],
		 FALSE);

	gtk_tree_view_column_set_attributes
		(library_column_browse[0],
		 library_renderer_browse[0],
		 "pixbuf",
		  0,
		  NULL);

	/* The track number renderer */
	library_renderer_browse[1] =
		gtk_cell_renderer_text_new ();

	g_object_set (G_OBJECT(library_renderer_browse[1]), 
		      "xalign", 1.0,
		      "xpad", 2,
		      NULL);

	gtk_tree_view_column_pack_start
		(library_column_browse[0],
		 library_renderer_browse[1],
		 FALSE);

	gtk_tree_view_column_set_cell_data_func
		(library_column_browse[0],
		 library_renderer_browse[1],
		 (GtkTreeCellDataFunc) tracknumber_renderer_func,
                 NULL, 
                 NULL); 

	/* The cell renderer for the actual text */
	library_renderer_browse[2] =
		gtk_cell_renderer_text_new ();

	g_object_set (G_OBJECT(library_renderer_browse[2]), 
		      "xpad", 12,
		      NULL);

	gtk_tree_view_column_pack_start
		(library_column_browse[0],
		 library_renderer_browse[2],
		 FALSE);

	gtk_tree_view_column_set_attributes
		(library_column_browse[0],
		 library_renderer_browse[2],
		 "text",
		  1,
		  NULL);

	gtk_tree_view_column_set_expand
		(library_column_browse[0],
		 FALSE);

	gtk_tree_view_append_column
		(GTK_TREE_VIEW (widget),
		 GTK_TREE_VIEW_COLUMN (library_column_browse[0]));

	gtk_tree_view_set_model (GTK_TREE_VIEW (widget), GTK_TREE_MODEL(tree_store_browse));
	gtk_widget_show_all (GTK_WIDGET (widget));

	/* Setup SHOUTCAST Tab */
	widget = glade_xml_get_widget_warn (xml, "image_shoutcast");
	gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/shoutcast.png");

	widget = glade_xml_get_widget_warn (xml, "tree_view_shoutcast");
	g_object_connect(G_OBJECT(widget), 
			 "swapped-signal::row-activated", 
			  G_CALLBACK(gtk_window_activate_default), 
			  GTK_WINDOW(dialog_window), 
			  NULL);

	for (n = 0; n < G_N_ELEMENTS (shoutcast_headers); n++)
	  {
	    GtkTreeViewColumn *column;
	    GtkCellRenderer   *cell;
	    SortDataTuple     *sort_data_tuple;

	    sort_data_tuple = g_new0 (SortDataTuple,1);

	    sort_data_tuple->tree_view = GTK_TREE_VIEW (widget);
	    sort_data_tuple->column_id = n;

	    column = 
	      gtk_tree_view_column_new ();
	    cell =
	      gtk_cell_renderer_text_new ();

	    gtk_tree_view_column_set_resizable (column, TRUE);
	    gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_GROW_ONLY);

	    gtk_tree_view_column_pack_start
		(column,
		 cell,
		 TRUE);  

	    gtk_tree_view_column_set_attributes
		(column,
		 cell, 
		 "text",
		  n,
		  NULL);

	    gtk_tree_view_column_set_clickable (column, TRUE);
	    g_object_connect (G_OBJECT(column),
			      "signal::clicked",
			      G_CALLBACK (column_clicked),
			      sort_data_tuple, 
			      NULL);

	    gtk_tree_view_column_set_title
		(column,
		 shoutcast_headers[n]);

	    gtk_tree_view_append_column (GTK_TREE_VIEW(widget), column);
	  }

	g_object_connect(G_OBJECT(glade_xml_get_widget_warn (xml, "refresh_shoutcast")),
			 "signal::clicked", 
			 G_CALLBACK(on_refresh_shoutcast_clicked), 
			 xml, 
		         NULL); 


	/* Setup ICECAST Tab */
	widget = glade_xml_get_widget_warn (xml, "image_icecast");
	gtk_image_set_from_file (GTK_IMAGE(widget), DATA_DIR "/images/xiph.png");

	widget = glade_xml_get_widget_warn (xml, "tree_view_icecast");
	g_object_connect(G_OBJECT(widget), 
			 "swapped-signal::row-activated", 
			  G_CALLBACK(gtk_window_activate_default), 
			  GTK_WINDOW(dialog_window), 
			  NULL);

	for (n = 0; n < G_N_ELEMENTS (icecast_headers); n++)
	  {
	    GtkTreeViewColumn *column;
	    GtkCellRenderer   *cell;
	    SortDataTuple     *sort_data_tuple;

	    sort_data_tuple = g_new0 (SortDataTuple,1);

	    sort_data_tuple->tree_view = GTK_TREE_VIEW (widget);
	    sort_data_tuple->column_id = n;

	    column = 
	      gtk_tree_view_column_new ();
	    cell =
	      gtk_cell_renderer_text_new ();

	    gtk_tree_view_column_set_resizable (column, TRUE);
	    gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_GROW_ONLY);

	    gtk_tree_view_column_pack_start
		(column,
		 cell,
		 TRUE);  

	    gtk_tree_view_column_set_attributes
		(column,
		 cell, 
		 "text",
		  n,
		  NULL);

	    gtk_tree_view_column_set_clickable (column, TRUE);
	    g_object_connect (G_OBJECT(column),
			      "signal::clicked",
			      G_CALLBACK (column_clicked),
			      sort_data_tuple, 
			      NULL);

	    gtk_tree_view_column_set_title
		(column,
		 icecast_headers[n]);

	    gtk_tree_view_append_column (GTK_TREE_VIEW(widget), column);
	  }

	g_object_connect(G_OBJECT(glade_xml_get_widget_warn (xml, "refresh_icecast")),
		         "signal::clicked", 
			 G_CALLBACK(on_refresh_icecast_clicked), 
			 xml, 
		         NULL); 

#if 0
	/* Setup PEERCAST Tab */
	widget = glade_xml_get_widget_warn (xml, "tree_view_peercast");
	g_object_connect(G_OBJECT(widget), 
			 "swapped-signal::row-activated", 
			  G_CALLBACK(gtk_window_activate_default), 
			  GTK_WINDOW(dialog_window), 
			  NULL);

	for (n = 0; n < G_N_ELEMENTS (peercast_headers); n++)
	  {
	    GtkTreeViewColumn *column;
	    GtkCellRenderer   *cell;
	    SortDataTuple     *sort_data_tuple;

	    sort_data_tuple = g_new0 (SortDataTuple,1);

	    sort_data_tuple->tree_view = widget;
	    sort_data_tuple->column_id = n;

	    column = 
	      gtk_tree_view_column_new ();
	    cell =
	      gtk_cell_renderer_text_new ();

	    gtk_tree_view_column_set_resizable (column, TRUE);
	    gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);

	    gtk_tree_view_column_pack_start
		(column,
		 cell,
		 TRUE);  

	    gtk_tree_view_column_set_attributes
		(column,
		 cell, 
		 "text",
		  n,
		  NULL);

	    gtk_tree_view_column_set_clickable (column, TRUE);
	    g_object_connect (G_OBJECT(column),
			      "signal::clicked",
			      G_CALLBACK (column_clicked),
			      sort_data_tuple, 
			      NULL);

	    gtk_tree_view_column_set_title
		(column,
		 icecast_headers[n]);

	    gtk_tree_view_append_column (GTK_TREE_VIEW(widget), column);
	  }

	g_object_connect(G_OBJECT(glade_xml_get_widget_warn (xml, "refresh_peercast")),
		         "signal::clicked", 
			 G_CALLBACK(on_refresh_peercast_clicked), 
			 xml, 
		         NULL); 
#endif

	/* Common widgets */
	widget = glade_xml_get_widget_warn (xml, "button_enqueue");
	g_object_connect(G_OBJECT(widget), "signal::clicked", G_CALLBACK(on_button_enqueue_clicked), xml, NULL);
	widget = glade_xml_get_widget_warn (xml, "button_play");
	g_object_connect(G_OBJECT(widget), "signal::clicked", G_CALLBACK(on_button_play_clicked), xml, NULL);
    }

    gtk_window_present (GTK_WINDOW(dialog_window));
    gdk_window_raise (dialog_window->window);
}
