/*  BMPx - Cross-platform multimedia player
 *  Copyright (C) 2003-2004  Edward Brocklesby, Chong Kai Xiong,
 *            Milosz Derezynski, David Lau, Michiel Sikkes
 *  Copyright (C) 2005 Chong Kai Xiong, Milosz Derezynski, Kenneth Ostby
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <X11/Xlib.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

#include "main.h"
#include "error.h"

#include "ui.h"
#include "ui_main.h"
#include "ui_prefswin.h"
#include "ui_util.h"
#include "ui_callbacks.h"
#include "ui_dialog_about.h"

/* bmp_dock.c/bmp_dock.h (C) 1998-2003 XMMS.org, (C) 2004-2005 BMPx DT */
#include "dock.h"

#include <libskinned/bmp_button.h>
#include <libskinned/bmp_button_toggle.h>
#include <libskinned/bmp_slider.h>
#include <libskinned/bmp_slider.h>
#include <libskinned/bmp_window.h>
#include <libskinned/bmp_tooltips.h>

#include "bmpx/util.h"

#include "goa/libgoa.h"

G_DEFINE_TYPE (BmpWindowMain, bmp_window_main, G_TYPE_OBJECT)

static void
window_main_create(BmpWindowMain *self);

static void
mw_adapt_cursor (BmpWindow *window, gint x, gint y);

static void
toggle_textbox_scroll (GtkToggleAction *action, BmpWindowMain *self);

static void
update_textbox (BmpWindowMain *self, const gchar *text);

static GtkActionEntry bmp_actions_MAIN[] = {
  { "dummy", NULL, "dummy" },
  { "dummy-view-menu", NULL, N_("View") },

  { BMP_ACTION_PREV,
    GTK_STOCK_MEDIA_PREVIOUS,
    N_("Previous"), "z",
    N_("Previous"),
    G_CALLBACK(bmp_ui_callback_prev)},

  { BMP_ACTION_EJECT,
    GTK_STOCK_OPEN,
    N_("Open files"), NULL,
    N_("Open Files"),
    G_CALLBACK(bmp_ui_callback_open_files)},

  { BMP_ACTION_PLAY,
    GTK_STOCK_MEDIA_PLAY,
    N_("Play"), "x",
    N_("Play"),
    G_CALLBACK(bmp_ui_callback_play)},

  { BMP_ACTION_PAUSE,
    GTK_STOCK_MEDIA_PAUSE,
    N_("Pause"), "c",
    N_("Pause"),
    G_CALLBACK(bmp_ui_callback_pause)},

  { BMP_ACTION_STOP,
    GTK_STOCK_MEDIA_STOP,
    N_("Stop"), "v",
    N_("Stop"),
    G_CALLBACK(bmp_ui_callback_stop)},

  { BMP_ACTION_NEXT,
    GTK_STOCK_MEDIA_NEXT,
    N_("Next"), "b",
    N_("Next"),
    G_CALLBACK(bmp_ui_callback_next)},

  { BMP_ACTION_ICONIFY,
    NULL,
    N_("Minimize BMPx"), NULL,
    N_("Minimize"),
    G_CALLBACK(bmp_ui_callback_iconify)},

  { BMP_ACTION_QUIT,
    GTK_STOCK_QUIT,
    N_("Quit BMPx"), "<control>w",
    N_("Quit BMPx"),
    G_CALLBACK(bmp_ui_callback_quit)},

  { BMP_ACTION_ABOUT,
    GTK_STOCK_ABOUT,
    N_("About BMPx"), NULL,
    N_("About BMPx"),
    G_CALLBACK(bmp_ui_callback_about)},

  { BMP_ACTION_PREFERENCES,
    GTK_STOCK_PREFERENCES,
    N_("Preferences"), "P",
    N_("Preferences"),
    G_CALLBACK(bmp_ui_callback_preferences)},

  { BMP_ACTION_LIBRARY,
    GTK_STOCK_FIND,
    N_("Music Library"), "L",
    N_("Music Library"),
    G_CALLBACK(bmp_ui_callback_library)}

};

static GtkToggleActionEntry bmp_toggle_actions_MAIN[] = {

  { BMP_TOGGLE_ACTION_STOP_AFTER_CURRENT,
	NULL,
	N_("Stop After Current Track"),
	NULL,
	NULL,	
	G_CALLBACK (bmp_ui_callback_stop_after_current_track),
	FALSE },

  { BMP_TOGGLE_ACTION_SHUFFLE,
	NULL,
	N_("_Shuffle"),
	NULL,
	N_("Switch between random and continuous playback"),
	G_CALLBACK (bmp_ui_callback_toggle_shuffle),
	FALSE },

  { BMP_TOGGLE_ACTION_REPEAT,
	NULL,
	N_("_Repeat"),
	NULL,
	N_("Whether to repeat the playlist or not"),
	G_CALLBACK (bmp_ui_callback_toggle_repeat),
	FALSE },

  { BMP_TOGGLE_ACTION_PL,
	NULL,
	N_("Show Playlist Editor"),
	NULL,
	N_("Toggles playlist visibility"),
	G_CALLBACK (bmp_ui_callback_toggle_pl),
	FALSE },

  { BMP_TOGGLE_ACTION_SCROLL,
	NULL,
	N_("Scroll Textbox"),
	NULL,
	N_("Toggles textbox scroll"),
	G_CALLBACK (toggle_textbox_scroll),
	TRUE }
};

static bmp_button_toggle_t bmp_window_main_buttons_toggle[] = {
    /* Shuffle */
    { BMP_TOGGLE_ACTION_SHUFFLE,
	N_("Toggle Shuffle Mode"),
	SKIN_COMPONENT_SHUFREP,
	28, 0,
	28, 15,
	28, 30,
	46, 15,
	164,89,
	TRUE },

    /* Repeat */
    { BMP_TOGGLE_ACTION_REPEAT,
	N_("Toggle Playlist Repeat"),
	SKIN_COMPONENT_SHUFREP,
	0,  0,
	0,  15,
	0,  30,
	28, 15,
	210,89,
	TRUE },

    /* EQ */
    { NULL,
	N_("Toggle Equalizer window visibility"),
	SKIN_COMPONENT_SHUFREP,
	0,  61,
	46, 73,
	0,  73,
	23, 12,
	219,58,
	TRUE },

    /* PL */
    {BMP_TOGGLE_ACTION_PL,
	N_("Toggle Playlist window visibility"),
	SKIN_COMPONENT_SHUFREP,
	23, 61,
	69, 73,
	23, 73,
	23, 12,
	242,58,
	TRUE }
};
static gint nbuttons_window_main_toggle = G_N_ELEMENTS(bmp_window_main_buttons_toggle);

enum {
    MAIN_BUTTON_EJECT,
    MAIN_BUTTON_PREV,
    MAIN_BUTTON_PLAY,
    MAIN_BUTTON_PAUSE,
    MAIN_BUTTON_STOP,
    MAIN_BUTTON_NEXT,
    MAIN_BUTTON_ICONIFY,
    MAIN_BUTTON_QUIT
};

static bmp_button_normal_t bmp_window_main_buttons[] =
{
    /* Eject button */
    { BMP_ACTION_EJECT,
	N_("Open Files"),
	SKIN_COMPONENT_CBUTTONS,
	114, 0,
	114, 16,
	22,  16,
	136, 89 },

    /* Prev button */
    { BMP_ACTION_PREV,
	N_("Previous Track"),
	SKIN_COMPONENT_CBUTTONS,
	0,  0,
	0,  18,
	23, 18,
	16, 88 },

    /* Play button */
    { BMP_ACTION_PLAY,
	N_("Play Current Track"),
	SKIN_COMPONENT_CBUTTONS,
	22, 0,
	22, 18,
	23, 18,
	38, 88 },

    /* Pause button */
    { BMP_ACTION_PAUSE,
	N_("Toggle Pause"),
	SKIN_COMPONENT_CBUTTONS,
	45, 0,
	45, 18,
	23, 18,
	61, 88 },

    /* Stop button */
    { BMP_ACTION_STOP,
	N_("Stop Playback"),
	SKIN_COMPONENT_CBUTTONS,
	68, 0,
	68, 18,
	23, 18,
	84, 88 },

    /* Next button */
    { BMP_ACTION_NEXT,
	N_("Next Track"),
	SKIN_COMPONENT_CBUTTONS,
	91,  0,
	91,  18,
	23,  18,
	107, 88 },

    /* Minimize button */
    { BMP_ACTION_ICONIFY,
	N_("Minimize BMPx"),
	SKIN_COMPONENT_TITLEBAR,
	9,   0,
	9,   9,
	8,   8,
	244, 3 },

    /* Close button */
    { BMP_ACTION_QUIT,
	N_("Quit BMPx"),
	SKIN_COMPONENT_TITLEBAR,
	18,  0,
	18,  9,
	8,   8,
	264, 3 }
};
gint nbuttons_window_main = G_N_ELEMENTS(bmp_window_main_buttons);

typedef struct _region_procedure_info_t region_procedure_info_t;
struct _region_procedure_info_t {
    gint x;
    gint y;
    gint w;
    gint h;
    void (*on_region_callback) (BmpWindowMain *self, GdkEvent *event, gpointer user_data);
};

/* FIXME: Move this inside Private */
static GPtrArray *array_buttons;
static GPtrArray *array_buttons_toggle;
static GdkPixbuf *numbers[12];

typedef struct {
	gint x;
	gint y;
} number_position_t;

static number_position_t number_positions[] = {
	{ 48, 26 },
	{ 60, 26 },
	{ 78, 26 },
	{ 90, 26 },
	{ 36, 26 }
};

#define KBPS_X 108
#define KBPS_Y 43

#define TEXTBOX_X 109
#define TEXTBOX_Y 23
#define TEXTBOX_FONT_SIZE 11 
#define TEXTBOX_WIDTH 154

#define SLIDER_SEEK_X 16
#define SLIDER_SEEK_Y 72

#define SLIDER_VOLUME_X 107
#define SLIDER_VOLUME_Y 57

#define SLIDER_BALANCE_X 177
#define SLIDER_BALANCE_Y 57

#define PLAYSTATUS_X 24
#define PLAYSTATUS_Y 28

G_LOCK_DEFINE(TEXTBOX_LOCK);

typedef enum {
	SCROLL_RIGHT,
	SCROLL_LEFT
} TextboxScrollDirection;

enum {
	NUMBER_MINUTES_10,
	NUMBER_MINUTES_1,
	NUMBER_SECONDS_10,
	NUMBER_SECONDS_1,
	NUMBER_MINUS
};

struct _BmpWindowMainPrivate {

    gboolean dispose_has_run;

    gboolean seeking;
    gint seek_request;
    gint last_position;
    gint time_minutes;
    gint time_seconds;
    gboolean time_inverse;

    /* Sliders */
    GtkWidget *sliders[BMP_UI_MAIN_N_SLIDERS];

    /* Textbox */
    GdkPixmap *textbox;
    TextboxScrollDirection tb_direction;
    gboolean tb_running;
    gboolean tb_dragging;
    gint tb_offset;
    gint tb_drag_offset;
    gint tb_drag_event_x;
};

void
reset_last_position (BmpWindowMain *self)
{
	g_return_if_fail (BMP_IS_WINDOW_MAIN(self));

	self->private->last_position = -1;
	self->private->seek_request = -1;
}

void
bmp_window_main_set_seek_request (BmpWindowMain *self, gint seek_request)
{
	g_return_if_fail (BMP_IS_WINDOW_MAIN(self));

	self->private->seek_request = seek_request;
}

static void
toggle_textbox_scroll (GtkToggleAction *action, BmpWindowMain *self)
{
    gboolean scroll;

    scroll = gtk_toggle_action_get_active (action);

    if (! scroll)
    {
	self->private->tb_running = FALSE;
	self->private->tb_offset = 0;
    }
    else
    {
	gchar *text = g_object_get_string (bmp_system_control, "current_title"); 
	update_textbox (self, text);
	g_free (text);
	g_usleep (100000);
	gtk_widget_queue_draw (self->window);
    }
}

static void
on_system_control_set_volume (BmpSystemControl *control, gint volume, BmpWindowMain *self)
{
    bmp_slider_set_position (BMP_SLIDER(self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]), volume);
}

static void
on_system_control_stream_pos (BmpSystemControl *control, gint stime, BmpWindowMain *self)
{
	gdouble position, percent;
	gint length;

	length = bmp_play_get_length (bmp_play);

	percent = length/100.;
	position = stime / percent;

	if (!self->private->seeking) {
		if (self->private->seek_request != -1) {
			if ( !(position < self->private->seek_request)) {
				self->private->seek_request = -1;
			}
		} else if (position > self->private->last_position) {
				bmp_slider_set_position (BMP_SLIDER(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), position);
				self->private->last_position = position;
		}
	}

	if (self->private->time_inverse)
	  {
	    stime = (length - stime) * -1;
	  }

	self->private->time_minutes = stime / 60;
	self->private->time_seconds = stime % 60;

	gtk_widget_queue_draw_area (self->window,36,26,99,39);
}

/* Functions */
static gboolean
seek_started (GtkWidget *widget, GdkEventButton *event, BmpWindowMain *self)
{
	g_return_val_if_fail (GTK_IS_WIDGET(widget), FALSE);

	self->private->seeking = TRUE;

	return FALSE;
}

static gboolean
seek_ended (GtkWidget *widget, GdkEventButton *event, BmpWindowMain *self)
{
	g_return_val_if_fail (GTK_IS_WIDGET(widget), FALSE);

	self->private->seeking = FALSE;

	return FALSE;
}

static gboolean
on_window_main_state_event (GtkWidget *widget, GdkEventWindowState *event, gpointer data)
{
    BmpWindowPlaylist *window_playlist;

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    if (event->changed_mask & GDK_WINDOW_STATE_ICONIFIED) 
      {
	   if (!(event->new_window_state & GDK_WINDOW_STATE_ICONIFIED) && 
		xcs_key_fetch_boolean (xcs, "playlist-window", "visible"))
	     {

		window_playlist = RM_GET_ITEM("windows", "window-playlist", NOLOCK);
#if 0
		gtk_window_deiconify (GTK_WINDOW(window_playlist->window));
	    	gtk_window_present (GTK_WINDOW(window_playlist->window));
#endif
	    	gdk_window_show (GTK_WIDGET(window_playlist->window)->window);
	    	gdk_window_raise (GTK_WIDGET(window_playlist->window)->window);

		return FALSE;
	     }

	   if (event->new_window_state & GDK_WINDOW_STATE_ICONIFIED)
             {
		window_playlist = RM_GET_ITEM("windows", "window-playlist", NOLOCK);
		gtk_window_iconify (GTK_WINDOW(window_playlist->window));
		return FALSE;
	     }

    }

    return TRUE;
}

static gboolean
on_window_main_focus_in (GtkWidget *widget, GdkEventFocus *event, gpointer data)
{
    BmpWindowPlaylist    *window_playlist;
    BmpWindowMain        *window_main;
    BmpWindowPreferences *window_preferences;

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);
    g_return_val_if_fail (GTK_IS_WIDGET (data), FALSE);

    gtk_widget_show (GTK_WIDGET(g_ptr_array_index (array_buttons, MAIN_BUTTON_ICONIFY)));
    gtk_widget_show (GTK_WIDGET(g_ptr_array_index (array_buttons, MAIN_BUTTON_QUIT)));

    gtk_widget_queue_draw_area (GTK_WIDGET(data), 0, 0, 275, 14);

    window_playlist = RM_GET_ITEM("windows", "window-playlist", NOLOCK);
    window_main = RM_GET_ITEM("windows", "window-main", NOLOCK);
    window_preferences = RM_GET_ITEM("windows", "window-preferences", NOLOCK);

    if (!bmp_preferences_window_raise(window_preferences))
	{
	    gdk_window_raise (window_playlist->window->window);
	    gdk_window_raise (window_main->window->window);
	    bmp_about_dialog_raise();
	}

    gtk_widget_queue_draw (window_playlist->window);

    return FALSE;
}

static gboolean
on_window_main_focus_out (GtkWidget *widget, GdkEventFocus *event, gpointer data)
{
	g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);
	g_return_val_if_fail (GTK_IS_WIDGET (data), FALSE);

	gtk_widget_hide (GTK_WIDGET(g_ptr_array_index (array_buttons, MAIN_BUTTON_ICONIFY)));
	gtk_widget_hide (GTK_WIDGET(g_ptr_array_index (array_buttons, MAIN_BUTTON_QUIT)));

	gtk_widget_queue_draw_area (GTK_WIDGET(data), 0, 0, 275, 14);

	return FALSE;
}

static gboolean
on_window_main_button_release (GtkWidget *widget,
		    GdkEventButton *event,
		    BmpWindowMain *self)
{
    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    gdk_pointer_ungrab(GDK_CURRENT_TIME);
    gdk_flush();
    mw_adapt_cursor (BMP_WINDOW(self->window), (gint)event->x, (gint)event->y);
    self->private->tb_dragging = FALSE;

    if (dock_is_moving(GTK_WINDOW(widget)))
    {
	dock_move_release(GTK_WINDOW(widget));
	bmp_system_control_move_end (bmp_system_control);
    }

    return FALSE;
}

static gboolean
on_window_main_scroll_event (GtkWidget *widget,
		    GdkEventScroll *event,
		    BmpWindowMain *self)
{
	gpointer instance;
	gboolean rv;

	g_return_val_if_fail (GTK_IS_WIDGET(widget), FALSE);

	instance = self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME];
	g_signal_emit_by_name (instance, "scroll-event", event, &rv);

	return TRUE;
}

static gboolean
on_window_main_key_press (GtkWidget *widget,
		    GdkEventKey *event,
		    BmpWindowMain *self)
{

	g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

	if (event->keyval == GDK_Escape)
		bmp_window_main_iconify (self);

	return FALSE;
}

static gboolean
on_window_main_stop_button_press (GtkWidget *widget,
		    GdkEventButton *event,
		    BmpWindowMain *self) {

    GtkWidget *menu;

    if (event->button == 3) {

	menu = ui_manager_get_popup (bmp_ui_manager, "/popup/menu-button-stop");
	gtk_widget_realize (menu);
	SET_CURSOR_FOR_WIDGET(CURSOR_MAINMENU,menu);
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);
	return FALSE;

    }

    return FALSE;
}

static gboolean
on_window_main_button_press (GtkWidget *widget,
		    GdkEventButton *event,
		    BmpWindowMain *self)
{
    GtkWidget *menu;
    gint x, y, w_width, w_height;
    gboolean grab = TRUE;

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    gtk_window_get_size (GTK_WINDOW(self->window), &w_width, &w_height);

    x = event->x;
    y = event->y;

    if (event->type != GDK_BUTTON_PRESS) return FALSE;

    if (event->button == 3) {

	menu = ui_manager_get_popup (bmp_ui_manager, "/popup/menu-main");
	gtk_widget_realize (menu);
	SET_CURSOR_FOR_WIDGET(CURSOR_MAINMENU,menu);
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);
	return FALSE;

    } else

	if (event->button == 1 && (x >= 36) && (x <= 99) && (y >= 26) && (y <= 39))
	  {
	    self->private->time_inverse = !self->private->time_inverse;
	    xcs_key_set_boolean (xcs,"bmp","time-remaining", self->private->time_inverse);
	    return FALSE;
	  }

    else
    /* Titlebar */
    if (event->button == 1 && ((y < 14) && (x < 240))) {

	dock_move_press (GTK_WINDOW(widget), event, TRUE);
	bmp_system_control_move_start (bmp_system_control);

    } else
    /* Textbox */
    if (event->button == 1 && self->private->textbox &&
	    ((x >= TEXTBOX_X) &&
	     (x <= TEXTBOX_X+TEXTBOX_WIDTH) &&
	     (y >= TEXTBOX_Y) &&
	     (y <= TEXTBOX_Y+11))) {

	gint w, h;

	gdk_drawable_get_size (self->private->textbox, &w, &h);

	if (w > TEXTBOX_WIDTH) {

	    self->private->tb_dragging = TRUE;
	    self->private->tb_drag_offset = self->private->tb_offset;
	    self->private->tb_drag_event_x = event->x;
	}
    }

    if (grab)
        gdk_pointer_grab(widget->window, FALSE,
                         GDK_BUTTON_MOTION_MASK |
                         GDK_BUTTON_RELEASE_MASK,
                         GDK_WINDOW(GDK_NONE), NULL, GDK_CURRENT_TIME);


    return FALSE;
}

static gboolean
on_window_main_motion_notify (GtkWidget *widget, GdkEventMotion *event, BmpWindowMain *self)
{
    GdkModifierType state;
    gint x, y;

    if (event->is_hint)
      {
	x = event->x;
        y = event->y;
        state = event->state;

      }
    else
      {
        x = event->x;
        y = event->y;
        state = event->state;

	if (self->private->tb_dragging && self->private->textbox)
	{
	    gint offset, w, h;

	    gdk_drawable_get_size (self->private->textbox, &w, &h);

	    offset = self->private->tb_drag_offset - (x - self->private->tb_drag_event_x);

	    if (offset < 0) offset = 0;
	    if (offset+TEXTBOX_WIDTH > w) offset = w - TEXTBOX_WIDTH;

	    if (offset < self->private->tb_offset)
	    {
		self->private->tb_direction = SCROLL_LEFT;
	    }
	    else if (offset > self->private->tb_offset)
	    {
		self->private->tb_direction = SCROLL_RIGHT;
	    }

	    self->private->tb_offset = offset;
	    gtk_widget_queue_draw (self->window);
	}
     }

  if (dock_is_moving(GTK_WINDOW(widget)))
      dock_move_motion(GTK_WINDOW(widget), event);

  return TRUE;
}

static gboolean
on_window_main_canvas_expose (GtkWidget *widget,
                     GdkEventExpose *event,
                     BmpWindowMain *self) {

	PangoFontDescription *pfd;
    	PangoLayout *layout;
	GtkWidget *window = self->window;
	GdkPixbuf *pb_aux;
	gchar *aux0, *aux1, *text;
	gint nums[4];
	gint n, playstatus;
	gboolean quit = FALSE;

	/* Draw background ("main") pixmap */
	gdk_draw_pixbuf(widget->window,
			NULL,
                        bmp_skin_component[SKIN_COMPONENT_MAIN]->pixbuf,
                        0,
                        0,
                        0,
                        0,
                        -1,
                        -1,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);

	/* Draw playstatus */
	playstatus = g_object_get_int (G_OBJECT(bmp_play), "playstatus");
	if (! bmp_skin_component[SKIN_COMPONENT_PLAYPAUSE]) goto skip1;

	pb_aux = bmp_skin_component[SKIN_COMPONENT_PLAYPAUSE]->pixbuf;
	if (! pb_aux) goto skip1;
	if (gdk_pixbuf_get_width (pb_aux) < 36) goto skip1;

	if (playstatus == PLAYSTATUS_PLAYING) {
        	gdk_draw_pixbuf(widget->window,
			NULL,
			pb_aux,
                        36, 0,
                        PLAYSTATUS_X, PLAYSTATUS_Y,
                        3,  9,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);
	} else {
        	gdk_draw_pixbuf(widget->window,
			NULL,
                        pb_aux,
                        27, 0,
                        PLAYSTATUS_X, PLAYSTATUS_Y,
                        2,  9,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);
	}

	switch (playstatus) {

	    case PLAYSTATUS_STOPPED: {

	    	gdk_draw_pixbuf(widget->window,
			NULL,
                        pb_aux,
                        18, 0,
                        PLAYSTATUS_X + 2,  PLAYSTATUS_Y,
                        9,  9,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);
	    }
	    break;

	    case PLAYSTATUS_PAUSED: {

	    	gdk_draw_pixbuf(widget->window,
			NULL,
                        pb_aux,
                        9, 0,
                        PLAYSTATUS_X + 2,  PLAYSTATUS_Y,
                        9, 9,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);
	    }
	    break;

	    case PLAYSTATUS_PLAYING: {

	    	gdk_draw_pixbuf(widget->window,
			NULL,
                        pb_aux,
                        1, 0,
                        PLAYSTATUS_X + 3,  PLAYSTATUS_Y,
                        8, 9,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);
	    }
	    break;

	    default: break;
	}

	skip1:

	if ( (event->area.x >= TEXTBOX_X) &&
	     (event->area.x <= TEXTBOX_X+TEXTBOX_WIDTH) &&
	     (event->area.width <= TEXTBOX_WIDTH) &&
	     (event->area.height <= 11)) {

	    quit = TRUE;
	    goto expose_draw_textbox;
	}

	expose_draw_textbox:

	if (self->private->textbox != NULL) {

	    gint w,h;

	    gdk_drawable_get_size (self->private->textbox, &w, &h);

	    gdk_draw_drawable(widget->window,
				widget->style->fg_gc[GTK_STATE_NORMAL],
				self->private->textbox,
				self->private->tb_offset, 0,
				TEXTBOX_X, TEXTBOX_Y,
				TEXTBOX_WIDTH,
				h);
	}
	if (quit) return TRUE;

	/* Bitrate */
	text = g_strdup_printf("%d", g_object_get_int(bmp_system_control, "bitrate")/1000);
	if (text) {

    		layout = gtk_widget_create_pango_layout (widget, "");
		aux0 = g_markup_escape_text (text, strlen(text));
		aux1 = g_strconcat ("<b>", aux0, "</b>", NULL);

		pfd = pango_font_description_copy (bmp_fonts[BMP_FONT_NORMAL]->font_description);
		pango_font_description_set_absolute_size (pfd, 8*PANGO_SCALE);

		pango_layout_set_markup (layout, aux1, strlen(aux1));
    		pango_layout_set_font_description (layout, pfd);
		pango_font_description_free (pfd);
    		gtk_paint_layout            (GTK_WIDGET (widget)->style,
                                             GTK_WIDGET (widget)->window,
                                             GTK_STATE_NORMAL,
                                             TRUE,
                                             NULL,
                                             widget,
                                             "",
                                             111,
                                             41,
                                             layout);
		g_free (aux0);
		g_free (aux1);
		g_free (text);
		g_object_unref (layout);
	}

	/* Samplerate */
	text = g_strdup_printf("%d", g_object_get_int(bmp_system_control, "samplerate")/1000);
	if (text)
	{
    		layout = gtk_widget_create_pango_layout (widget, "");

		pfd = pango_font_description_copy (bmp_fonts[BMP_FONT_NORMAL]->font_description);
		pango_font_description_set_absolute_size (pfd, 8*PANGO_SCALE);

		aux0 = g_markup_escape_text (text, strlen(text));
		aux1 = g_strconcat ("<b>", aux0, "</b>", NULL);
		pango_layout_set_markup (layout, aux1, strlen(aux1));
    		pango_layout_set_font_description (layout, pfd);
		pango_font_description_free (pfd);
    		gtk_paint_layout            (GTK_WIDGET (widget)->style,
                                             GTK_WIDGET (widget)->window,
                                             GTK_STATE_NORMAL,
                                             TRUE,
                                             NULL,
                                             widget,
                                             "",
                                             155,
                                             41,
                                             layout);

		g_free (aux0);
		g_free (aux1);
		g_free (text);
		g_object_unref (layout);
	}

	if ( (event->area.x == 0) &&
	     (event->area.y == 0) &&
	     (event->area.width == 275) &&
	     (event->area.height == 14)) {

			quit = TRUE;
			goto on_window_main_canvas_expose_DRAW_TITLEBAR;
	}

	if ( (event->area.x == 36) &&
	     (event->area.y == 26) &&
	     (event->area.width == 99) &&
	     (event->area.height == 39)) {

			quit = TRUE;
			goto on_window_main_canvas_expose_DRAW_NUMBERS;
	}

	on_window_main_canvas_expose_DRAW_TITLEBAR:
	/* Draw titlebar */
	pb_aux = gdk_pixbuf_new (gdk_pixbuf_get_colorspace(bmp_skin_component[SKIN_COMPONENT_TITLEBAR]->pixbuf),
                 		 gdk_pixbuf_get_has_alpha(bmp_skin_component[SKIN_COMPONENT_TITLEBAR]->pixbuf),
                                 gdk_pixbuf_get_bits_per_sample(bmp_skin_component[SKIN_COMPONENT_TITLEBAR]->pixbuf),
                                 275,
				 14);

	/* FIXME: Precache these pixbufs */
	if (gtk_window_has_toplevel_focus (GTK_WINDOW(window))) {

		gdk_pixbuf_copy_area (bmp_skin_component[SKIN_COMPONENT_TITLEBAR]->pixbuf,
				27,
				0,
				275,
				14,
				pb_aux,
				0,
				0);
	} else {

		gdk_pixbuf_copy_area (bmp_skin_component[SKIN_COMPONENT_TITLEBAR]->pixbuf,
				27,
				15,
				275,
				14,
				pb_aux,
				0,
				0);
	}

	gdk_draw_pixbuf(widget->window,
			NULL,
                        pb_aux,
                        0,
                        0,
                        0,
                        0,
                        -1,
                        -1,
                        GDK_RGB_DITHER_NONE,
                        0,
                        0);
	g_object_unref (pb_aux);
	if (quit) return FALSE;

	on_window_main_canvas_expose_DRAW_NUMBERS:

	/* FIXME: Hack */
	if (playstatus == PLAYSTATUS_STOPPED) {
	    self->private->time_minutes = 0;
	    self->private->time_seconds = 0;
	}

	/* Draw numbers */
	if (abs(self->private->time_minutes) < 100)
	  {
	    nums[NUMBER_MINUTES_10] = abs(self->private->time_minutes) / 10;
	    nums[NUMBER_MINUTES_1]  = abs(self->private->time_minutes) % 10;
	    nums[NUMBER_SECONDS_10] = abs(self->private->time_seconds) / 10;
	    nums[NUMBER_SECONDS_1]  = abs(self->private->time_seconds) % 10;
	  }
	else
	  {
	    gint time_hours, time_minutes;
  
	    time_hours   = self->private->time_minutes / 60;
	    time_minutes = self->private->time_minutes % 60;

	    nums[NUMBER_MINUTES_10] = abs(time_hours) / 10;
	    nums[NUMBER_MINUTES_1]  = abs(time_hours) % 10;
	    nums[NUMBER_SECONDS_10] = abs(time_minutes) / 10;
	    nums[NUMBER_SECONDS_1]  = abs(time_minutes) % 10;
	  }

	/* Draw Time */
	for (n = 0; n < 4; n++) {

		if (!numbers[nums[n]])
		  continue;

		gdk_draw_pixbuf(widget->window,
				NULL,
                	        numbers[nums[n]],
	                        0,
        	                0,
                	        number_positions[n].x,
	                        number_positions[n].y,
        	                -1,
                	        (gdk_pixbuf_get_height(numbers[nums[n]]) < 13) ? gdk_pixbuf_get_height(numbers[nums[n]]) : 13,
	                        GDK_RGB_DITHER_NONE,
        	                0,
                	        0);
	}

	if ((((self->private->time_minutes*60)+(self->private->time_seconds)) > 0) && (numbers[10] != NULL)) {

		gdk_draw_pixbuf(widget->window,
				NULL,
                	        numbers[10],
	                        0,
        	                0,
                	        number_positions[NUMBER_MINUS].x,
	                        number_positions[NUMBER_MINUS].y,
        	                -1,
                	        (gdk_pixbuf_get_height(numbers[10]) < 13) ? gdk_pixbuf_get_height(numbers[10]) : 13,
	                        GDK_RGB_DITHER_NONE,
        	                0,
                	        0);
	}
	else
	if ((((self->private->time_minutes*60)+(self->private->time_seconds)) < 0) && (numbers[11] != NULL)) {

		gdk_draw_pixbuf(widget->window,
				NULL,
                	        numbers[11],
	                        0,
        	                0,
                	        number_positions[NUMBER_MINUS].x,
	                        number_positions[NUMBER_MINUS].y,
        	                -1,
                	        (gdk_pixbuf_get_height(numbers[11]) < 13) ? gdk_pixbuf_get_height(numbers[11]) : 13,
	                        GDK_RGB_DITHER_NONE,
        	                0,
                	        0);
	}
	if (quit) return FALSE;

	return FALSE;
}

static gboolean
tb_offset_timeout (BmpWindowMain *self)
{
	static gint wait_cycle = 0;

	if (self->private->tb_dragging)
	    return TRUE;

	G_LOCK (TEXTBOX_LOCK);

	if (! self->private->textbox)
	{
	    self->private->tb_running = FALSE;
	    G_UNLOCK (TEXTBOX_LOCK);
	    return FALSE;
	}

	if (! self->private->tb_running)
	{
	    G_UNLOCK (TEXTBOX_LOCK);
	    return FALSE;
	}

	if (wait_cycle != 0)
	{
	    wait_cycle--;
	    G_UNLOCK (TEXTBOX_LOCK);
	    return TRUE;
	}

	switch (self->private->tb_direction)
	{
		case SCROLL_RIGHT:
		{
			gint offset, w, h;

			gdk_drawable_get_size (self->private->textbox, &w, &h);

			offset = self->private->tb_offset + 1;

			if ((offset+TEXTBOX_WIDTH) > w)
			{
			    self->private->tb_direction = SCROLL_LEFT;
			    wait_cycle = 20;
			    G_UNLOCK (TEXTBOX_LOCK);
			    return TRUE;
			}
			else
			{
			    self->private->tb_offset = offset;
			    G_UNLOCK (TEXTBOX_LOCK);
			    gtk_widget_queue_draw_area (self->window, TEXTBOX_X, TEXTBOX_Y, TEXTBOX_X+TEXTBOX_WIDTH-1, TEXTBOX_Y+10);
			    return TRUE;
			}
		}
		break;

		case SCROLL_LEFT:
		{
			gint offset, w, h;

			gdk_drawable_get_size (self->private->textbox, &w, &h);

			offset = self->private->tb_offset - 1;

			if (offset < 0)
			{
			    self->private->tb_direction = SCROLL_RIGHT;
			    wait_cycle = 20;
			    G_UNLOCK (TEXTBOX_LOCK);
			    return TRUE;
			}
			else
			{
			    self->private->tb_offset = offset;
			    G_UNLOCK (TEXTBOX_LOCK);
			    gtk_widget_queue_draw_area (self->window, TEXTBOX_X, TEXTBOX_Y, TEXTBOX_X+TEXTBOX_WIDTH-1, TEXTBOX_Y+10);
			    return TRUE;
			}
		}
		break;

		default: {
			    self->private->tb_running = FALSE;
			    G_UNLOCK (TEXTBOX_LOCK);
			    return FALSE; }
	}

	self->private->tb_running = FALSE;
	G_UNLOCK(TEXTBOX_LOCK);
	return FALSE;
}

static void
update_textbox (BmpWindowMain *self, const gchar *text) 
{

    PangoFontDescription *pfd;
    PangoLayout *layout;
    PangoRectangle rect_ink, rect_logical;
    GdkGC *gc;

    /* Recreate textbox pixmap */
    /* FIXME: Operate with Cairo here instead of GDK */

    G_LOCK(TEXTBOX_LOCK);

    if (text)
      {
	layout =
	    gtk_widget_create_pango_layout (BMP_WINDOW(self->window)->canvas, "");
	pfd =
	    pango_font_description_copy (bmp_fonts[BMP_FONT_NORMAL]->font_description);

	pango_font_description_set_absolute_size (pfd, TEXTBOX_FONT_SIZE*PANGO_SCALE);
	pango_layout_set_font_description (layout, pfd);
	pango_font_description_free (pfd);
	pango_layout_set_text (layout, text, strlen(text));
	pango_layout_get_pixel_extents (layout, &rect_ink, &rect_logical);

	if (rect_logical.width < TEXTBOX_WIDTH) rect_logical.width = TEXTBOX_WIDTH;

	if (self->private->textbox != NULL) g_object_unref (self->private->textbox);
	self->private->textbox = gdk_pixmap_new(GTK_WIDGET(BMP_WINDOW(self->window)->canvas)->window,
						rect_logical.width,
						rect_logical.height,
						-1);

	gc =
	    gdk_gc_new (self->private->textbox);
	gdk_gc_set_rgb_fg_color (gc, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));
	gdk_draw_rectangle(self->private->textbox,
			    gc,
			    TRUE,
			    0, 0,
			    rect_logical.width, rect_logical.height);

	gdk_gc_set_rgb_fg_color (gc, bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT));
	gdk_draw_layout(self->private->textbox,
			gc,
			0, 0,
			layout);
	gdk_gc_destroy (gc);

	self->private->tb_offset = 0;
	self->private->tb_direction = SCROLL_RIGHT;

	if (! self->private->tb_running)
          {
		g_timeout_add (50, (GSourceFunc)tb_offset_timeout, self);
		self->private->tb_running = TRUE;
	  } 

      }
    else if (self->private->textbox)
      {

	gc = gdk_gc_new (self->private->textbox);
	gdk_gc_set_rgb_fg_color (gc, bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));
	gdk_draw_rectangle(self->private->textbox,
			    gc,
			    TRUE,
			    0, 0,
			    TEXTBOX_WIDTH, 13);

	gdk_gc_destroy (gc);

	self->private->tb_offset = 0;
	self->private->tb_running = FALSE;
      }

    G_UNLOCK(TEXTBOX_LOCK);
}

static void
on_system_control_set_title (BmpSystemControl *control, BmpWindowMain *self)
{
    gchar *text;

    text = g_object_get_string (control, "current_title");
    update_textbox (self, text);
    g_free (text);
    gtk_widget_queue_draw (self->window);
}

static void
on_system_control_track_change (BmpSystemControl *control, BmpWindowMain *self)
{
    gchar *text;

    g_return_if_fail (BMP_IS_SYSTEM_CONTROL(control));
    g_return_if_fail (BMP_IS_WINDOW_MAIN(self));

    text = g_object_get_string (bmp_system_control, "current_title"); 
    update_textbox (self, text);
    g_free (text);

    bmp_window_main_set_seek_request (self, -1);
    reset_last_position (self);

    gtk_widget_queue_draw (self->window);
}

static void
on_bmp_font_changed (Xcs *configuration, BmpWindowMain *self)
{
    gchar *text;

    if (GTK_IS_WIDGET(self->window))
       {
	  text = g_object_get_string (bmp_system_control, "current_title"); 
	  update_textbox (self, text);
	  g_free (text);
	  gtk_widget_queue_draw (self->window);
       }
}

static void
create_actions(BmpWindowMain *self)
{
  GError *error = NULL;

  gtk_action_group_add_actions (bmp_actions, bmp_actions_MAIN, G_N_ELEMENTS (bmp_actions_MAIN), self);
  gtk_action_group_add_toggle_actions (bmp_toggle_actions, bmp_toggle_actions_MAIN, G_N_ELEMENTS (bmp_toggle_actions_MAIN), self);

  if (!gtk_ui_manager_add_ui_from_file (bmp_ui_manager, DATA_DIR G_DIR_SEPARATOR_S BMP_UI_BASENAME G_DIR_SEPARATOR_S "main.ui", &error))
    {
		g_message ("building menus failed: %s", error->message);
		g_error_free (error);
		error = NULL;
    }
}

static void
allocate_widgets(BmpWindowMain *self) {

  BmpTooltips *tooltips;
  BmpWindow *window = BMP_WINDOW(self->window);
  BmpButton *button;
  BmpButtonToggle *button_toggle;
  GtkAction *action;
  gint i;

  array_buttons = g_ptr_array_new();
  array_buttons_toggle = g_ptr_array_new();

  tooltips = bmp_tooltips_new();
  bmp_tooltips_set_draw_hotspot (tooltips, TRUE);
  RM_ITEM_NEW("tooltips", "tooltips-main", tooltips, NULL);

  /* Normal buttons */
  for (i = 0; i < nbuttons_window_main; ++i) {

      button = bmp_button_new((bmp_window_main_buttons[i].action_id != NULL) ? FALSE : TRUE);

      if (bmp_window_main_buttons[i].action_id != NULL) {
             action = gtk_action_group_get_action(bmp_actions, bmp_window_main_buttons[i].action_id);
             if (action != NULL) {
                     gtk_action_connect_proxy (GTK_ACTION (action), GTK_WIDGET (button));
             }
      }
      else {
	    gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);
      }

     bmp_tooltips_set_tip (tooltips,
                           GTK_WIDGET (button),
                           GTK_STOCK_INFO,
                           Q_(bmp_window_main_buttons[i].tooltip),
                           NULL,
                           FALSE);

     /* FIXME: Make this an appropriate method of BmpWindow */
     gtk_fixed_put(GTK_FIXED (BMP_WINDOW(window)->canvas),
		GTK_WIDGET (button),
		bmp_window_main_buttons[i].pos_x,
		bmp_window_main_buttons[i].pos_y);
     g_ptr_array_add(array_buttons,button);
  }

  /* Toggle buttons */
  for (i = 0; i < nbuttons_window_main_toggle; ++i) {

      button_toggle = bmp_button_toggle_new();

      if (bmp_window_main_buttons_toggle[i].action_id != NULL) {
            action = gtk_action_group_get_action(bmp_toggle_actions, bmp_window_main_buttons_toggle[i].action_id);
            if (action != NULL)
		gtk_action_connect_proxy(GTK_ACTION (action), GTK_WIDGET (button_toggle));
        }
      else
	    gtk_widget_set_sensitive(GTK_WIDGET(button_toggle), FALSE);

      bmp_tooltips_set_tip (tooltips,
                            GTK_WIDGET (button_toggle),
                            GTK_STOCK_INFO,
                            Q_(bmp_window_main_buttons_toggle[i].tooltip),
                            NULL,
                            TRUE);

      gtk_fixed_put (GTK_FIXED(BMP_WINDOW(window)->canvas),
		GTK_WIDGET(button_toggle),
		bmp_window_main_buttons_toggle[i].pos_x,
		bmp_window_main_buttons_toggle[i].pos_y);
      g_ptr_array_add (array_buttons_toggle, button_toggle);
  }

  /* Posbar slider */
  self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK] = bmp_slider_new();
  bmp_slider_set_scroll_step (BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), BMP_SLIDER_DEFAULT_STEP);
  g_object_connect (G_OBJECT(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]),
		    "signal::moved",
		    G_CALLBACK (bmp_ui_callback_seek_moved),
		    self,
		    "signal::button-press-event",
		    G_CALLBACK (seek_started),
		    self,
		    "signal::button-release-event",
		    G_CALLBACK (seek_ended),
		    self,
		    NULL);
  gtk_fixed_put (GTK_FIXED(BMP_WINDOW(window)->canvas),
	    GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]),
	    SLIDER_SEEK_X,
	    SLIDER_SEEK_Y);
  bmp_tooltips_set_tip (tooltips,
		self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK],
		GTK_STOCK_INFO,
		N_("With this slider you can seek within the track"),
		NULL,
		TRUE);
  gtk_widget_set_sensitive (GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), FALSE);

  /* Volume slider */
  self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME] = bmp_slider_new();
  bmp_slider_set_scroll_step (BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]),
			     BMP_SLIDER_DEFAULT_STEP);
  bmp_slider_set_report_on_motion (BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]),
			     TRUE);
  g_object_connect (G_OBJECT(self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]),
		    "signal::moved",
		    G_CALLBACK(bmp_ui_callback_volume_moved),
		    NULL,
		    NULL);
  gtk_fixed_put (GTK_FIXED(BMP_WINDOW(window)->canvas),
	    GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]),
	    SLIDER_VOLUME_X,
	    SLIDER_VOLUME_Y);
  bmp_tooltips_set_tip (tooltips,
		self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME],
		GTK_STOCK_INFO,
		N_("This slider controls the volume"),
		NULL,
		TRUE);

  /* Balance slider */
  self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE] = bmp_slider_new();
  bmp_slider_set_scroll_step (BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE]),
			     BMP_SLIDER_DEFAULT_STEP);
  g_object_connect (G_OBJECT(self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE]),
		    "signal::moved",
		    G_CALLBACK(bmp_ui_callback_balance_moved),
		    NULL,
		    NULL);
  gtk_fixed_put (GTK_FIXED(BMP_WINDOW(window)->canvas),
	    GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE]),
	    SLIDER_BALANCE_X,
	    SLIDER_BALANCE_Y);
  bmp_tooltips_set_tip (tooltips,
		self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE],
		GTK_STOCK_INFO,
		N_("This slider controls the left/right balance"),
		NULL,
		TRUE);

  for (i = 0; i < 10; numbers[i++] = NULL);

  if (xcs_key_fetch_boolean (xcs, "bmp", "display-tooltips"))
    bmp_tooltips_enable (tooltips);
  else
    bmp_tooltips_disable (tooltips);

    gtk_widget_set_sensitive (GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), FALSE);

    gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PREV), FALSE);
    gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_NEXT), FALSE);
    gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PAUSE), FALSE);
    gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_STOP), FALSE);
}

void
bmp_window_main_iconify (BmpWindowMain *self) {

	GtkAction *action;

	action = gtk_action_group_get_action (bmp_actions, BMP_ACTION_ICONIFY);
	gtk_action_activate (action);
}

static void
mw_adapt_cursor (BmpWindow *window, gint x, gint y) {

#if 0
   GdkCursor *cursor;
   GdkCursorType cursor_type = GDK_LEFT_PTR;
   gint w_width, w_height;

   w_width = GTK_WIDGET(window)->allocation.width;
   w_height = GTK_WIDGET(window)->allocation.height;

   // FIXME: Inconsistency. This should use a stat val too. Also non-unified check for condition.
   if ((y < 16) && (w_width-x > 32))
   	{
		cursor_type = GDK_FLEUR;
		goto mw_adapt_cursor__set_cursor;
	}

  mw_adapt_cursor__set_cursor:

  cursor = gdk_cursor_new_for_display (gdk_display_get_default(), cursor_type);
  gdk_window_set_cursor (GDK_WINDOW (GTK_WIDGET (BMP_WINDOW(window)->canvas)->window), cursor);
  return;
#endif
}

static void
on_system_control_set_playstatus (BmpSystemControl *system_control, gint status, BmpWindowMain *self) {
    gchar *title;

    g_return_if_fail (BMP_IS_WINDOW_MAIN(self));

    if (status & PLAYSTATUS_SEEKING) {

	gtk_widget_queue_draw (self->window);
        return;
    }

    if (status & PLAYSTATUS_STOPPED) {

	bmp_ui_set_trayicon_text (bmp_ui, "(Not Playing)"); 

	reset_last_position (self);
	bmp_slider_set_position (BMP_SLIDER(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), 0);
	gtk_widget_set_sensitive (GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), FALSE);

	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PREV), FALSE);
        gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_NEXT), FALSE);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PAUSE), FALSE);
        gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_STOP), FALSE);

	gtk_widget_queue_draw (self->window);
        return;
    }

    if (status & PLAYSTATUS_PAUSED) {

        title = bmp_playlist_get_current_title (bmp_playlist);
	bmp_ui_set_trayicon_text (bmp_ui, title);
	g_free (title);

	gtk_widget_set_sensitive (GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), FALSE);
	gtk_widget_queue_draw (self->window);
        return;
    }

    if (status & PLAYSTATUS_PLAYING) {

        title = bmp_playlist_get_current_title (bmp_playlist);
	bmp_ui_set_trayicon_text (bmp_ui, title);
	g_free (title);

	gtk_widget_set_sensitive (GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), TRUE);

	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PREV), TRUE);
        gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_NEXT), TRUE);
	gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_PAUSE), TRUE);
        gtk_action_set_sensitive (gtk_action_group_get_action (bmp_actions, BMP_ACTION_STOP), TRUE);

	gtk_widget_queue_draw (self->window);
        return;
    }
}

static void
window_main_create(BmpWindowMain *self)
{
    BmpWindow *window;

    self->window = GTK_WIDGET(bmp_window_new(GTK_WINDOW_TOPLEVEL));
    window = BMP_WINDOW(self->window);

    gtk_widget_set_size_request (GTK_WIDGET(BMP_WINDOW(window)->canvas), BMP_WINDOW_MAIN_WIDTH, BMP_WINDOW_MAIN_HEIGHT);

    create_actions (self);
    allocate_widgets (self);

    bmp_window_set_icon_list (GTK_WIDGET(window), "player");
    gtk_window_set_title (GTK_WINDOW(window), "BMPx");

    g_object_set_data (G_OBJECT(window), "id", g_strdup("main"));

    dock_window_add (GTK_WINDOW(window));

    g_object_connect (G_OBJECT(g_ptr_array_index(array_buttons, MAIN_BUTTON_STOP)),
			"signal::button-press-event",
			G_CALLBACK(on_window_main_stop_button_press),
			self,
			NULL);

    g_object_connect (G_OBJECT (BMP_WINDOW(window)),
			"signal::focus-in-event",
			G_CALLBACK(on_window_main_focus_in),
		        BMP_WINDOW(window)->canvas,
 			"signal::focus-out-event",
			G_CALLBACK(on_window_main_focus_out),
			BMP_WINDOW(window)->canvas,
  			"signal::button-press-event",
			G_CALLBACK(on_window_main_button_press),
			self,
			"signal::button-release-event",
			G_CALLBACK(on_window_main_button_release),
			self,
			"signal::motion-notify-event",
			G_CALLBACK(on_window_main_motion_notify),
			self,
			"signal::key-press-event",
			G_CALLBACK(on_window_main_key_press),
			self,
			"signal::scroll-event",
			G_CALLBACK(on_window_main_scroll_event),
			self,
			"signal::window-state-event",
			G_CALLBACK(on_window_main_state_event),
			NULL,
			"swapped-signal::delete-event",
			G_CALLBACK(bmp_system_control_quit),
			bmp_system_control,
			NULL);

    g_object_connect (G_OBJECT(bmp_system_control),
			"signal::track_change",
			G_CALLBACK(on_system_control_track_change),
			self,
			"signal::set-stream-pos",
			G_CALLBACK(on_system_control_stream_pos),
			self,
			"signal::set-volume",
			G_CALLBACK(on_system_control_set_volume),
			self,
			"signal::set-playstatus",
			G_CALLBACK(on_system_control_set_playstatus),
			self,
			"signal::set-title",
			G_CALLBACK(on_system_control_set_title),
			self,
			NULL);

    g_object_connect (G_OBJECT(xcs),
			"signal::changed::bmp-font",
			G_CALLBACK(on_bmp_font_changed),
			self,
			NULL);

    g_object_connect (G_OBJECT(BMP_WINDOW(window)->canvas),
			"signal::expose-event",
			G_CALLBACK(on_window_main_canvas_expose),
			self,
			NULL);

    gtk_window_add_accel_group (GTK_WINDOW(window), gtk_ui_manager_get_accel_group(bmp_ui_manager));
}

void
bmp_window_main_configure(BmpWindowMain *self)
{
    BmpWindow *window = BMP_WINDOW(self->window);
    BmpButton *button;
    BmpButtonToggle *button_toggle;
    GdkPixbuf *pixbuf_aux;
    gchar *text;
    gint i;

    gdk_window_shape_combine_mask (GTK_WIDGET(window)->window, bmp_ui_get_mask_window_main(bmp_ui), 0, 0);
    gdk_flush();

    /* Normal buttons */
    for (i = 0; i < nbuttons_window_main; ++i) {

        button =
	    g_ptr_array_index(array_buttons, i);

        bmp_button_set_from_table(BMP_BUTTON (button),
				  bmp_skin_component[bmp_window_main_buttons[i].pixmap_id]->pixbuf,
				  bmp_window_main_buttons[i].width,
				  bmp_window_main_buttons[i].height,
				  bmp_window_main_buttons[i].img_x,
				  bmp_window_main_buttons[i].img_y,
				  bmp_window_main_buttons[i].img_x_pressed,
				  bmp_window_main_buttons[i].img_y_pressed);

        gtk_widget_set_size_request (GTK_WIDGET (button),
	    bmp_window_main_buttons[i].width,
	    bmp_window_main_buttons[i].height);
    }
    BMP_BUTTON(g_ptr_array_index(array_buttons, MAIN_BUTTON_ICONIFY))->passtrough = FALSE;
    BMP_BUTTON(g_ptr_array_index(array_buttons, MAIN_BUTTON_QUIT))->passtrough = FALSE;

    /* Toggle buttons */
    for (i = 0; i < nbuttons_window_main_toggle; ++i) {

        button_toggle =
	    g_ptr_array_index(array_buttons_toggle, i);

        bmp_button_toggle_set_from_table (BMP_BUTTON_TOGGLE (button_toggle),
					  bmp_skin_component[bmp_window_main_buttons_toggle[i].pixmap_id]->pixbuf,
					  bmp_window_main_buttons_toggle[i].width,
					  bmp_window_main_buttons_toggle[i].height,
					  bmp_window_main_buttons_toggle[i].img_x,
					  bmp_window_main_buttons_toggle[i].img_y,
					  bmp_window_main_buttons_toggle[i].img_x_pressed,
					  bmp_window_main_buttons_toggle[i].img_y_pressed,
					  bmp_window_main_buttons_toggle[i].img_x_active,
					  bmp_window_main_buttons_toggle[i].img_y_active );

        gtk_widget_set_size_request (GTK_WIDGET (button_toggle),
	    bmp_window_main_buttons_toggle[i].width,
	    bmp_window_main_buttons_toggle[i].height);
    }

    /* Posbar slider */
    if (gdk_pixbuf_get_width(bmp_posbar->pixbuf) == 307) {

        bmp_slider_set_from_table (BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), bmp_posbar->pixbuf,
				   247,
				   gdk_pixbuf_get_height(bmp_posbar->pixbuf),
				   29,
				   gdk_pixbuf_get_height(bmp_posbar->pixbuf),
				   0,
				   0,
				   248,
				   0,
				   278,
				   0,
				   TRUE);

	gtk_widget_set_size_request (GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]),
				      247,
				      gdk_pixbuf_get_height(bmp_posbar->pixbuf));

        gtk_widget_show (GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]));

  	SET_CURSOR_FOR_WIDGET(CURSOR_POSBAR, self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]);
    }
    else
        gtk_widget_hide (GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]));

    /* Volume slider */
    bmp_slider_set_from_table (BMP_SLIDER(self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]), bmp_volume->pixbuf,
				(gdk_pixbuf_get_width(bmp_volume->pixbuf) > 68) ? 68 : gdk_pixbuf_get_width(bmp_volume->pixbuf),
                		gdk_pixbuf_get_height(bmp_volume->pixbuf),
				14,
				bmp_volume->height,
				0,
				0,
                		15,
                	 	422,
                		0,
                	 	422,
				bmp_volume->has_button);



    if (bmp_volume->has_button) {
  	gtk_widget_set_size_request (GTK_WIDGET (self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]),
	    (gdk_pixbuf_get_width(bmp_volume->pixbuf) > 68) ? 68 : gdk_pixbuf_get_width(bmp_volume->pixbuf),
	     gdk_pixbuf_get_height(bmp_volume->pixbuf) - 421);
    } else {
  	gtk_widget_set_size_request (GTK_WIDGET (self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]),
	    (gdk_pixbuf_get_width(bmp_volume->pixbuf) > 68) ? 68 : gdk_pixbuf_get_width(bmp_volume->pixbuf),
	     13);
    }

    bmp_slider_set_slice_data(BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]),
	     TRUE,
	     15,
	     27,
	     bmp_volume->has_button);

    BMP_SLIDER(self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME])->button_yoffset = 1;
    SET_CURSOR_FOR_WIDGET(CURSOR_VOLBAR,self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]);
    bmp_slider_set_position (BMP_SLIDER(self->private->sliders[BMP_UI_MAIN_SLIDER_VOLUME]),
 		g_object_get_int (G_OBJECT(bmp_system_control), "volume"));

    /* Balance slider */
    if (bmp_balance)
      {
	BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE])->max_adjust = 2;
	bmp_slider_set_from_table (BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE]), bmp_balance->pixbuf,
                		38,
                		gdk_pixbuf_get_height(bmp_balance->pixbuf),
				14,
				bmp_balance->height-1,
				9,
				0,
                		15,
                		422,
                		0,
                		422,
				TRUE);
	if ((!bmp_balance->has_button) && (bmp_volume->has_button) )
          {
	    pixbuf_aux = gdk_pixbuf_new (gdk_pixbuf_get_colorspace(bmp_volume->pixbuf),
                                            gdk_pixbuf_get_has_alpha(bmp_volume->pixbuf),
                                            gdk_pixbuf_get_bits_per_sample(bmp_volume->pixbuf),
                                            14, 11);
	    gdk_pixbuf_copy_area(bmp_volume->pixbuf,15,421,14,gdk_pixbuf_get_height(bmp_volume->pixbuf)-421,pixbuf_aux,0,0);
	    g_object_unref (pixbuf_aux);
	    BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE])->pb_pressed = pixbuf_aux;
	    BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE])->pb_normal  = pixbuf_aux;
	    bmp_balance->has_button = TRUE;
	  }

	BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE])->button_yoffset = 1;
	gtk_widget_set_size_request (GTK_WIDGET(self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE]), 38, 13);
	bmp_slider_set_slice_data(BMP_SLIDER (self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE]), TRUE, 15, 27, bmp_balance->has_button);
    }

    reset_last_position (self);
    bmp_slider_set_position (BMP_SLIDER(self->private->sliders[BMP_UI_MAIN_SLIDER_SEEK]), 0);

    gtk_widget_queue_draw (self->window);

#ifdef HAVE_XINE
    /* XINE has no balance support FIXME: Is that right? */
    gtk_widget_set_sensitive (self->private->sliders[BMP_UI_MAIN_SLIDER_BALANCE], FALSE);
#endif

    /* Main window numbers */
    for (i = 0; i < 12; ++i) {

        if (numbers[i]) {
	    g_object_unref (numbers[i]);
	    numbers[i] = NULL;
        }
    }

    for (i = 0; i < GPOINTER_TO_INT(bmp_skin_component[SKIN_COMPONENT_NUMBERS]->user_data); ++i)
    {
	numbers[i] =
		gdk_pixbuf_new (gdk_pixbuf_get_colorspace(bmp_skin_component[SKIN_COMPONENT_NUMBERS]->pixbuf),
                                            gdk_pixbuf_get_has_alpha(bmp_skin_component[SKIN_COMPONENT_NUMBERS]->pixbuf),
                                            gdk_pixbuf_get_bits_per_sample(bmp_skin_component[SKIN_COMPONENT_NUMBERS]->pixbuf),
                                            9,
					    gdk_pixbuf_get_height(bmp_skin_component[SKIN_COMPONENT_NUMBERS]->pixbuf));

	gdk_pixbuf_copy_area (bmp_skin_component[SKIN_COMPONENT_NUMBERS]->pixbuf,
				0 + (i*9),
				0,
				9,
				gdk_pixbuf_get_height(bmp_skin_component[SKIN_COMPONENT_NUMBERS]->pixbuf),
				numbers[i],
				0,
				0);
    }

    SET_CURSOR_FOR_WIDGET(CURSOR_NORMAL,BMP_WINDOW(window)->canvas);
    SET_CURSOR_FOR_WIDGET(CURSOR_NORMAL,window);

    gtk_widget_modify_bg
	(GTK_WIDGET(BMP_WINDOW(window)->canvas),
	    GTK_STATE_NORMAL,
	    bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));

    gtk_widget_modify_base
	(GTK_WIDGET(BMP_WINDOW(window)->canvas),
	    GTK_STATE_NORMAL,
	    bmp_colors_get_color(BMP_SKINCOLOR_PL_BG_NORMAL));

    gtk_widget_modify_fg
	(GTK_WIDGET(BMP_WINDOW(window)->canvas),
	    GTK_STATE_NORMAL,
	    bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT));

    gtk_widget_modify_text
	(GTK_WIDGET(BMP_WINDOW(window)->canvas),
	    GTK_STATE_NORMAL,
	    bmp_colors_get_color(BMP_SKINCOLOR_PL_FG_CURRENT));

    text = g_object_get_string (bmp_system_control, "current_title"); 
    update_textbox (self, text);
    g_free (text);

    gtk_widget_queue_draw (GTK_WIDGET (window));
    gdk_flush();
}

/* GObject */
BmpWindowMain*
bmp_window_main_new (void)
{
    BmpWindowMain *self;

    self =
        g_object_new (BMP_TYPE_WINDOW_MAIN, NULL);

    return self;
}

static GObject *
bmp_window_main_constructor (GType type,
	                     guint n_construct_properties,
         	             GObjectConstructParam * construct_properties)
{
    GObject *obj;

    {
        BmpWindowMainClass *klass;
        GObjectClass *parent_class;
        klass = BMP_WINDOW_MAIN_CLASS (g_type_class_peek (BMP_TYPE_WINDOW_MAIN));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj =
            parent_class->constructor (type,
                                       n_construct_properties,
                                       construct_properties);
    }

    return obj;
}

static void
bmp_window_main_dispose (GObject *gobj)
{
    BmpWindowMain *self = (BmpWindowMain *) gobj;

#if 0
    gint n;
#endif

    if (self->private->dispose_has_run)
         return;

#if 0
    g_object_unref (BMP_WINDOW(self->window)->canvas);
    g_object_unref (self->window);

    for (n = 0; n < BMP_UI_MAIN_N_SLIDERS; n++)
    {
	gtk_widget_destroy(self->private->sliders[n]);
    }

    g_ptr_array_foreach (array_buttons, gtk_widget_destroy, NULL);
    g_ptr_array_foreach (array_buttons_toggle, gtk_widget_destroy, NULL);
#endif

    self->private->dispose_has_run = TRUE;
}

static void
bmp_window_main_finalize (GObject *gobj)
{}

static void
bmp_window_main_class_init (BmpWindowMainClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

#if 0
    gobject_class->set_property = bmp_play_set_property;
    gobject_class->get_property = bmp_play_get_property;
#endif
    gobject_class->finalize = bmp_window_main_finalize;
    gobject_class->dispose = bmp_window_main_dispose;
    gobject_class->constructor = bmp_window_main_constructor;
}

static void
bmp_window_main_init (BmpWindowMain *self)
{
    self->private = g_new (BmpWindowMainPrivate, 1);

    self->private->dispose_has_run = FALSE;

    self->private->seeking = FALSE;

    self->private->seek_request = -1;
    self->private->last_position = -1;
    self->private->time_minutes = 0;
    self->private->time_seconds = 0;
    self->private->time_inverse = xcs_key_fetch_boolean (xcs, "bmp", "time-remaining"); 

    self->private->textbox = NULL;

    self->private->tb_offset = 0;
    self->private->tb_drag_offset = 0;
    self->private->tb_running = FALSE;
    self->private->tb_dragging = FALSE;

    window_main_create(self);
}
