#ifndef _XCS_H_
#define _XCS_H_

#ifdef HAVE_GUI
#  include <gtk/gtk.h>
#endif

#include <glib-object.h>
#include <config.h>

G_BEGIN_DECLS

#define XCS_TYPE_CONFIGURATION		  (xcs_get_type ())
#define XCS_CONFIGURATION(obj)		  (G_TYPE_CHECK_INSTANCE_CAST ((obj), XCS_TYPE_CONFIGURATION, Xcs))
#define XCS_CONFIGURATION_CLASS(klass)	  (G_TYPE_CHECK_CLASS_CAST ((klass), XCS_TYPE_CONFIGURATION, XcsClass))
#define XCS_IS_CONFIGURATION(obj)	  (G_TYPE_CHECK_INSTANCE_TYPE ((obj), XCS_TYPE_CONFIGURATION))
#define XCS_IS_CONFIGURATION_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), XCS_TYPE_CONFIGURATION))
#define XCS_CONFIGURATION_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), XCS_TYPE_CONFIGURATION, XcsClass))

typedef struct _Xcs Xcs;
typedef struct _XcsClass XcsClass;
typedef struct _XcsPrivate XcsPrivate;

struct _Xcs {
    GObject parent;

    XcsPrivate *private;
};

struct _XcsClass {
    GObjectClass parent;

    void (*removed) (Xcs *self);
    void (*changed) (Xcs *self);
    void (*added) (Xcs *self);
};

GType
xcs_get_type (void);

Xcs*
xcs_new (gboolean *new_configuration);

gboolean
xcs_save (Xcs *self);

gboolean
xcs_domain_new (Xcs *self,
                const gchar * domain_id);

gboolean
xcs_key_new (Xcs *self,
             const gchar * domain_id,
             const gchar * key_id,
             GValue * key_value);

gboolean
xcs_key_set (Xcs *self,
             const gchar * domain_id,
             const gchar * key_id,
             GValue * key_value,
             gboolean create);

gboolean
xcs_key_unset (Xcs *self,
               const gchar * domain_id,
               const gchar * key_id,
               gboolean create);

GHashTable*
xcs_domain_fetch (Xcs *configuration,
                  const gchar * domain_id);
void
xcs_ht_destroy (Xcs *self,
                GHashTable * ht);


const gchar*
xcs_key_fetch (Xcs *self,
               const gchar * domain_id,
               const gchar * key_id);
gint
xcs_key_fetch_int (Xcs *self,
                   const gchar * domain_id,
                   const gchar * key_id);
gboolean
xcs_key_fetch_boolean (Xcs *self,
                       const gchar * domain_id,
                       const gchar * key_id);

void
xcs_key_set_string (Xcs *self,
                    const gchar * domain_id,
                    const gchar * key_id,
                    const gchar *value_arg);
void
xcs_key_set_int (Xcs *configuration,
                 const gchar	*domain_id,
                 const gchar	*key_id,
                 gint		 value_arg);
void
xcs_key_set_boolean (Xcs *configuration,
                     const gchar	    *domain_id,
                     const gchar	    *key_id,
                     gboolean	     value_arg);


gboolean
xcs_domain_exists (Xcs *configuration,
                   const gchar	  *domain_id);
gboolean
xcs_domain_has_keys (Xcs *configuration,
                     const gchar	    *domain_id);

gboolean
xcs_key_find (Xcs *self,
              const gchar * domain_id,
              const gchar * key_id, 
	      const gchar *key_value); /* Either key_id or key_value can be null, but of course not both */


gboolean
xcs_remove_domain (Xcs *self,
                   const gchar * domain_id);
gboolean
xcs_remove_key (Xcs *configuration,
                const gchar      *domain_id,
                const gchar      *key_id,
		const gchar      *key_value);


#define LOOKUP_OR_CREATE_DEFAULT(domain_id,key_id,default_value) \
  if (!xcs_key_find (xcs, domain_id, key_id, NULL)) \
  { \
      xcs_key_set_string (xcs, domain_id, key_id, default_value); \
  }

#define LOOKUP_OR_CREATE_DEFAULT_INT(domain_id,key_id,default_value) \
  if (!xcs_key_find (xcs, domain_id, key_id, NULL)) \
  { \
      xcs_key_set_int (xcs, domain_id, key_id, default_value); \
  }

#define LOOKUP_OR_CREATE_DEFAULT_BOOLEAN(domain_id,key_id,default_value) \
  if (!xcs_key_find (xcs, domain_id, key_id, NULL)) \
  { \
      xcs_key_set_boolean (xcs, domain_id, key_id, default_value); \
  }

typedef void (*BindingDataFunc) (GObject *object, GValue *value);

void
xcs_bind_object (Xcs *self, 
			       GObject *object, 
			       BindingDataFunc func, 
			       const gchar *signal_spec, 
			       const gchar *domain, 
			       const gchar *key);

#ifdef HAVE_GUI
/* Convenience functions since we deal a lot with ToggleActions and ToggleButtons */
void
xcs_bind_toggle_action (Xcs *self, 
				      GtkToggleAction *action, 
				      const gchar *domain, 
				      const gchar *key);

void
xcs_bind_toggle_button (Xcs *self, 
				      GtkToggleButton *button, 
				      const gchar *domain, 
				      const gchar *key);
#endif
 
#endif /* _XCS_H_ */
